/*
 * (c) Copyright 2003, 2005 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <string.h>

#include "bfirdata.h"
#include "bfirmem.h"
#include "idarray.h"
#include "elog.h"

static int n_reverb_channels[2] = { 0, 0 };
static int n_rates = 0;
static int rates[FILTERPRG_RATES_MAX];
static int block_sizes[FILTERPRG_RATES_MAX];
static idarray_t *filterprgs = NULL;
static int filterprgs_count[BFIR_FILTER_TYPE_COUNT];
static int filter_bytes = 0;
static struct _bfirfilter dummy_filters[10 * BFIR_MAX_OUTPUT_CHANNELS];
static int n_dummy_filters = 0;
static int realsize;

static double
get_capacity(double *max_capacity,
	     int n_channels)
{
    int n, totrate;

    for (n = totrate = 0; n < n_rates; n++) {
	totrate += rates[n];
    }
    if (max_capacity != NULL) {
	*max_capacity = (double)filter_bytes /
	    (double)(2 * totrate * realsize) /
	    (double)(n_channels);
    }
    return (double)bfirmem_bytes_left() /
	(double)(2 * totrate * realsize) /
	(double)(n_channels);
}

static bool_t
alloc_filter(struct bfirfilter *filter,
             const int lengths[],
             int type)
{
    int n;

    filter->n_rates = n_rates;
    for (n = 0; n < n_rates; n++) {
        if (lengths[n] == 0) {
            elog(LOGDEBUG, "WARNING: length is zero for rate index %d (%d Hz)",
                 n, rates[n]);
        }
        filter->type = type;
	filter->rates[n] = rates[n];
	filter->f[n].block_size = block_sizes[n];
	filter->f[n].n_blocks = (lengths[n] + filter->f[n].block_size - 1) /
            filter->f[n].block_size;
	/* allocate for frequency domain representation of filter, hence twice
	   the size */
	if (!bfirmem_alloc(&filter->f[n].memseg,
			   2 * filter->f[n].n_blocks * filter->f[n].block_size *
			   realsize,
			   2 * filter->f[n].block_size * realsize))
	{
	    elog(LOGDEBUG, "failed to allocate filter memory");
	    while (n--) {
		bfirmem_free(&filter->f[n].memseg);
	    }
	    return false;
	}
    }
    return true;
}

static void
free_filter(struct bfirfilter *filter)
{
    int n = n_rates;
    
    while (n--) {
	bfirmem_free(&filter->f[n].memseg);
    }
}    

static int
get_filter_index(int type,
                 int id)
{
    struct bfirfilter *filter;
    int idx, n;

    n = -1;
    idx = 0;
    while ((filter = (struct bfirfilter *)idarray_next(filterprgs, &n))
           != NULL)
    {
        if (filter->type == type) {
            if (n == id) {
                return idx;
            }
            idx++;
        }
    }
    return -1;
}

static struct bfirfilter *
get_next_filter(int type,
                int *idx)
{
    struct bfirfilter *filter;
    
    while ((filter = (struct bfirfilter *)idarray_next(filterprgs, idx))
           != NULL &&
           filter->type != type);
    return filter;
}

static int
get_coeff_base(int type,
               int id)
{
    int coeff_base, idx;

    if (id < 0) {
        return -1;
    }
    switch (type) {
    case BFIR_FILTER_TYPE_REVERB:
        coeff_base = 0;
        break;
    case BFIR_FILTER_TYPE_EQ:
        coeff_base = n_reverb_channels[IN] * n_reverb_channels[OUT] *
            filterprgs_count[BFIR_FILTER_TYPE_REVERB];
        break;
    case BFIR_FILTER_TYPE_XTC:
        coeff_base = n_reverb_channels[IN] * n_reverb_channels[OUT] *
            filterprgs_count[BFIR_FILTER_TYPE_REVERB] +
            filterprgs_count[BFIR_FILTER_TYPE_EQ];
        break;
    case BFIR_FILTER_TYPE_UHJ:
        coeff_base = n_reverb_channels[IN] * n_reverb_channels[OUT] *
            filterprgs_count[BFIR_FILTER_TYPE_REVERB] +
            filterprgs_count[BFIR_FILTER_TYPE_EQ] +
            4 * filterprgs_count[BFIR_FILTER_TYPE_XTC];
        break;
    default:
        coeff_base = n_reverb_channels[IN] * n_reverb_channels[OUT] *
            filterprgs_count[BFIR_FILTER_TYPE_REVERB] +
            filterprgs_count[BFIR_FILTER_TYPE_EQ] +
            4 * filterprgs_count[BFIR_FILTER_TYPE_XTC] +
            3 * filterprgs_count[BFIR_FILTER_TYPE_UHJ];
        return coeff_base;
    }
    if ((idx = get_filter_index(type, id)) == -1) {
        return -1;
    }
    coeff_base += idx * bfirdata_filter_count(type);
    
    return coeff_base;
}

static bool_t
shorten_filter(struct bfirfilter *filter,
               const int lengths[])
{
    static uint8_t *buf = NULL, *ptr;
    int buf_size;
    int n, i, n_blocks;

    for (n = 0; n < n_rates; n++) {
	n_blocks = (lengths[n] + filter->f[n].block_size - 1) /
            filter->f[n].block_size;
        if (n_blocks > filter->f[n].n_blocks) {
            elog(LOGDEBUG, "length is longer than previous");
            return false;
        }
    }
    filter->n_rates = n_rates;
    for (n = 0; n < n_rates; n++) {
	n_blocks = (lengths[n] + filter->f[n].block_size - 1) /
            filter->f[n].block_size;
        if (filter->f[n].n_blocks == n_blocks) {
            continue;
        }
        if (lengths[n] == 0) {
            elog(LOGDEBUG, "WARNING: length is zero");
        }
        buf_size = 2 * filter->f[n].block_size * filter->f[n].n_blocks *
            realsize;
        buf = realloc(buf, buf_size);
        if (buf == NULL) {
            elog(LOGERROR, "Out of memory");
            exit(EXIT_FAILURE);
        }
        ptr = buf;
        for (i = 0; i < filter->f[n].memseg.n_segments; i++) {
            memcpy(ptr, filter->f[n].memseg.segments[i],
                   filter->f[n].memseg.seglengths[i]);
            ptr += filter->f[n].memseg.seglengths[i];
        }
        bfirmem_free(&filter->f[n].memseg);
	filter->f[n].n_blocks = n_blocks;
	if (!bfirmem_alloc(&filter->f[n].memseg,
			   2 * filter->f[n].n_blocks * filter->f[n].block_size *
			   realsize,
			   2 * filter->f[n].block_size * realsize))
	{
            /* Since we are shortening the filter, this should not happen */
	    elog(LOGERROR, "failed to allocate filter memory");
            exit(EXIT_FAILURE);
	}
        ptr = buf;
        for (i = 0; i < filter->f[n].memseg.n_segments; i++) {
            memcpy(filter->f[n].memseg.segments[i], ptr,
                   filter->f[n].memseg.seglengths[i]);
            ptr += filter->f[n].memseg.seglengths[i];
        }
    }
    return true;
}

int
bfirdata_new_dummy_filter(int n_blocks,
                          int block_size)
{
    struct memseg *memseg;
    int retval, n;
    
    elog(LOGDEBUG, "dummy filter alloc: %d blocks of size %d",
         n_blocks, block_size);
    if (!bfirmem_alloc(&dummy_filters[n_dummy_filters].memseg,
                       2 * n_blocks * block_size * realsize,
                       2 * block_size * realsize))
    {
        return -1;
    }
    memseg = &dummy_filters[n_dummy_filters].memseg;
    for (n = 0; n < memseg->n_segments; n++) {
        memset(memseg->segments[n], 0, memseg->seglengths[n]);
    }
    dummy_filters[n_dummy_filters].n_blocks = n_blocks;
    dummy_filters[n_dummy_filters].block_size = block_size;
    retval = n_dummy_filters;
    n_dummy_filters++;
    return retval;
}

const struct brutefir_coeff *
bfirdata_get_dummy_coeffs(int *n_coeffs)
{
    static struct brutefir_coeff *coeffs = NULL, *coeff;
    struct _bfirfilter *_f;
    int n, i;

    elog(LOGDEBUG, "getting %d dummy filters", n_dummy_filters);
    
    if (n_dummy_filters == 0) {
        return NULL;        
    }

    coeffs = realloc(coeffs, n_dummy_filters * sizeof(struct brutefir_coeff));
    memset(coeffs, 0, n_dummy_filters * sizeof(struct brutefir_coeff));
    
    for (n = 0; n < n_dummy_filters; n++) {
        coeff = &coeffs[n];
        _f = &dummy_filters[n];
        coeff->n_blocks = _f->n_blocks;
        coeff->n_segments = _f->memseg.n_segments;
        for (i = 0; i < coeff->n_segments; i++) {
            coeff->segids[i] = _f->memseg.segids[i];
            coeff->segoffsets[i] = _f->memseg.segoffsets[i];
            coeff->seglengths[i] = _f->memseg.seglengths[i] /
                _f->block_size / 2 / realsize;
        }
    }
    *n_coeffs = n_dummy_filters;
    return coeffs;
}

int
bfirdata_filter_count(int type)
{
    switch (type) {
    case BFIR_FILTER_TYPE_REVERB:
        return n_reverb_channels[IN] * n_reverb_channels[OUT];
    case BFIR_FILTER_TYPE_EQ:
        return 1;
    case BFIR_FILTER_TYPE_XTC:
        return 4;
    case BFIR_FILTER_TYPE_UHJ:
        return 3;
    }
    return -1;
}

int
bfirdata_longest_filter_blocks(int sample_rate)
{
    struct bfirfilter *filter;
    int n, maxblocks, ri;

    n = -1;
    maxblocks = 1;
    while ((filter = (struct bfirfilter *)idarray_next(filterprgs, &n))
           != NULL)
    {
        for (ri = 0; ri < filter->n_rates; ri++) {
            if (filter->rates[ri] == sample_rate) {
                break;
            }
        }
        if (ri == filter->n_rates) {
            continue;
        }
        if (filter->f[ri].n_blocks > maxblocks) {
            maxblocks = filter->f[ri].n_blocks;
        }
    }
    return maxblocks;
}

int
bfirdata_coeff_index(int type,
                     int id,
                     int sub_index)
{
    int coeff_base;

    if ((coeff_base = get_coeff_base(type, id)) < 0) {
        return -1;
    }
    return coeff_base + sub_index;
}

int
bfirdata_new_filters(int type,
                     struct bfirfilter _filters[],
                     const int lengths[])
{
    struct bfirfilter *filters;
    int n, id, n_filters; 
    
    if ((n_filters = bfirdata_filter_count(type)) == -1) {
        return -1;
    }
    if (n_filters == 0) {
        elog(LOGDEBUG, "no filters with current config");
        return -1;
    }
    if ((filters = malloc(n_filters * sizeof(struct bfirfilter))) == NULL) {
	elog(LOGDEBUG, "memory allocation failed");
	return -1;
    }
    if ((id = idarray_new_id(filterprgs, filters)) == -1) {
	elog(LOGDEBUG, "failed to get new id");
	free(filters);
	return -1;
    }
    for (n = 0; n < n_filters; n++) {
	if (!alloc_filter(&filters[n], lengths, type)) {
	    while (n--) {
		free_filter(&filters[n]);
	    }
	    free(filters);
	    idarray_delete_id(filterprgs, id);
	    return -1;
	}
    }
    filterprgs_count[type] += 1;
    memcpy(_filters, filters, n_filters * sizeof(struct bfirfilter));

    elog(LOGDEBUG, "memseg.segids[0] %d", filters[0].f[0].memseg.segids[0]);
    elog(LOGDEBUG, "added filter of type %s (id %d)",
         bfirdata_typename(type), id);
    
    return id;
}

void
bfirdata_delete_filters(int id)
{
    struct bfirfilter *filters;
    int n_filters;
    
    if ((filters = (struct bfirfilter *)idarray_delete_id(filterprgs, id))
	== NULL)
    {
	return;
    }
    n_filters = bfirdata_filter_count(filters->type);
    filterprgs_count[filters->type] -= 1;
    while (n_filters--) {
	free_filter(&filters[n_filters]);
    }
    free(filters);
}

int
bfirdata_shorten_filters(int id,
                         const int lengths[])
{
    struct bfirfilter *filter;
    int n_filters, n;
    
    if ((filter = (struct bfirfilter *)idarray_get(filterprgs, id)) == NULL) {
        return VCU_ERR_INVALID_VALUE;
    }
    n_filters = bfirdata_filter_count(filter->type);
    for (n = 0; n < n_filters; n++) {
        if (!shorten_filter(&filter[n], lengths)) {
            return VCU_ERR_INVALID_VALUE;
        }
    }
    return 0;
}

const struct brutefir_coeff *
bfirdata_get_coeffs(int *n_coeffs,
                    int sample_rate)
{
    static struct brutefir_coeff *coeffs = NULL, *coeff;
    int ri, coeffs_offset, n, i, j, k, type, len;
    struct _bfirfilter *_f;
    struct bfirfilter *f;

    for (ri = 0; ri < n_rates; ri++) {
	if (rates[ri] == sample_rate) {
	    break;
	}
    }
    if (ri == n_rates) {
	elog(LOGDEBUG, "rate not found\n");
	return NULL;
    }

    len = 0;
    for (type = 0; type < BFIR_FILTER_TYPE_COUNT; type++) {
        len += bfirdata_filter_count(type) * filterprgs_count[type];
    }
    *n_coeffs = len;
    if (len == 0) {
        len = 1; /* just so we get a valid pointer */
    }
    coeffs = realloc(coeffs, len * sizeof(struct brutefir_coeff));
    memset(coeffs, 0, len * sizeof(struct brutefir_coeff));
    if (*n_coeffs == 0) {
        /* no filters */
        return coeffs;
    }
    
    coeffs_offset = 0;
    for (type = 0; type < BFIR_FILTER_TYPE_COUNT; type++) {
        i = -1;
        for (n = 0; n < filterprgs_count[type]; n++) {            
            for (k = 0; (f = get_next_filter(type, &i)) != NULL; k++) {
                for (j = 0; j < bfirdata_filter_count(type); j++) {
                    coeff = &coeffs[coeffs_offset++];
                    _f = &f[j].f[ri];
                    coeff->type = type;
                    coeff->id = i;
                    coeff->index = k;
                    coeff->sub_index = j;
                    coeff->n_blocks = _f->n_blocks;
                    coeff->n_segments = _f->memseg.n_segments;
                    for (n = 0; n < coeff->n_segments; n++) {
                        coeff->segids[n] = _f->memseg.segids[n];
                        coeff->segoffsets[n] = _f->memseg.segoffsets[n];
                        coeff->seglengths[n] = _f->memseg.seglengths[n] /
                            _f->block_size / 2 / realsize;
                    }
                }
            }
        }
    }
    return coeffs;
}

double
bfirdata_filter_capacity(int type,
                         double *max_capacity)
{
    return get_capacity(max_capacity, bfirdata_filter_count(type));
}

void
bfirdata_reset(void)
{
    bfirdata_settings(n_reverb_channels[IN], n_reverb_channels[OUT],
                      n_rates, rates, block_sizes, realsize);
}

void
bfirdata_settings(int _n_reverb_input_channels,
                  int _n_reverb_output_channels,
                  int _n_rates,
                  const int _rates[],
                  const int _block_sizes[],
                  int _realsize)
{
    int id, n;

    if (filterprgs == NULL) {
        elog(LOGERROR, "not initialised");
        exit(EXIT_FAILURE);
    }
    
    n_reverb_channels[IN] = _n_reverb_input_channels;
    n_reverb_channels[OUT] = _n_reverb_output_channels;
    n_rates = _n_rates;
    realsize = _realsize;
    for (n = 0; n < n_rates; n++) {
        rates[n] = _rates[n];
        block_sizes[n] = _block_sizes[n];
    }

    /* free all filters */
    n_dummy_filters = 0;
    bfirmem_freeall();    
    for (id = -1;
	 idarray_next(filterprgs, &id) != NULL;
	 free(idarray_delete_id(filterprgs, id)));
    memset(filterprgs_count, 0, sizeof(filterprgs_count));
    filter_bytes = bfirmem_bytes_left();
    elog(LOGDEBUG, "filter_bytes: %d", filter_bytes);
}

bool_t
bfirdata_init(int alloc_bytes)
{
    memset(filterprgs_count, 0, sizeof(filterprgs_count));
    if ((filterprgs = idarray_new(false)) == NULL) {
        elog(LOGERROR, "idarray allocation failure");
        return false;
    }   
    if (alloc_bytes < BFIR_MIN_REVERB_LENGTH * sizeof(float) * 20) {
	elog(LOGERROR, "alloc_bytes is unreasonable small (%d)", alloc_bytes);
	return false;
    }
    filter_bytes = alloc_bytes;
    return bfirmem_init(alloc_bytes);
}
