/*
 * (c) Copyright 2002 - 2004 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dlfcn.h>
#include <unistd.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

#include "bit.h"
#include "defs.h"
#include "conf.h"
#include "log2.h"
#include "elog.h"
#include "path.h"
#include "bfir.h"
#include "reverb.h"
#include "grammar.h"
#include "copyfile.h"
#include "filterprg.h"

#include <bfmod.h>

/* FIXME: make parsing tolerant (so it can accept constant changes, minor format
   changes and so on */

#define BOOLEAN_BIT 10000

struct conf_static conf_static;
struct conf_user cu;

static bool_t has_loaded_static = false;
static FILE *custom_user_config_stream;

static int
dummy_get_config_token(union vcu_lexval *lexval)
{
    if (lexval == NULL) {
	elog(LOGERROR, "NULL lexval, bug in module.\n");
	exit(EXIT_FAILURE);
    }
    return 0;
}

static int
get_config_token(union vcu_lexval *lexval)
{
    static int brace_depth = 0;
    int tok;

    if (lexval == NULL) {
	elog(LOGERROR, "NULL lexval, bug in module.\n");
	exit(EXIT_FAILURE);
    }
    switch (tok = yylex()) {
    case VCU_LEX_LBRACE:
	brace_depth++;
	break;
    case VCU_LEX_RBRACE:
	if (brace_depth == 0) {
	    return 0;
	}
	brace_depth--;
	break;
    }
    memcpy(lexval, &yylval, sizeof(yylval));

    return tok;
}

static void
put_config_token(int tok,
		 const union vcu_lexval *lexval)
{
    FILE *stream = custom_user_config_stream;
    
    switch (tok) {
    case VCU_LEX_EOS:
	fprintf(stream, ";\n");
	break;
    case VCU_LEX_COMMA:
	fprintf(stream, ", ");
	break;
    case VCU_LEX_SLASH:
	fprintf(stream, "/");
	break;
    case VCU_LEX_LBRACE:
	fprintf(stream, "{");
	break;
    case VCU_LEX_RBRACE:
	fprintf(stream, "}");
	break;
    case VCU_LEXVAL_BOOLEAN:
	fprintf(stream, (lexval->boolean) ? "true" : "false");
	break;
    case VCU_LEXVAL_REAL:
	fprintf(stream, "%g", lexval->real);
	break;
    case VCU_LEXVAL_STRING:
	fprintf(stream, "\"%s\"", lexval->string);
	break;
    case VCU_LEXVAL_FIELD:
	fprintf(stream, "    %s: ", lexval->field);
	break;
    default:
	elog(LOGDEBUG, "Ignoring invalid token %d from module save config.",
             tok);
	break;
    }
}

static void
limit_range(double *real,
	    double min,
	    double max)
{
    if (*real < min) {
	*real = min;
    }
    if (*real > max) {
	*real = max;
    }
}

static void
limit_range_int(int *integer,
		int min,
		int max)
{
    if (*integer < min) {
	*integer = min;
    }
    if (*integer > max) {
	*integer = max;
    }
}

static void
init_channel_mapping(void)
{
    int io, type, logi, phys;

    memset(cu.channel.map, 0, sizeof(cu.channel.map));
    FOR_IN_AND_OUT {
        for (type = 0; type < VCU_CH_COUNT; type++) {
            for (logi = 0; logi < cu.channel.count[io][type]; logi++) {
                phys = logi % cu.channel.count[io][VCU_CH_PHYSICAL];
                bit_set(cu.channel.map[io][type][logi], phys);
            }
        }
    }
}

static void
stabilise_conf_user(void)
{
    int n, i, io, j, type, count;
    prgid_t zeroid;

    memset(&zeroid, 0, sizeof(zeroid));
    
    /* volume */
    limit_range(&cu.volume.master, VCU_VOLUME_MIN, VCU_VOLUME_MAX);
    for (n = 0; n < VCU_MAX_CHANNEL_GROUPS; n++) {
        limit_range(&cu.volume.group[n], VCU_VOLUME_MIN, VCU_VOLUME_MAX);
    }

    /* signal */
    for (n = 0; n < cu.signal.n_rates; n++) {
	for (i = 0; i < conf_static.n_rates; i++) {
	    if (cu.signal.rates[n] == conf_static.rates[i]) {
		break;
	    }
	}
	if (i == conf_static.n_rates ||
	    (n > 0 && cu.signal.rates[n] <= cu.signal.rates[n-1]))
	{
	    cu.signal.n_rates = 1;
	    cu.signal.rates[0] = conf_static.rates[0];
	    break;
	}
    }
    for (n = 0; n < cu.signal.n_rates; n++) {
	if (cu.signal.suggested_rate == cu.signal.rates[n]) {
	    break;
	}
    }
    if (n == cu.signal.n_rates) {
	cu.signal.suggested_rate =
	    cu.signal.rates[cu.signal.n_rates - 1];
    }
    for (n = 0; n < conf_static.output.n_sample_formats; n++) {
        if (cu.signal.output_sample_format ==
            conf_static.output.sample_formats[n])
        {
            break;
        }
    }
    if (n == conf_static.output.n_sample_formats) {
        cu.signal.output_sample_format =
            conf_static.output.sample_formats[0];
    }

    /* channel */
    limit_range_int(&cu.channel.active_mode, -1, cu.channel.mode_count - 1);
    for (n = 0; n < cu.channel.mode_count; n++) {
        limit_range_int(&cu.channel.mode[n].active_eq_mode, -1,
                        cu.eq.mode_count - 1);
        if (!bit_isset(cu.channel.mode[n].eqmodes,
                       cu.channel.mode[n].active_eq_mode))
        {
            cu.channel.mode[n].active_eq_mode = -1;
        }
    }
    cu.channel.count[IN][VCU_CH_PHYSICAL] =
        conf_static.input.total_open_channels;
    cu.channel.count[OUT][VCU_CH_PHYSICAL] =
        conf_static.output.total_open_channels;
    for (type = 0; type < VCU_CH_COUNT; type++) {
        limit_range_int(&cu.channel.count[OUT][type], 0, VCU_MAX_CHANNELS);
        if (type == VCU_CH_AMBIOPOLE &&
            cu.channel.count[OUT][type] % 2 != 0)            
        {
            cu.channel.count[OUT][type] -= 1;
        }
        if (cu.channel.count[OUT][type] > 0) {
            switch (type) {
            case VCU_CH_REVERB:
                switch (cu.reverb.source_layout) {
                case VCU_REVERB_SOURCE_LAYOUT_CENTER:
                    cu.channel.count[IN][type] = 1;
                    break;
                case VCU_REVERB_SOURCE_LAYOUT_STEREO:
                    cu.channel.count[IN][type] = 2;
                    break;
                case VCU_REVERB_SOURCE_LAYOUT_LRC:
                    cu.channel.count[IN][type] = 3;
                    break;
                default:
                    cu.reverb.source_layout = VCU_REVERB_SOURCE_LAYOUT_STEREO;
                    cu.channel.count[IN][type] = 2;
                    break;
                }
                break;
            case VCU_CH_AMBISONIC:
                switch (cu.ambisonics.input_format) {
                case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:
                    cu.channel.count[IN][type] = 4;
                    break;
                case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
                    cu.channel.count[IN][type] = 9;
                    break;
                case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
                    cu.channel.count[IN][type] = 2;
                    break;
                default:
                    cu.ambisonics.input_format =
                        VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER;
                    cu.channel.count[IN][type] = 4;
                    break;
                }
                break;
            default:
                cu.channel.count[IN][type] = cu.channel.count[OUT][type];
                break;
            }
        } else {
            cu.channel.count[IN][type] = 0;
        }
    }
    FOR_IN_AND_OUT {
        count = 0;
        for (type = 0; type < VCU_CH_COUNT; type++) {
            for (n = 0; n < cu.channel.count[io][type]; n++) {
                i = bit_count(cu.channel.map[io][type][n],
                              cu.channel.count[io][VCU_CH_PHYSICAL]);
                count += i;
                if (i == 0 || count > VCU_MAX_CHANNELS) {
                    break;
                }
            }
            if (n != cu.channel.count[io][type]) {
                break;
            }
        }
        if (type != VCU_CH_COUNT) {
            init_channel_mapping();
            break;
        }
    }
    for (type = 0; type < VCU_CH_WP_COUNT; type++) {
        FOR_IN_AND_OUT {
            limit_range(&cu.channel.trim.volume[io][type][n],
                        VCU_VOLUME_MIN, VCU_VOLUME_MAX);
            limit_range(&cu.channel.trim.delay[io][type][n],
                        0, VCU_TRIM_DELAY_RANGE);
        }
    }
    for (n = 0; n < VCU_MAX_CHANNELS; n++) {
	limit_range(&cu.channel.speakerpos[n].azimut, -180, +180);
	limit_range(&cu.channel.speakerpos[n].elevation, -90, +90);
	limit_range(&cu.channel.speakerpos[n].distance,
                    0, VCU_DISTANCE_MAX);
    }
    for (n = 0; n < VCU_MAX_STAGES; n++) {
	limit_range(&cu.channel.stage[n].azimut, -180, +180);
	limit_range(&cu.channel.stage[n].elevation, -90, +90);
	limit_range(&cu.channel.stage[n].width, 0, +360);
	limit_range(&cu.channel.stage[n].height, 0, +180);
    }
    
    /* eq */
    limit_range_int(&cu.eq.active_mode, -1, cu.eq.mode_count - 1);
    for (n = 0; n < cu.eq.eqprg_count; n++) {
	if (cu.eq.eqprg[n].part_count > VCU_EQ_COMPONENTS_MAX) {
	    cu.eq.eqprg[n].part_count = VCU_EQ_COMPONENTS_MAX;
	}
        /* some sanity checks are done when parsing */
        for (i = 0; i < cu.eq.eqprg[n].part_count; i++) {
            switch (cu.eq.eqprg[n].parts[i].type) {
            case VCU_EQPRG_TYPE_OP_MULTIPLY:
                limit_range(&cu.eq.eqprg[n].parts[i].params.multiply.amp,
                            VCU_VOLUME_MIN, VCU_VOLUME_MAX);
                break;
            case VCU_EQPRG_TYPE_FFT_STATIC:
            case VCU_EQPRG_TYPE_FFT_DYNAMIC:
                for (j = 0;
                     j < cu.eq.eqprg[n].parts[i].params.fft.n_bands;
                     j++)
                {
                    limit_range(&cu.eq.eqprg[n].parts[i].params.fft.amp[j],
                                VCU_FFTEQ_VOLUME_MIN, VCU_FFTEQ_VOLUME_MAX);
                }
                break;
            }
        }
    }
    for (i = 0; i < cu.eq.mode_count; i++) {
        FOR_IN_AND_OUT {
            for (type = 0; type < VCU_CH_WP_COUNT; type++) {
                for (n = 0; n < VCU_MAX_CHANNELS; n++) {
                    if (cu.eq.mode[i].prgs[io][type][n] >= cu.eq.eqprg_count ||
                        cu.eq.mode[i].prgs[io][type][n] < -1)
                    {
                        cu.eq.mode[i].prgs[io][type][n] = -1;
                    }
                }
            }
        }
    }

    /* ambiopole */
    for (n = 0; n < BFIR_MAX_OUTPUT_CHANNELS/2; n++) {
        for (i = 0; i < cu.ambiopole.n_loaded_programs; i++) {
            if (memcmp(&cu.ambiopole.current[n],
                       &cu.ambiopole.loaded_programs[i],
                       sizeof(prgid_t)) == 0)
            {
                break;
            }
        }
        if (i == cu.ambiopole.n_loaded_programs) {
            cu.ambiopole.current[n] = zeroid;
        }
        if (memcmp(&cu.ambiopole.current[n], &zeroid,
                   sizeof(prgid_t)) == 0 &&
            cu.ambiopole.use_xtalk_cancel[n])
        {
            cu.ambiopole.use_xtalk_cancel[n] = false;
        }
    }

    /* reverb */
    for (n = 0; n < cu.signal.n_rates; n++) {
	cu.reverb.block_sizes[n] = 1 << log2_roof(cu.reverb.block_sizes[n]);
	i = cu.signal.rates[n] / 48010 + 1;
	limit_range_int(&cu.reverb.block_sizes[n],
			BFIR_MIN_BLOCKSIZE_LOWLAT,
			i * BFIR_MAX_BLOCKSIZE_48000HZ);
	i = BFIR_MIN_REVERB_LENGTH / cu.reverb.block_sizes[n];
	if (i < BFIR_MIN_BLOCKS) {
	    i = BFIR_MIN_BLOCKS;
	}
	limit_range_int(&cu.reverb.n_blocks[n], i, BFIR_MAX_BLOCKS);
    }
    limit_range(&cu.reverb.max_benchmark_load, 0, 1);
    limit_range(&cu.reverb.max_benchmark_length, VCU_BENCHMARK_MIN_SECONDS,
                VCU_BENCHMARK_MAX_SECONDS);
    i = cu.reverb.n_blocks[cu.signal.n_rates - 1] *
	cu.reverb.block_sizes[cu.signal.n_rates - 1];
    for (n = 0; n < cu.reverb.n_loaded_programs; n++) {
	if (memcmp(&cu.reverb.current,
		   &cu.reverb.loaded_programs[n], sizeof(prgid_t)) == 0)
	{
	    break;
	}
    }
    if (n == cu.reverb.n_loaded_programs) {
	cu.reverb.current = zeroid;
    }
    if (cu.reverb.response_mode == 0) {
        cu.reverb.response_mode = VCU_REVERB_RESP_MODE_FULL;
    }
    if (cu.reverb.decorrelate_mode == 0) {
        cu.reverb.decorrelate_mode = VCU_REVERB_DECORR_MODE_NEEDED;
    }
    limit_range(&cu.reverb.bformat_directivity,
                REVERB_BFORMAT_DIRECTIVITY_MIN,
                REVERB_BFORMAT_DIRECTIVITY_MAX);
    limit_range(&cu.reverb.dry_volume, VCU_VOLUME_MIN, VCU_VOLUME_MAX);

    /* ctl */
    for (n = 0; n < cu.ctl.n_fprogrefs; n++) {
        if (cu.ctl.fprogrefs[n] == NULL) {
            cu.ctl.n_fprogrefs = n;
            break;
        }
    }
}

static void
init_conf_user(void)
{
    int rates[VCU_VALID_RATES_COUNT] = VCU_VALID_RATES;
    int n;
    
    memset(&cu, 0, sizeof(cu));
    cu.volume.master = VCU_VOLUME_MIN;
    for (n = 0; n < VCU_MAX_CHANNEL_GROUPS; n++) {
        cu.volume.group[n] = VCU_VOLUME_MIN;
    }
    cu.signal.apply_dither = true;
    cu.signal.n_rates = 1;
    cu.signal.rates[0] = rates[0];
    cu.signal.suggested_rate = rates[0];
    cu.signal.output_sample_format =
        conf_static.output.sample_formats[0];
    cu.channel.count[OUT][VCU_CH_AMBIOPOLE] = 2;
    cu.channel.count[OUT][VCU_CH_REVERB] = 2;
    cu.channel.active_mode = -1;
    cu.eq.active_mode = -1;
    cu.reverb.max_benchmark_load = 0.85;
    cu.reverb.max_benchmark_length = VCU_BENCHMARK_MAX_SECONDS;
    cu.reverb.block_sizes[0] = BFIR_MIN_BLOCKSIZE_48000HZ;
    cu.reverb.n_blocks[0] = BFIR_MIN_BLOCKS;
    cu.reverb.response_mode = VCU_REVERB_RESP_MODE_FULL;
    cu.reverb.decorrelate_mode = VCU_REVERB_DECORR_MODE_NEEDED;
    cu.reverb.force_decay = true;
    cu.reverb.bformat_directivity = REVERB_BFORMAT_DIRECTIVITY_DEFAULT;
    cu.reverb.source_layout = VCU_REVERB_SOURCE_LAYOUT_STEREO;
    cu.ambisonics.input_format = VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER;
    cu.ctl.powersave = true;
    cu.ctl.load_user_config = true;
    stabilise_conf_user();
}

static int
parse_phase(const char str[])
{
    if (strcmp(str, "minimum") == 0) {
	return VCU_PHASE_MINIMUM;
    } else if (strcmp(str, "maximum") == 0) {
	return VCU_PHASE_MAXIMUM;
    } else if (strcmp(str, "linear") == 0) {
	return VCU_PHASE_LINEAR;
    } else {
	return -1;
    }
}

static const char *
phasestr(int phase)
{
    switch (phase) {
    case VCU_PHASE_MINIMUM:
	return "minimum";
    case VCU_PHASE_MAXIMUM:
	return "maximum";
    case VCU_PHASE_LINEAR:
	return "linear";
    default:
	return "#invalid phase#";
    }
}

static bool_t
get_eqpart(int eq_index)
{
    struct vcu_eqpart eqpart;
    int n;
    
    memset(&eqpart, 0, sizeof(eqpart));
    GET_TOKEN(LBRACE);
    GET_TOKEN(STRING);
    if (strcmp(yylval.string, "convolve") == 0) {
        eqpart.type = VCU_EQPRG_TYPE_OP_CONVOLVE;
    } else if (strcmp(yylval.string, "add") == 0) {
        eqpart.type = VCU_EQPRG_TYPE_OP_ADD;
    } else if (strcmp(yylval.string, "subtract") == 0) {
        eqpart.type = VCU_EQPRG_TYPE_OP_SUBTRACT;
    } else if (strcmp(yylval.string, "multiply") == 0) {
        eqpart.type = VCU_EQPRG_TYPE_OP_MULTIPLY;
        GET_TOKEN(SLASH);
	GET_TOKEN(LBRACE);
	GET_TOKEN(REAL);
	eqpart.params.multiply.amp = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(REAL);
	eqpart.params.multiply.sign = (int)yylval.real;
	GET_TOKEN(RBRACE);
    } else if (strcmp(yylval.string, "lowpass") == 0) {
	eqpart.type = VCU_EQPRG_TYPE_LOWPASS;
        GET_TOKEN(SLASH);
	GET_TOKEN(LBRACE);
	GET_TOKEN(REAL);
	eqpart.params.lowpass.cutoff = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(REAL);
	eqpart.params.lowpass.slope = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(STRING);
	if ((eqpart.params.lowpass.phase = parse_phase(yylval.string)) == -1) {
	    return false;
	}
	GET_TOKEN(RBRACE);
    } else if (strcmp(yylval.string, "highpass") == 0) {
	eqpart.type = VCU_EQPRG_TYPE_HIGHPASS;
        GET_TOKEN(SLASH);
	GET_TOKEN(LBRACE);
	GET_TOKEN(REAL);
	eqpart.params.highpass.cutoff = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(REAL);
	eqpart.params.highpass.slope = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(STRING);
	if ((eqpart.params.highpass.phase = parse_phase(yylval.string)) == -1) {
	    return false;
	}
	GET_TOKEN(RBRACE);
    } else if (strcmp(yylval.string, "bandpass") == 0) {
	eqpart.type = VCU_EQPRG_TYPE_BANDPASS;
        GET_TOKEN(SLASH);
	GET_TOKEN(LBRACE);
	GET_TOKEN(REAL);
	eqpart.params.bandpass.cutoff[0] = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(REAL);
	eqpart.params.bandpass.cutoff[1] = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(REAL);
	eqpart.params.bandpass.slope = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(STRING);
	if ((eqpart.params.bandpass.phase = parse_phase(yylval.string)) == -1) {
	    return false;
	}
	GET_TOKEN(RBRACE);
    } else if (strcmp(yylval.string, "bandstop") == 0) {
	eqpart.type = VCU_EQPRG_TYPE_BANDSTOP;
        GET_TOKEN(SLASH);
	GET_TOKEN(LBRACE);
	GET_TOKEN(REAL);
	eqpart.params.bandstop.cutoff[0] = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(REAL);
	eqpart.params.bandstop.cutoff[1] = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(REAL);
	eqpart.params.bandstop.slope = yylval.real;
	GET_TOKEN(COMMA);
	GET_TOKEN(STRING);
	if ((eqpart.params.bandstop.phase = parse_phase(yylval.string)) == -1) {
	    return false;
	}
	GET_TOKEN(RBRACE);
    } else if (strcmp(yylval.string, "allpass") == 0) {
	eqpart.type = VCU_EQPRG_TYPE_BANDSTOP;
        GET_TOKEN(SLASH);
	GET_TOKEN(LBRACE);
	GET_TOKEN(RBRACE);
    } else if (strcmp(yylval.string, "custom") == 0) {
	eqpart.type = VCU_EQPRG_TYPE_CUSTOM;
        GET_TOKEN(SLASH);
	GET_TOKEN(LBRACE);
	GET_TOKEN(STRING);
	eqpart.params.custom.id = filterprg_prgid_aton(yylval.string);
	GET_TOKEN(RBRACE);
    } else if (strcmp(yylval.string, "random_phase") == 0) {
	eqpart.type = VCU_EQPRG_TYPE_RANDOM_PHASE;
        GET_TOKEN(SLASH);
	GET_TOKEN(LBRACE);
	GET_TOKEN(REAL);
	eqpart.params.random_phase.length = yylval.real;
	GET_TOKEN(RBRACE);
    } else if (strcmp(yylval.string, "fft_dynamic") == 0 ||
               strcmp(yylval.string, "fft_static") == 0)
    {
        if (strcmp(yylval.string, "fft_static") == 0) {
            eqpart.type = VCU_EQPRG_TYPE_FFT_STATIC;
        } else {
            eqpart.type = VCU_EQPRG_TYPE_FFT_DYNAMIC;
        }
        GET_TOKEN(SLASH);
        GET_TOKEN(LBRACE);
        GET_TOKEN(REAL);
        eqpart.params.fft.n_bands = (int)yylval.real;
        if (eqpart.params.fft.n_bands < 1 ||
            eqpart.params.fft.n_bands > VCU_EQPRG_MAX_BANDS)
        {
            return false;
        }
        GET_TOKEN(COMMA);
        for (n = 0; n < eqpart.params.fft.n_bands; n++) {
            GET_TOKEN(REAL);
            if (yylval.real < VCU_EQPRG_MIN_BAND_FREQUENCY ||
                yylval.real > VCU_EQPRG_MAX_BAND_FREQUENCY)
            {
                return false;
            }
            eqpart.params.fft.band[n] = yylval.real;
            if (n > 0 &&
                yylval.real - VCU_FFTEQ_SMALLEST_BAND <
                eqpart.params.fft.band[n-1])
            {
                return false;
            }
            GET_TOKEN(SLASH);
            GET_TOKEN(REAL);
            eqpart.params.fft.amp[n] = yylval.real;
            if (n < eqpart.params.fft.n_bands - 1) {
                GET_TOKEN(COMMA);
            }
        }
        GET_TOKEN(RBRACE);
    } else {
	return false;
    }
    GET_TOKEN(RBRACE);

    if ((eqpart.type == VCU_EQPRG_TYPE_FFT_DYNAMIC &&
         cu.eq.eqprg[eq_index].part_count != 0) ||
        (cu.eq.eqprg[eq_index].part_count == 1 &&
         cu.eq.eqprg[eq_index].parts[0].type == VCU_EQPRG_TYPE_FFT_DYNAMIC))
    {
        return false;
    }
    
    cu.eq.eqprg[eq_index].part_count++;

    cu.eq.eqprg[eq_index].parts =
	realloc(cu.eq.eqprg[eq_index].parts,
		cu.eq.eqprg[eq_index].part_count *
		sizeof(struct vcu_eqpart));
    if (cu.eq.eqprg[eq_index].parts == NULL) {
	elog(LOGERROR, "realloc failed\n");
	exit(EXIT_FAILURE);
    }
    cu.eq.eqprg[eq_index].parts[cu.eq.eqprg[eq_index].part_count - 1] = eqpart;

    return true;
}

#define INTEGER 1000

static int
parse_array(int token,
            void *array_base,
            int array_len)
{
    uint8_t *ptr;
    int n, tok, subidx;

    ptr = (uint8_t *)array_base;
    if ((tok = yylex()) != LBRACE) {
        yy_perror_tok(LBRACE, tok);
        return -1;
    }    
    for (n = 0, subidx = 0, tok = COMMA; n < array_len && tok == COMMA; n++) {
        if ((tok = yylex()) == RBRACE) {
            break;
        }
        switch (token) {
        case INTEGER:
            if (tok != REAL) {
                yy_perror_tok(REAL, tok);
                return -1;
            }
            *(int *)ptr = (int)yylval.real;
            ptr += sizeof(int);
            break;
        case BOOLEAN:
            if (tok != BOOLEAN) {
                yy_perror_tok(BOOLEAN, tok);
                return -1;
            }
            *(bool_t *)ptr = yylval.boolean;
            ptr += sizeof(bool_t);
            break;
        case BOOLEAN_BIT:
            if (tok != BOOLEAN) {
                yy_perror_tok(BOOLEAN, tok);
                return -1;
            }
            if (yylval.boolean) {
                bit_set((uint32_t *)ptr, subidx);
            } else {
                bit_clr((uint32_t *)ptr, subidx);
            }
            subidx++;
            if (subidx == 32) {
                ptr += sizeof(uint32_t);
                subidx = 0;
            }
            break;
        case REAL:
            if (tok != REAL) {
                yy_perror_tok(REAL, tok);
                return -1;
            }
            *(double *)ptr = yylval.real;
            ptr += sizeof(double);
            break;
        }
        tok = yylex();
    }
    if (tok != RBRACE) {
        yy_perror_tok(RBRACE, tok);
        return -1;
    }
    return n;
}

static bool_t
parse_matrix(int token,
             void *matrix_base,
             int dim1,
             int dim2)
{
    uint8_t *ptr;
    int n, tok;

    ptr = (uint8_t *)matrix_base;
    GET_TOKEN(LBRACE);
    for (n = 0, tok = COMMA; n < dim1 && tok == COMMA; n++) {
        if (parse_array(token, ptr, dim2) == -1) {
            return false;
        }
        switch (token) {
        case INTEGER:
            ptr += dim2 * sizeof(int);
            break;
        case BOOLEAN:
            ptr += dim2 * sizeof(bool_t);
            break;
        case BOOLEAN_BIT:
            ptr += dim2 / 8;
            break;
        case REAL:
            ptr += dim2 * sizeof(double);
            break;
        }
        tok = yylex();
    }
    if (tok != RBRACE) {
        yy_perror_tok(RBRACE, tok);
        return false;
    }
    return true;
}

static void
print_array(FILE *stream,
            int token,
            void *array_base,
            int array_len)
{
    int n, subidx;
    uint8_t *ptr;

    if (array_len == 0) {
        fprintf(stream, "{ }");
        return;
    }
    ptr = (uint8_t *)array_base;
    fprintf(stream, "{ ");
    for (n = 0, subidx = 0; n < array_len - 1; n++) {
        switch (token) {
        case INTEGER:
            fprintf(stream, "%d, ", *(int *)ptr);
            ptr += sizeof(int);
            break;
        case BOOLEAN:
            fprintf(stream, "%s, ", (*(bool_t *)ptr) ? "true" : "false");
            ptr += sizeof(bool_t);
            break;
        case BOOLEAN_BIT:
            fprintf(stream, "%s, ", (bit_isset((uint32_t *)ptr, subidx)) ?
                    "true" : "false");
            subidx++;
            if (subidx == 32) {
                subidx = 0;
                ptr += sizeof(uint32_t);
            }
            break;
        case REAL:
            fprintf(stream, "%g, ", *(double *)ptr);
            ptr += sizeof(double);
            break;
        }
    }
    switch (token) {
    case INTEGER:
        fprintf(stream, "%d }", *(int *)ptr);
        break;
    case BOOLEAN:
        fprintf(stream, "%s }", (*(bool_t *)ptr) ? "true" : "false");
        break;
    case BOOLEAN_BIT:
        fprintf(stream, "%s }", (bit_isset((uint32_t *)ptr, subidx)) ?
                "true" : "false");
        break;
    case REAL:
        fprintf(stream, "%g }", *(double *)ptr);
        break;
    }
}

static void
ignore_to_eos(void)
{
    int tok;
    
    while ((tok = yylex() != EOS) && tok > 0);
}

static bool_t
parse_conf_user(FILE *stream)
{
    uint32_t isset[__LINE__ / 32 + 100];
    int tok, n, i, logi, phys, type, io;
    char module_name[PATH_MAX];
    struct fprogref fref;
    
    /* FIXME: too indented, too much copied code */

    elog(LOGDEBUG, "enter");

    memset(isset, 0, sizeof(isset));
    
    init_conf_user();

    while ((tok = yylex()) > 0) {
	if (tok != STRING) {
	    yy_perror_tok(STRING, tok);
	    return false;
	}
	if (strcmp(yylval.string, "volume") == 0) {
	    GET_TOKEN(LBRACE);
	    while ((tok = yylex()) != RBRACE) {
		if (tok != FIELD) {
		    yy_perror_tok(FIELD, tok);
		    return false;
		}
		if (strcmp(yylval.field, "master") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(REAL);
		    cu.volume.master = yylval.real;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "group") == 0) {
		    TEST_AND_SET(__LINE__);
                    if (parse_array(REAL, cu.volume.group,
                                    VCU_MAX_CHANNEL_GROUPS) == -1)
                    {
                        return false;
                    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "mute") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
                    if (!parse_matrix(BOOLEAN, cu.volume.mute[IN],
                                      VCU_CH_WP_COUNT,
                                      BFIR_MAX_OUTPUT_CHANNELS))
                    {
                        return false;
                    }
                    GET_TOKEN(COMMA);
                    if (!parse_matrix(BOOLEAN, cu.volume.mute[OUT],
                                      VCU_CH_WP_COUNT,
                                      BFIR_MAX_OUTPUT_CHANNELS))
                    {
                        return false;
                    }
                    GET_TOKEN(RBRACE);
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "group_mute") == 0) {
		    TEST_AND_SET(__LINE__);
                    if (parse_array(BOOLEAN, cu.volume.group_mute,
                                    VCU_MAX_CHANNEL_GROUPS) == -1)
                    {
                        return false;
                    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "master_temp_mute") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
		    cu.volume.master_temp_mute = yylval.boolean;
		    GET_TOKEN(EOS);
		} else {
		    elog(LOGDEBUG, "Ignoring unknown volume section "
                         "field \"%s\".", yylval.string);
                    ignore_to_eos();
		}
	    }
	    GET_TOKEN(EOS);
	} else if (strcmp(yylval.string, "signal") == 0) {
	    GET_TOKEN(LBRACE);
	    while ((tok = yylex()) != RBRACE) {
		if (tok != FIELD) {
		    yy_perror_tok(FIELD, tok);
		    return false;
		}
		if (strcmp(yylval.field, "rates") == 0) {
		    TEST_AND_SET(__LINE__);
                    cu.signal.n_rates = parse_array(INTEGER, cu.signal.rates,
                                                    VCU_VALID_RATES_COUNT);
                    if (cu.signal.n_rates == -1) {
                        return false;
                    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "suggested_rate") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_INTEGER();
		    cu.signal.suggested_rate = (int)yylval.real;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "output_format") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(STRING);
		    for (n = BF_SAMPLE_FORMAT_MIN;
			 n <= BF_SAMPLE_FORMAT_MAX;
			 n++)
		    {
			if (strcasecmp(bf_strsampleformat(n),
				       yylval.string) == 0)
			{
			    cu.signal.output_sample_format = n;
			    break;
			}
		    }
		    if (n == BF_SAMPLE_FORMAT_MAX + 1) {
			yy_perror("unknown sample format: \"%s\"",
				  yylval.string);
			return false;
		    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "soft_clip") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
		    cu.signal.use_soft_clip = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "dither") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
		    cu.signal.apply_dither = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "watch_clock") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
		    cu.signal.watch_clock = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "double_precision") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
		    cu.signal.double_precision = yylval.boolean;
		    GET_TOKEN(EOS);
		} else {
		    elog(LOGDEBUG, "Ignoring unknown signal section "
                         "field \"%s\".", yylval.string);
                    ignore_to_eos();
		}
	    }
	    GET_TOKEN(EOS);
	} else if (strcmp(yylval.string, "channel") == 0) {
	    GET_TOKEN(LBRACE);
	    while ((tok = yylex()) != RBRACE) {
		if (tok != FIELD) {
		    yy_perror_tok(FIELD, tok);
		    return false;
		}
		if (strcmp(yylval.field, "count") == 0) {
		    TEST_AND_SET(__LINE__);
                    if (parse_array(INTEGER, cu.channel.count[OUT],
                                    VCU_CH_COUNT) == -1)
                    {
                        return false;
                    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "mapping") == 0) {
                    memset(cu.channel.map, 0, sizeof(cu.channel.map));
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
		    for (n = 0; ; n++) {
                        tok = yylex();
                        if (tok == RBRACE || (n == 0 && tok != LBRACE) ||
                            (n > 0 && tok != COMMA))
                        {
                            break;
                        }
                        if (n > 0) {
                            GET_TOKEN(LBRACE);
                        }
                        GET_TOKEN(STRING);
                        if (strlen(yylval.string) < 3 ||
                            (yylval.string[0] != 'i' &&
                             yylval.string[0] != 'o') ||
                            (yylval.string[1] != 'A' &&
                             yylval.string[1] != 'R' &&
                             yylval.string[1] != 'B' &&
                             yylval.string[1] != 'P') ||
                            (logi = atoi(&yylval.string[2])) < 0 ||
                            logi > VCU_MAX_CHANNELS)
                        {
                            yy_perror("channel name \"%s\"", yylval.string);
                            return false;
                        }
                        io = yylval.string[0] == 'i' ? IN : OUT;
                        type = 0;
                        switch (yylval.string[1]) {
                        case 'A': type = VCU_CH_AMBIOPOLE; break;
                        case 'R': type = VCU_CH_REVERB; break;
                        case 'B': type = VCU_CH_AMBISONIC; break;
                        case 'P': type = VCU_CH_PASSTHRU; break;
                        }
                        GET_TOKEN(COMMA);
                        GET_TOKEN(LBRACE);
                        tok = COMMA;
                        for (i = 0; tok == COMMA && i < VCU_MAX_CHANNELS; i++) {
                            GET_TOKEN(STRING);
                            if (strlen(yylval.string) < 2 ||
                                (yylval.string[0] != 'i' &&
                                 yylval.string[0] != 'o'))
                            {
                                yy_perror("channel name \"%s\"", yylval.string);
                                return false;
                            }
                            if ((yylval.string[0] == 'i' && io == OUT) ||
                                (yylval.string[0] == 'o' && io == IN) ||
                                (phys = atoi(&yylval.string[1])) < 0 ||
                                phys >= VCU_MAX_CHANNELS)
                            {
                                yy_perror("I/O mismatch");
                                return false;
                            }
                            bit_set(cu.channel.map[io][type][logi], phys);
                            tok = yylex();
                        }
                        if (tok != RBRACE) {
                            yy_perror_tok(RBRACE, tok);
                            return false;
                        }
                        GET_TOKEN(RBRACE);
                    }
		    if (tok != RBRACE) {
			yy_perror_tok(RBRACE, tok);
			return false;
		    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "speaker_positions") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
		    for (n = 0; n < VCU_MAX_CHANNELS; n++) {
                        tok = yylex();
                        if (tok == RBRACE || (n == 0 && tok != LBRACE) ||
                            (n > 0 && tok != COMMA))
                        {
                            break;
                        }
                        if (n > 0) {
                            GET_TOKEN(LBRACE);
                        }
			GET_TOKEN(BOOLEAN);
			cu.channel.speakerpos[n].active = yylval.boolean;
			GET_TOKEN(COMMA);
			GET_TOKEN(REAL);
			cu.channel.speakerpos[n].azimut = yylval.real;
			GET_TOKEN(COMMA);
			GET_TOKEN(REAL);
			cu.channel.speakerpos[n].elevation = yylval.real;
			GET_TOKEN(COMMA);
			GET_TOKEN(REAL);
			cu.channel.speakerpos[n].distance = yylval.real;
			GET_TOKEN(RBRACE);
		    }
		    if (tok != RBRACE) {
			yy_perror_tok(RBRACE, tok);
			return false;
		    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "stages") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
		    for (n = 0; n < VCU_MAX_STAGES; n++) {
                        tok = yylex();
                        if (tok == RBRACE || (n == 0 && tok != LBRACE) ||
                            (n > 0 && tok != COMMA))
                        {
                            break;
                        }
                        if (n > 0) {
                            GET_TOKEN(LBRACE);
                        }
			GET_TOKEN(REAL);
			cu.channel.stage[n].azimut = yylval.real;
			GET_TOKEN(COMMA);
			GET_TOKEN(REAL);
			cu.channel.stage[n].elevation = yylval.real;
			GET_TOKEN(COMMA);
			GET_TOKEN(REAL);
			cu.channel.stage[n].width = yylval.real;
			GET_TOKEN(COMMA);
			GET_TOKEN(REAL);
			cu.channel.stage[n].height = yylval.real;
			GET_TOKEN(RBRACE);
		    }
                    cu.channel.stage_count = n;
		    if (tok != RBRACE) {
			yy_perror_tok(RBRACE, tok);
			return false;
		    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "distance_compensate") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(BOOLEAN);
                    cu.channel.distance_compensate = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "trim_volumes") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
                    if (!parse_matrix(REAL, cu.channel.trim.volume[IN],
                                      VCU_CH_WP_COUNT,
                                      BFIR_MAX_OUTPUT_CHANNELS))
                    {
                        return false;
                    }
                    GET_TOKEN(COMMA);
                    if (!parse_matrix(REAL, cu.channel.trim.volume[OUT],
                                      VCU_CH_WP_COUNT,
                                      BFIR_MAX_OUTPUT_CHANNELS))
                    {
                        return false;
                    }
                    GET_TOKEN(RBRACE);
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "trim_delays") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
                    if (!parse_matrix(REAL, cu.channel.trim.delay[IN],
                                      VCU_CH_WP_COUNT,
                                      BFIR_MAX_OUTPUT_CHANNELS))
                    {
                        return false;
                    }
                    GET_TOKEN(COMMA);
                    if (!parse_matrix(REAL, cu.channel.trim.delay[OUT],
                                      VCU_CH_WP_COUNT,
                                      BFIR_MAX_OUTPUT_CHANNELS))
                    {
                        return false;
                    }
                    GET_TOKEN(RBRACE);
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "groups") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
                    for (n = 0; n < VCU_MAX_CHANNEL_GROUPS; n++) {
                        tok = yylex();
                        if (tok == RBRACE || (n == 0 && tok != LBRACE) ||
                            (n > 0 && tok != COMMA))
                        {
                            break;
                        }
                        if (n > 0) {
                            GET_TOKEN(LBRACE);
                        }
                        GET_TOKEN(STRING);
                        strncpy(cu.channel.group[n].name, yylval.string,
                                VCU_CHGRP_NAME_SIZE);
                        cu.channel.group[n].name[VCU_CHGRP_NAME_SIZE-1] = '\0';
                        FOR_IN_AND_OUT {
                            GET_TOKEN(COMMA);
                            if (!parse_matrix(BOOLEAN_BIT,
                                              (uint8_t *)cu.channel.group[n].
                                              channels[io],
                                              VCU_CH_COUNT,
                                              VCU_MAX_CHANNELS))
                            {
                                return false;
                            }
                        }
                        GET_TOKEN(RBRACE);
                    }
                    cu.channel.group_count = n;
                    if (tok != RBRACE) {
                        yy_perror_tok(RBRACE, tok);
                        return false;
                    }
                    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "modes") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
                    for (n = 0; n < VCU_MAX_CHANNEL_MODES; n++) {
                        tok = yylex();
                        if (tok == RBRACE || (n == 0 && tok != LBRACE) ||
                            (n > 0 && tok != COMMA))
                        {
                            break;
                        }
                        if (n > 0) {
                            GET_TOKEN(LBRACE);
                        }
                        GET_TOKEN(STRING);
                        strncpy(cu.channel.mode[n].name, yylval.string,
                                VCU_CHMODE_NAME_SIZE);
                        cu.channel.mode[n].name[VCU_CHMODE_NAME_SIZE-1] = '\0';
                        GET_TOKEN(COMMA);
                        if (parse_array(BOOLEAN_BIT, cu.channel.mode[n].groups,
                                        VCU_MAX_CHANNEL_GROUPS) == -1)
                        {
                            return false;
                        }
                        GET_TOKEN(COMMA);
                        if (parse_array(BOOLEAN_BIT, cu.channel.mode[n].eqmodes,
                                        VCU_MAX_EQ_MODES) == -1)
                        {
                            return false;
                        }
                        GET_TOKEN(COMMA);
                        GET_INTEGER();
                        cu.channel.mode[n].active_eq_mode = (int)yylval.real;
                        GET_TOKEN(RBRACE);
                    }
                    cu.channel.mode_count = n;
                    if (tok != RBRACE) {
                        yy_perror_tok(RBRACE, tok);
                        return false;
                    }
                    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "active_mode") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_INTEGER();
		    cu.channel.active_mode = (int)yylval.real;
		    GET_TOKEN(EOS);
		} else {
		    elog(LOGDEBUG, "Ignoring unknown channel section "
                         "field \"%s\".", yylval.string);
                    ignore_to_eos();
		}
	    }
	    GET_TOKEN(EOS);
	} else if (strcmp(yylval.string, "equalisation") == 0) {
	    GET_TOKEN(LBRACE);
	    while ((tok = yylex()) != RBRACE) {
		if (tok != FIELD) {
		    yy_perror_tok(FIELD, tok);
		    return false;
		}
		if (strcmp(yylval.field, "modes") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
                    for (n = 0; n < VCU_MAX_EQ_MODES; n++) {
                        tok = yylex();
                        if (tok == RBRACE || (n == 0 && tok != LBRACE) ||
                            (n > 0 && tok != COMMA))
                        {
                            break;
                        }
                        if (n > 0) {
                            GET_TOKEN(LBRACE);
                        }
                        GET_TOKEN(STRING);
                        strncpy(cu.eq.mode[n].name, yylval.string,
                                VCU_EQMODE_NAME_SIZE);
                        cu.eq.mode[n].name[VCU_EQMODE_NAME_SIZE-1] = '\0';
                        memset(cu.eq.mode[n].prgs, -1,
                               sizeof(cu.eq.mode[n].prgs));
                        FOR_IN_AND_OUT {
                            GET_TOKEN(COMMA);
                            if (!parse_matrix(INTEGER,
                                              (uint8_t *)cu.eq.mode[n].prgs[io],
                                              VCU_CH_WP_COUNT,
                                              VCU_MAX_CHANNELS))
                            {
                                return false;
                            }
                        }
                        GET_TOKEN(RBRACE);
                    }
                    cu.eq.mode_count = n;
                    if (tok != RBRACE) {
                        yy_perror_tok(RBRACE, tok);
                        return false;
                    }
                    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "active_mode") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_INTEGER();
		    cu.eq.active_mode = (int)yylval.real;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "equalisers") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
                    for (n = 0; n < VCU_MAX_EQ_PROGRAMS; n++) {
                        tok = yylex();
                        if (tok == RBRACE || (n == 0 && tok != LBRACE) ||
                            (n > 0 && tok != COMMA))
                        {
                            break;
                        }
                        if (n > 0) {
                            GET_TOKEN(LBRACE);
                        }
                        GET_TOKEN(STRING);
                        strncpy(cu.eq.eqprg[n].name, yylval.string,
                                VCU_EQPRG_NAME_SIZE);
                        cu.eq.eqprg[n].name[VCU_EQPRG_NAME_SIZE-1] = '\0';
                        GET_TOKEN(COMMA);
			do {
			    if (!get_eqpart(n)) {
				yy_perror("invalid equaliser");
				return false;
			    }
			} while ((tok = yylex()) == COMMA);
			if (tok != RBRACE) {
			    yy_perror_tok(RBRACE, tok);
			    return false;
			}
		    }
                    cu.eq.eqprg_count = n;
		    if (tok != RBRACE) {
			yy_perror_tok(RBRACE, tok);
			return false;
		    }
                    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "programs") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
		    for (n = 0, tok = COMMA;
			 n < CONF_MAX_EQ_PROGRAMS && tok == COMMA;
			 n++)
		    {
			GET_TOKEN(STRING);
			cu.eq.loaded_programs[n] =
			    filterprg_prgid_aton(yylval.string);
			tok = yylex();
		    }
		    if (tok != RBRACE) {
			yy_perror_tok(RBRACE, tok);
			return false;
		    }
                    GET_TOKEN(EOS);
		    cu.eq.n_loaded_programs = n;
		} else {
		    elog(LOGDEBUG, "Ignoring unknown equalisation section "
                         "field \"%s\".", yylval.string);
                    ignore_to_eos();
		}
	    }
	    GET_TOKEN(EOS);
	} else if (strcmp(yylval.string, "ambiopole") == 0) {
	    GET_TOKEN(LBRACE);
	    while ((tok = yylex()) != RBRACE) {
		if (tok != FIELD) {
		    yy_perror_tok(FIELD, tok);
		    return false;
		}
		if (strcmp(yylval.field, "xtalk_cancel") == 0) {
		    TEST_AND_SET(__LINE__);
                    if (parse_array(BOOLEAN, cu.ambiopole.use_xtalk_cancel,
                                    BFIR_MAX_OUTPUT_CHANNELS) == -1)
                    {
                        return false;
                    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "current_program") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
		    for (n = 0, tok = COMMA;
			 n < BFIR_MAX_OUTPUT_CHANNELS/2 && tok == COMMA;
			 n++)
		    {
                        GET_TOKEN(STRING);
                        cu.ambiopole.current[n] =
                            filterprg_prgid_aton(yylval.string);
			tok = yylex();
		    }
		    if (tok != RBRACE) {
			yy_perror_tok(RBRACE, tok);
			return false;
		    }
                    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "programs") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
		    for (n = 0, tok = COMMA;
			 n < CONF_MAX_XTC_PROGRAMS && tok == COMMA;
			 n++)
		    {
			GET_TOKEN(STRING);
			cu.ambiopole.loaded_programs[n] =
			    filterprg_prgid_aton(yylval.string);
			tok = yylex();
		    }
		    if (tok != RBRACE) {
			yy_perror_tok(RBRACE, tok);
			return false;
		    }
                    GET_TOKEN(EOS);
		    cu.ambiopole.n_loaded_programs = n;
		} else {
		    elog(LOGDEBUG, "Ignoring unknown ambiopole section "
                         "field \"%s\".", yylval.string);
                    ignore_to_eos();                                    
		}
	    }
	    GET_TOKEN(EOS);
	} else if (strcmp(yylval.string, "reverb") == 0) {
	    GET_TOKEN(LBRACE);
	    while ((tok = yylex()) != RBRACE) {
		if (tok != FIELD) {
		    yy_perror_tok(FIELD, tok);
		    return false;
		}
		if (strcmp(yylval.field, "block_sizes") == 0) {
		    TEST_AND_SET(__LINE__);
		    for (n = 0, tok = COMMA;
			 n < VCU_VALID_RATES_COUNT && tok == COMMA; n++)
		    {
			GET_TOKEN(LBRACE);
			GET_INTEGER();
			cu.reverb.block_sizes[n] = (int)yylval.real;
			GET_TOKEN(COMMA);
			GET_INTEGER();
			cu.reverb.n_blocks[n] = (int)yylval.real;
			GET_TOKEN(RBRACE);
			tok = yylex();
		    }
		    if (tok != EOS) {
			yy_perror_tok(EOS, tok);
			return false;
		    }
		} else if (strcmp(yylval.field, "benchmark_cpu_count") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_INTEGER();
		    cu.reverb.benchmark_sysinfo.cpu_count =
			(int)yylval.real;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "benchmark_cpu_speed") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_INTEGER();
		    cu.reverb.benchmark_sysinfo.cpu_speed =
			(int)yylval.real;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "benchmark_cpu_model") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(STRING);
		    strncpy(cu.reverb.benchmark_sysinfo.cpu_modelname,
			    yylval.string,
			    sizeof(cu.reverb.benchmark_sysinfo.
				   cpu_modelname));
		    cu.reverb.benchmark_sysinfo.cpu_modelname
			[sizeof(cu.reverb.benchmark_sysinfo.
				cpu_modelname) - 1] = '\0';
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "benchmark_memsize") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_INTEGER();
		    cu.reverb.benchmark_sysinfo.mem_total =
			(int)yylval.real;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "max_benchmark_load") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(REAL);
		    cu.reverb.max_benchmark_load = yylval.real;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "max_benchmark_length") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(REAL);
		    cu.reverb.max_benchmark_length = yylval.real;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "iodelay_manually_set") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
		    cu.reverb.iodelay_manually_set = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "current_program") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(STRING);
		    cu.reverb.current =
			filterprg_prgid_aton(yylval.string);
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "programs") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
		    for (n = 0, tok = COMMA;
			 n < BFIR_MAX_REVERB_PROGRAMS && tok == COMMA;
			 n++)
		    {
			GET_TOKEN(STRING);
			cu.reverb.loaded_programs[n] =
			    filterprg_prgid_aton(yylval.string);
			tok = yylex();
		    }
		    if (tok != RBRACE) {
			yy_perror_tok(RBRACE, tok);
			return false;
		    }
                    GET_TOKEN(EOS);
		    cu.reverb.n_loaded_programs = n;
		} else if (strcmp(yylval.field, "response_mode") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(STRING);
                    if (strcmp(yylval.string, "full") == 0) {
                        cu.reverb.response_mode = VCU_REVERB_RESP_MODE_FULL;
                    } else if (strcmp(yylval.string, "mute_direct") == 0) {
                        cu.reverb.response_mode =
                            VCU_REVERB_RESP_MODE_MUTE_DIRECT;
                    } else if (strcmp(yylval.string, "mute_early") == 0) {
                        cu.reverb.response_mode =
                            VCU_REVERB_RESP_MODE_MUTE_EARLY;
                    } else if (strcmp(yylval.string, "stage_adapted") == 0) {
                        cu.reverb.response_mode =
                            VCU_REVERB_RESP_MODE_STAGE_ADAPTED;
                    } else {
                        yy_perror("invalid response mode \"%s\"",
                                  yylval.string);
                        return false;
                    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "decorrelate_mode") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(STRING);
                    if (strcmp(yylval.string, "none") == 0) {
                        cu.reverb.decorrelate_mode =
                            VCU_REVERB_DECORR_MODE_NONE;
                    } else if (strcmp(yylval.string, "needed") == 0) {
                        cu.reverb.decorrelate_mode =
                            VCU_REVERB_DECORR_MODE_NEEDED;
                    } else if (strcmp(yylval.string, "all") == 0) {
                        cu.reverb.decorrelate_mode =
                            VCU_REVERB_DECORR_MODE_ALL;
                    } else {
                        yy_perror("invalid decorrelate mode \"%s\"",
                                  yylval.string);
                        return false;
                    }
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "force_delay_diff") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(BOOLEAN);
                    cu.reverb.force_delay_diff = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "force_decay") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(BOOLEAN);
                    cu.reverb.force_decay = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "mix_with_dry") == 0) {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(BOOLEAN);
                    cu.reverb.mix_with_dry = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "dry_volume") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(REAL);
		    cu.reverb.dry_volume = yylval.real;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "bformat_directivity") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(REAL);
		    cu.reverb.bformat_directivity = yylval.real;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "source_layout") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(STRING);
                    if (strcmp(yylval.string, "center") == 0) {
                        cu.reverb.source_layout =
                            VCU_REVERB_SOURCE_LAYOUT_CENTER;
                    } else if (strcmp(yylval.string, "stereo") == 0) {
                        cu.reverb.source_layout =
                            VCU_REVERB_SOURCE_LAYOUT_STEREO;
                    } else if (strcmp(yylval.string, "lrc") == 0) {
                        cu.reverb.source_layout = VCU_REVERB_SOURCE_LAYOUT_LRC;
                    } else {
                        yy_perror("invalid reverb input format \"%s\"",
                                  yylval.string);
                        return false;
                    }
		    GET_TOKEN(EOS);
		} else {
		    elog(LOGDEBUG, "Ignoring unknown reverb section field "
                         "\"%s\".", yylval.string);
                    ignore_to_eos();
		}
	    }
	    GET_TOKEN(EOS);
	} else if (strcmp(yylval.string, "ambisonics") == 0) {
	    GET_TOKEN(LBRACE);
	    while ((tok = yylex()) != RBRACE) {
		if (tok != FIELD) {
		    yy_perror_tok(FIELD, tok);
		    return false;
		}
		if (strcmp(yylval.field, "input_format") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(STRING);
                    if (strcmp(yylval.string, "first order") == 0) {
                        cu.ambisonics.input_format =
                            VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER;
                    } else if (strcmp(yylval.string, "second order") == 0) {
                        cu.ambisonics.input_format =
                            VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER;
                    } else if (strcmp(yylval.string, "uhj") == 0) {
                        cu.ambisonics.input_format =
                            VCU_AMBISONIC_INPUT_FORMAT_UHJ;
                    } else {
                        yy_perror("invalid Ambisonics input format \"%s\"",
                                  yylval.string);
                        return false;
                    }
		    GET_TOKEN(EOS);
		} else {
		    elog(LOGDEBUG, "Ignoring unknown ambisonics section "
                         "field \"%s\".", yylval.string);
                    ignore_to_eos();
		}
            }
	    GET_TOKEN(EOS);
	} else if (strcmp(yylval.string, "ctl") == 0) {
	    GET_TOKEN(LBRACE);
	    while ((tok = yylex()) != RBRACE) {
		if (tok != FIELD) {
		    yy_perror_tok(FIELD, tok);
		    return false;
		}
		if (strcmp(yylval.field, "powersave") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
                    cu.ctl.powersave = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "debug") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
                    cu.ctl.debug = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "load_user_config") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
                    cu.ctl.load_user_config = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "load_revprg_config") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
                    cu.ctl.load_revprg_config = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field, "load_static_config") == 0) {
		    TEST_AND_SET(__LINE__);
		    GET_TOKEN(BOOLEAN);
                    cu.ctl.load_static_config = yylval.boolean;
		    GET_TOKEN(EOS);
		} else if (strcmp(yylval.field,
                                  "filter_program_references") == 0)
                {
		    TEST_AND_SET(__LINE__);
                    GET_TOKEN(LBRACE);
		    for (n = 0, tok = COMMA;
			 n < CONF_MAX_FPROGREFS && tok == COMMA;
			 n++)
		    {
                        memset(&fref, 0, sizeof(fref));
                        GET_TOKEN(LBRACE);
			GET_TOKEN(STRING);
			fref.id = filterprg_prgid_aton(yylval.string);
                        GET_TOKEN(COMMA);
			GET_TOKEN(STRING);
                        strncpy(fref.path, yylval.string, PATH_MAX - 1);
                        GET_TOKEN(RBRACE);
			cu.ctl.fprogrefs[n] = malloc(sizeof(struct fprogref));
                        memcpy(cu.ctl.fprogrefs[n], &fref,
                               sizeof(struct fprogref));
			tok = yylex();
		    }
		    if (tok != RBRACE) {
                        while (n--) {
                            free(cu.ctl.fprogrefs[n]);
                        }
			yy_perror_tok(RBRACE, tok);
			return false;
		    }
                    GET_TOKEN(EOS);
		    cu.ctl.n_fprogrefs = n;
		} else {
		    elog(LOGDEBUG, "Ignoring unknown ctl section "
                         "field \"%s\".", yylval.string);
                    ignore_to_eos();
		}
            }
	    GET_TOKEN(EOS);
	} else {
	    strncpy(module_name, yylval.string, PATH_MAX);
	    module_name[PATH_MAX - 1] = '\0';
	    /* see if it matches a custom section */
	    for (n = 0; n < conf_static.n_uimodules; n++) {
		if (conf_static.uimodules[n]->loaduserconf != NULL &&
		    strcmp(module_name, conf_static.uimodules[n]->name()) == 0)
		{
		    GET_TOKEN(LBRACE);
		    if (!conf_static.uimodules[n]->
			loaduserconf(get_config_token))
		    {
			elog(LOGDEBUG, "Ignoring invalid user config for "
                             "module \"%s\".", module_name);
		    }
		    GET_TOKEN(EOS);
		    break;
		}
	    }
	    if (n == conf_static.n_uimodules) {
		elog(LOGDEBUG, "Ignoring section \"%s\" in user configuration.",
		     module_name);
	    }
	}
    }
    if (tok == INVALID_INPUT) {
        yy_perror("syntax error");
        exit(EXIT_FAILURE);
    }
    stabilise_conf_user();
    n = 0;
    for (type = 0; type < VCU_CH_COUNT; type++) {
        n += cu.channel.count[OUT][type];        
    }
    if (n > BFIR_MAX_OUTPUT_CHANNELS) {
        init_conf_user();
    }
    elog(LOGDEBUG, "normal exit");
    return true;
}

static void *
load_modfun(void *handle,
	    const char modname[],
	    const char name[],
	    bool_t required)
{
    char *funname;
    void *fun;

    funname = alloca(strlen(name) + 1);
    strcpy(funname, name);
    if ((fun = dlsym(handle, funname)) == NULL && required) {
	yy_perror("failed to resolve function \"%s\" in module \"%s\": %s",
		  name, modname, dlerror());
	exit(EXIT_FAILURE);
    }
    return fun;
}

static bool_t
load_module(struct ui_module_functions **uim,
	    int *tok)
{
    char *fullname;
    void *handle;
    
    GET_TOKEN(STRING);

    fullname = alloca(strlen(yylval.string) + PATH_MAX);
    if (conf_static.uimodules_path[strlen(conf_static.uimodules_path) - 1]
        == '/')
    {        
        sprintf(fullname, "%s%s.vcu", conf_static.uimodules_path,
                yylval.string);
    } else {
        sprintf(fullname, "%s/%s.vcu", conf_static.uimodules_path,
                yylval.string);
    }
    
    if ((handle = dlopen(fullname, RTLD_NOW | RTLD_GLOBAL)) == NULL) {
	yy_perror("failed to open module \"%s\": %s", yylval.string, dlerror());
	exit(EXIT_FAILURE);
    }
    (*uim) = malloc(sizeof(struct ui_module_functions));
    memset(*uim, 0, sizeof(struct ui_module_functions));
    (*uim)->name = load_modfun(handle, yylval.string, VCU_FUN_NAME, true);
    (*uim)->preinit = load_modfun(handle, yylval.string, VCU_FUN_PREINIT,
				  false);
    (*uim)->init = load_modfun(handle, yylval.string, VCU_FUN_INIT, false);
    (*uim)->finish = load_modfun(handle, yylval.string, VCU_FUN_FINISH, false);
    (*uim)->needterm = load_modfun(handle, yylval.string, VCU_FUN_NEEDTERM,
                                 false);
    (*uim)->share = load_modfun(handle, yylval.string, VCU_FUN_SHARE, false);
    (*uim)->select_wrapper = load_modfun(handle, yylval.string,
                                         VCU_FUN_SELECT_WRAPPER, false);
    (*uim)->loaduserconf = load_modfun(handle, yylval.string,
				       VCU_FUN_LOADUSERCONF, false);
    (*uim)->saveuserconf = load_modfun(handle, yylval.string,
				       VCU_FUN_SAVEUSERCONF, false);
    (*uim)->progress.init_phase =
        load_modfun(handle, yylval.string, VCU_FUN_PROGRESS_INIT_PHASE, false);
    (*uim)->progress.benchmark =
        load_modfun(handle, yylval.string, VCU_FUN_PROGRESS_BENCHMARK, false);
    (*uim)->progress.load_program =
        load_modfun(handle, yylval.string, VCU_FUN_PROGRESS_LOAD_PROGRAM,
                    false);
    (*uim)->progress.generate_eq =
        load_modfun(handle, yylval.string, VCU_FUN_PROGRESS_GENERATE_EQ, false);
    (*uim)->progress.generate_ambisonic =
        load_modfun(handle, yylval.string, VCU_FUN_PROGRESS_GENERATE_AMBISONIC,
                    false);
    
    if ((*tok = yylex()) == LBRACE) {
	if ((*uim)->preinit == NULL) {
	    yy_perror("private configuration for non-configurable module");
	    exit(EXIT_FAILURE);
	}
	/* let the module parse its private configuration */
	if (!(*uim)->preinit(get_config_token)) {
	    yy_perror("private module failed to parse its configuration");
	    exit(EXIT_FAILURE);
	}
	*tok = yylex();
    } else if ((*uim)->preinit != NULL) {
	if (!(*uim)->preinit(dummy_get_config_token)) {
	    yy_perror("private module failed to parse its configuration");
	    exit(EXIT_FAILURE);
	}
    }

    return true;
}

static bool_t
parse_device(char device[],
	     char device_params[])
{
    int tok;
    
    GET_TOKEN(STRING);
    strncpy(device, yylval.string, PATH_MAX);
    device[PATH_MAX - 1] = '\0';
    if ((tok = yylex()) == EOS) {
        device_params[0] = '\0';
	return true;
    } else if (tok != SLASH) {
	yy_perror_tok(EOS, tok);
	return false;
    }    
    GET_TOKEN(STRING);
    strncpy(device_params, yylval.string, PATH_MAX);
    device_params[PATH_MAX - 1] = '\0';
    GET_TOKEN(EOS);
    return true;
}

static bool_t
parse_channels(int *open_channels,
	       int channel_index[])
{
    int tok, n;
    
    GET_INTEGER();
    if ((*open_channels = (int)yylval.real) < 0 ||
	*open_channels > VCU_MAX_CHANNELS || *open_channels < 1)
    {
	yy_perror("channel amount %d is out of range, valid range is 1 - %d",
                  *open_channels, VCU_MAX_CHANNELS);
	return false;
    }
    if ((tok = yylex()) == EOS) {
	for (n = 0; n < *open_channels; n++) {
	    channel_index[n] = n;
	}
	return true;
    } else if (tok != SLASH) {
	yy_perror_tok(EOS, tok);
	return false;
    }
    GET_TOKEN(LBRACE);
    for (n = 0; n < *open_channels; n++) {
	GET_INTEGER();
	if ((channel_index[n] = (int)yylval.real) < 0 ||
	    channel_index[n] > *open_channels)
	{
	    yy_perror("channel index is out of range");
	    return false;
	}
	if (n != *open_channels - 1) {
	    GET_TOKEN(COMMA);
	} else {
	    GET_TOKEN(RBRACE);
	}
    }    
    GET_TOKEN(EOS);
    return true;
}

static int
sampleformat2bits(int format)
{
    switch (format) {
    case BF_SAMPLE_FORMAT_S8:
	return 8;
    case BF_SAMPLE_FORMAT_S16_LE:
    case BF_SAMPLE_FORMAT_S16_BE:
    case BF_SAMPLE_FORMAT_S16_4LE:
    case BF_SAMPLE_FORMAT_S16_4BE:
	return 16;
    case BF_SAMPLE_FORMAT_S24_LE:
    case BF_SAMPLE_FORMAT_S24_BE:
    case BF_SAMPLE_FORMAT_S24_4LE:
    case BF_SAMPLE_FORMAT_S24_4BE:
	return 24;
    case BF_SAMPLE_FORMAT_S32_LE:
    case BF_SAMPLE_FORMAT_S32_BE:
    case BF_SAMPLE_FORMAT_FLOAT_LE:
    case BF_SAMPLE_FORMAT_FLOAT_BE:
	return 32;
    case BF_SAMPLE_FORMAT_FLOAT64_LE:
    case BF_SAMPLE_FORMAT_FLOAT64_BE:
    default:	    
	return 64;
    }
}

static bool_t
parse_formats(int *n_sample_formats,
	      int sample_formats[])
{
    int n, i, tok;

    GET_TOKEN(LBRACE);
    do {
	GET_TOKEN(STRING);
	for (n = BF_SAMPLE_FORMAT_MIN; n <= BF_SAMPLE_FORMAT_MAX; n++) {
	    if (strcasecmp(bf_strsampleformat(n), yylval.string) == 0) {
		sample_formats[*n_sample_formats] = n;
		(*n_sample_formats)++;
		break;
	    }
	}
	if (n == BF_SAMPLE_FORMAT_MAX + 1) {
	    return false;
	}
	tok = yylex();
    } while (*n_sample_formats != CONF_SAMPLE_FORMATS_MAX && tok == COMMA);
    if (tok != RBRACE) {
	yy_perror_tok(RBRACE, tok);
    }
    GET_TOKEN(EOS);
    for (n = 0; n < *n_sample_formats; n++) {
	for (i = 0; i < *n_sample_formats; i++) {
	    if (i != n && sample_formats[n] == sample_formats[i]) {
		yy_perror("duplicate sample formats");
		return false;
	    }
	    if (i != n &&
		sampleformat2bits(sample_formats[n]) ==
		sampleformat2bits(sample_formats[i]))
	    {
		yy_perror("sample formats of same resolution (bits), "
			  "pick only one of them");
		return false;
	    }
	}
    }
    return true;
}

#warning all static config errors should get easy-to-understand error messages

#define MISSING_FIELD(fieldname, idx)                                          \
    if (!bit_isset(isset, idx)) {                                              \
        elog(LOGERROR, "\"" fieldname "\" must be set in the static "          \
             "configuration file.");                                           \
        missing_field = true;                                                  \
    }

/* we have to return bool because of the macros, although we rather would exit
   directly on parse error */
static bool_t
parse_static(void)
{
    int rates[VCU_VALID_RATES_COUNT] = VCU_VALID_RATES;
    uint32_t isset[1] = { 0 };
    bool_t missing_field;
    int tok, n, i;
    
    while ((tok = yylex()) > 0) {
	switch (tok) {
	case FIELD:
	    if (strcmp(yylval.field, "allocate_ram") == 0) {
		TEST_AND_SET(0);
		GET_INTEGER();
		conf_static.allocate_ram = (int)yylval.real;
		if (conf_static.allocate_ram < CONF_ALLOCATE_RAM_MIN) {
		    yy_perror("too small \"allocate_ram\" value, need "
			      "at least %d megabytes", CONF_ALLOCATE_RAM_MIN);
		    exit(EXIT_FAILURE);
		}
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "rates") == 0) {
		TEST_AND_SET(1);
                GET_TOKEN(LBRACE);
		for (n = 0, tok = COMMA;
		     n < VCU_VALID_RATES_COUNT && tok == COMMA;
		     n++)
		{
		    GET_INTEGER();
		    conf_static.rates[n] = (int)yylval.real;
		    if (n > 0 &&
			conf_static.rates[n] <= conf_static.rates[n-1])
		    {
			yy_perror("rates are not in order");
			exit(EXIT_FAILURE);
		    }
		    for (i = 0; i < VCU_VALID_RATES_COUNT; i++) {
			if (conf_static.rates[n] == rates[i]) {
			    break;
			}
		    }
		    if (i == VCU_VALID_RATES_COUNT) {
			yy_perror("unsupported rate: %d Hz",
				  conf_static.rates[n]);
			exit(EXIT_FAILURE);
		    }
		    tok = yylex();
		}
		if (tok != RBRACE) {
		    yy_perror_tok(RBRACE, tok);
		    exit(EXIT_FAILURE);
		}
                GET_TOKEN(EOS);
		conf_static.n_rates = n;
            } else if (strcmp(yylval.field, "master_input_device") == 0) {
                TEST_AND_SET(2);
                GET_INTEGER();
                conf_static.master_input_device = (int)yylval.real - 1;
                if (conf_static.master_input_device < 0) {
                    yy_perror("invalid index");
                    exit(EXIT_FAILURE);
                }
                GET_TOKEN(EOS);
            } else if (strcmp(yylval.field, "allow_poll_mode") == 0) {
                TEST_AND_SET(3);
                GET_TOKEN(BOOLEAN);
                conf_static.allow_poll_mode = yylval.boolean;
                GET_TOKEN(EOS);
	    } else if (strstr(yylval.field, "input_device") == yylval.field) {
                if ((n = atoi(&yylval.field[12])) != 0) {
                    n -= 1;
                }
                if (n < 0 || n >= CONF_DEVICES_MAX) {
                    yy_perror("invalid input_device index %d, valid range "
                              "is 1 - %d", n + 1, CONF_DEVICES_MAX);
		    exit(EXIT_FAILURE);
                }
                if (conf_static.input.device[n][0] != '\0') {
                    yy_perror("input_device %d is already set", n + 1);
		    exit(EXIT_FAILURE);
                }
		if (!parse_device(conf_static.input.device[n],
				  conf_static.input.device_params[n]))
		{
		    exit(EXIT_FAILURE);
		}
	    } else if (strstr(yylval.field, "input_channels") == yylval.field) {
                if ((n = atoi(&yylval.field[14])) != 0) {
                    n -= 1;
                }
                if (n < 0 || n >= CONF_DEVICES_MAX) {
                    yy_perror("invalid input_channels index %d, valid range "
                              "is 1 - %d", n + 1, CONF_DEVICES_MAX);
		    exit(EXIT_FAILURE);
                }
                if (conf_static.input.open_channels[n] != 0) {
                    yy_perror("input_channels %d is already set", n + 1);
		    exit(EXIT_FAILURE);
                }
		if (!parse_channels(&conf_static.input.open_channels[n],
				    conf_static.input.channel_index[n]))
		{
		    exit(EXIT_FAILURE);
		}
	    } else if (strcmp(yylval.field, "input_formats") == 0) {
		TEST_AND_SET(4);
		if (!parse_formats(&conf_static.input.n_sample_formats,
				   conf_static.input.sample_formats))
		{
		    exit(EXIT_FAILURE);
		}
            } else if (strcmp(yylval.field, "min_device_blocksize") == 0) {
                TEST_AND_SET(5);
                GET_INTEGER();
                conf_static.min_device_blocksize = (int)yylval.real;
                if (conf_static.min_device_blocksize != 0 &&
                    conf_static.min_device_blocksize <
                    BFIR_MIN_BLOCKSIZE_LOWLAT)
                {
                    yy_perror("too small min_device_blocksize, %d is the "
                              "smallest allowed", BFIR_MIN_BLOCKSIZE_LOWLAT);
                    exit(EXIT_FAILURE);
                }
                if (conf_static.min_device_blocksize != 0 &&
                    conf_static.min_device_blocksize >
                    BFIR_MAX_BLOCKSIZE_48000HZ)
                {
                    yy_perror("too large min_device_blocksize, %d is the "
                              "largest allowed", BFIR_MAX_BLOCKSIZE_48000HZ);
                    exit(EXIT_FAILURE);
                }
                if (conf_static.min_device_blocksize != 0 &&
                    log2_get(conf_static.min_device_blocksize) == -1)
                {
                    yy_perror("min_device_blocksize is not a power of two");
                    exit(EXIT_FAILURE);
                }
                GET_TOKEN(EOS);
	    } else if (strstr(yylval.field, "output_device") == yylval.field) {
                if ((n = atoi(&yylval.field[13])) != 0) {
                    n -= 1;
                }
                if (n < 0 || n >= CONF_DEVICES_MAX) {
                    yy_perror("invalid output_device index %d, valid range "
                              "is 1 - %d", n + 1, CONF_DEVICES_MAX);
		    exit(EXIT_FAILURE);
                }
                if (conf_static.output.device[n][0] != '\0') {
                    yy_perror("output_device %d is already set", n + 1);
		    exit(EXIT_FAILURE);
                }
		if (!parse_device(conf_static.output.device[n],
				  conf_static.output.device_params[n]))
		{
		    exit(EXIT_FAILURE);
		}
	    } else if (strstr(yylval.field, "output_channels")
                       == yylval.field)
            {
                if ((n = atoi(&yylval.field[15])) != 0) {
                    n -= 1;
                }
                if (n < 0 || n >= CONF_DEVICES_MAX) {
                    yy_perror("invalid input_channels index %d, valid range "
                              "is 1 - %d", n + 1, CONF_DEVICES_MAX);
		    exit(EXIT_FAILURE);
                }
                if (conf_static.output.open_channels[n] != 0) {
                    yy_perror("output_channels %d is already set", n + 1);
		    exit(EXIT_FAILURE);
                }
		if (!parse_channels(&conf_static.output.open_channels[n],
				    conf_static.output.channel_index[n]))
		{
		    exit(EXIT_FAILURE);
		}
	    } else if (strcmp(yylval.field, "output_formats") == 0) {
		TEST_AND_SET(7);
		if (!parse_formats(&conf_static.output.n_sample_formats,
				   conf_static.output.sample_formats))
		{
		    exit(EXIT_FAILURE);
		}
	    } else if (strcmp(yylval.field, "ui_modules") == 0) {
                if ((isset[0] & 0x200) == 0) {
                    yy_perror("\"ui_modules_path must be set before "
                              "\"ui_modules\" is specified\n");
                    exit(EXIT_FAILURE);
                }
		TEST_AND_SET(8);
                GET_TOKEN(LBRACE);
		do {
		    if (!load_module(
			&conf_static.uimodules[conf_static.n_uimodules], &tok))
		    {
			exit(EXIT_FAILURE);
		    }
		    conf_static.n_uimodules++;
		} while (tok == COMMA &&
			 conf_static.n_uimodules < CONF_UIMODULES_MAX);
		if (tok != RBRACE) {
		    yy_perror_tok(RBRACE, tok);
		    exit(EXIT_FAILURE);
		}
                GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "uimodules_path") == 0) {
		TEST_AND_SET(9);
		GET_TOKEN(STRING);
		strncpy(conf_static.uimodules_path, yylval.string,
                        PATH_MAX - 1);
		GET_TOKEN(EOS);
                path_make_absolute(conf_static.uimodules_path, true, true);
	    } else if (strcmp(yylval.field, "brutefir_path") == 0) {
		TEST_AND_SET(10);
		GET_TOKEN(STRING);
		strncpy(conf_static.brutefir_path, yylval.string, PATH_MAX - 1);
		GET_TOKEN(EOS);
                path_make_absolute(conf_static.brutefir_path, false, true);
	    } else if (strcmp(yylval.field, "brutefir_modules_path") == 0) {
		TEST_AND_SET(11);
		GET_TOKEN(STRING);
		strncpy(conf_static.brutefir_modules_path, yylval.string,
                        PATH_MAX - 1);
		GET_TOKEN(EOS);
                path_make_absolute(conf_static.brutefir_modules_path, true,
                                   true);
	    } else if (strcmp(yylval.field, "log_path") == 0) {
		TEST_AND_SET(12);
		GET_TOKEN(STRING);
		strncpy(conf_static.log_path, yylval.string, PATH_MAX - 1);
                path_make_absolute(conf_static.log_path, true, true);
                path_create(conf_static.log_path);
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "storage_path") == 0) {
		TEST_AND_SET(13);
		GET_TOKEN(STRING);
		strncpy(conf_static.storage_path, yylval.string, PATH_MAX - 1);
                path_make_absolute(conf_static.storage_path, true, true);
                path_create(conf_static.storage_path);
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "cpu_temp_proc_path") == 0) {
		TEST_AND_SET(14);
		GET_TOKEN(STRING);
		strncpy(conf_static.cpu_temp_proc_path, yylval.string,
                        PATH_MAX - 1);
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "motherboard_temp_proc_path")
                       == 0)
            {
		TEST_AND_SET(15);
		GET_TOKEN(STRING);
		strncpy(conf_static.motherboard_temp_proc_path, yylval.string,
                        PATH_MAX - 1);
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "storage_size") == 0) {
		TEST_AND_SET(16);
		GET_INTEGER();
		conf_static.storage_size = (int)yylval.real;
		if (conf_static.storage_size < CONF_STORAGE_MIN_SIZE) {
		    yy_perror("too small \"storage_size\" value, need "
			      "at least %d megabytes", CONF_STORAGE_MIN_SIZE);
		    exit(EXIT_FAILURE);
		}
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "run_as_daemon") == 0) {
                TEST_AND_SET(17);
                GET_TOKEN(BOOLEAN);
                conf_static.run_as_daemon = yylval.boolean;
                GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "suspend_in_runtime") == 0) {
                TEST_AND_SET(18);
                GET_TOKEN(BOOLEAN);
                conf_static.suspend_in_runtime = yylval.boolean;
                GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "suspend_command") == 0) {
                TEST_AND_SET(19);
		GET_TOKEN(STRING);
		strncpy(conf_static.suspend_command, yylval.string,
                        PATH_MAX - 1);
                GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "log_rotate_size") == 0) {
		TEST_AND_SET(20);
		GET_INTEGER();
		conf_static.log_rotate_size = (int)yylval.real;
		GET_TOKEN(EOS);
	    } else {
		yy_perror("unrecognised field \"%s\"", yylval.string);
		exit(EXIT_FAILURE);
	    }
	    break;
	default:
	    yy_perror_tok(FIELD, tok);
	    exit(EXIT_FAILURE);
	}
    }

    if (tok == INVALID_INPUT) {
        yy_perror("syntax error");
        exit(EXIT_FAILURE);
    }
    for (n = 0; n < CONF_DEVICES_MAX; n++) {
        if (conf_static.input.device[n][0] != '\0' &&
            conf_static.input.open_channels[n] == 0)
        {
            yy_perror("missing input_channels%d", n + 1);
            exit(EXIT_FAILURE);
        }
        if (conf_static.input.device[n][0] == '\0' &&
            conf_static.input.open_channels[n] != 0)
        {
            yy_perror("missing input_device%d", n + 1);
            exit(EXIT_FAILURE);
        }
        conf_static.input.total_open_channels +=
            conf_static.input.open_channels[n];
        if (conf_static.input.device[n][0] == '\0') {
            if (conf_static.input.n_devices == 0) {
                conf_static.input.n_devices = n;
            }
        } else if (conf_static.input.n_devices != 0) {
            yy_perror("input device list has holes");
            exit(EXIT_FAILURE);
        } else if (conf_static.input.total_open_channels > VCU_MAX_CHANNELS) {
            yy_perror("too many input channels (%d max allowed)",
                      VCU_MAX_CHANNELS);
            exit(EXIT_FAILURE);
        }
    }
    if (conf_static.input.n_devices == 0 &&
        conf_static.input.device[0][0] != '\0')
    {
        conf_static.input.n_devices = CONF_DEVICES_MAX;
    }
    if (conf_static.input.n_devices == 0) {
        yy_perror("missing input_device");
        exit(EXIT_FAILURE);
    }
    if (conf_static.master_input_device >= conf_static.input.n_devices) {
        yy_perror("master_input_device is out of range");
        exit(EXIT_FAILURE);
    }
    
    for (n = 0; n < CONF_DEVICES_MAX; n++) {
        if (conf_static.output.device[n][0] != '\0' &&
            conf_static.output.open_channels[n] == 0)
        {
            yy_perror("missing output_channels%d", n + 1);
            exit(EXIT_FAILURE);
        }
        if (conf_static.output.device[n][0] == '\0' &&
            conf_static.output.open_channels[n] != 0)
        {
            yy_perror("missing output_device%d", n + 1);
            exit(EXIT_FAILURE);
        }
        conf_static.output.total_open_channels +=
            conf_static.output.open_channels[n];
        if (conf_static.output.device[n][0] == '\0') {
            if (conf_static.output.n_devices == 0) {
                conf_static.output.n_devices = n;
            }
        } else if (conf_static.output.n_devices != 0) {
            yy_perror("output device list has holes");
            exit(EXIT_FAILURE);
        } else if (conf_static.output.total_open_channels > VCU_MAX_CHANNELS) {
            yy_perror("too many output channels (%d max allowed)",
                      VCU_MAX_CHANNELS);
            exit(EXIT_FAILURE);
        }
    }
    if (conf_static.output.n_devices == 0 &&
        conf_static.output.device[0][0] != '\0')
    {
        conf_static.output.n_devices = CONF_DEVICES_MAX;
    }
    if (conf_static.output.n_devices == 0) {
        yy_perror("missing output_device");
        exit(EXIT_FAILURE);
    }
    missing_field = false;
    MISSING_FIELD("allocate_ram", 0);
    MISSING_FIELD("rates", 1);
    MISSING_FIELD("input_formats", 4);
    MISSING_FIELD("output_formats", 7);
    MISSING_FIELD("ui_modules", 8);
    MISSING_FIELD("uimodules_path", 9);
    MISSING_FIELD("brutefir_path", 10);
    MISSING_FIELD("brutefir_modules_path", 11);
    MISSING_FIELD("log_path", 12);
    MISSING_FIELD("storage_path", 13);
    MISSING_FIELD("storage_size", 16);
    if (missing_field) {
        elog(LOGERROR, "field missing in static configuration file, aborting.");
	exit(EXIT_FAILURE);
    }
    return true;
}

#undef MISSING_FIELD

static bool_t
parse_revprg(struct conf_revprg *revprg,
             const struct filterprg *fprg)
{
    uint32_t isset[1] = { 0 };
    int tok;
    
    while ((tok = yylex()) > 0) {
	switch (tok) {
	case FIELD:
	    if (strcmp(yylval.field, "volume") == 0) {
		TEST_AND_SET(0);
		GET_TOKEN(REAL);
		revprg->volume = yylval.real;
		limit_range(&revprg->volume, VCU_VOLUME_MIN, VCU_VOLUME_MAX);
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "delay") == 0) {
		TEST_AND_SET(1);
		GET_TOKEN(REAL);
		revprg->delay = yylval.real;
		limit_range(&revprg->delay, 0, VCU_REVERB_DELAY_RANGE);
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "scale_length") == 0) {
		TEST_AND_SET(2);
		GET_TOKEN(REAL);
		revprg->scale_length = yylval.real;
		GET_TOKEN(EOS);
		limit_range(&revprg->scale_length, 0, 1);
            } else if (strcmp(yylval.field, "response_mode") == 0) {
                TEST_AND_SET(3);
                GET_TOKEN(STRING);
                if (strcmp(yylval.string, "full") == 0) {
                    revprg->response_mode = VCU_REVERB_RESP_MODE_FULL;
                } else if (strcmp(yylval.string, "mute_direct") == 0) {
                    revprg->response_mode = VCU_REVERB_RESP_MODE_MUTE_DIRECT;
                } else if (strcmp(yylval.string, "mute_early") == 0) {
                    revprg->response_mode = VCU_REVERB_RESP_MODE_MUTE_EARLY;
                } else if (strcmp(yylval.string, "stage_adapted") == 0) {
                    revprg->response_mode = VCU_REVERB_RESP_MODE_STAGE_ADAPTED;
                } else {
                    revprg->response_mode = VCU_REVERB_RESP_MODE_FULL;
                }
                GET_TOKEN(EOS);
            } else if (strcmp(yylval.field, "decorrelate_mode") == 0) {
                TEST_AND_SET(4);
                GET_TOKEN(STRING);
                if (strcmp(yylval.string, "none") == 0) {
                    revprg->decorrelate_mode = VCU_REVERB_DECORR_MODE_NONE;
                } else if (strcmp(yylval.string, "needed") == 0) {
                    revprg->decorrelate_mode = VCU_REVERB_DECORR_MODE_NEEDED;
                } else if (strcmp(yylval.string, "all") == 0) {
                    revprg->decorrelate_mode = VCU_REVERB_DECORR_MODE_ALL;
                } else {
                    revprg->decorrelate_mode = VCU_REVERB_DECORR_MODE_NEEDED;
                }
                GET_TOKEN(EOS);
            } else if (strcmp(yylval.field, "force_delay_diff") == 0) {
                TEST_AND_SET(5);
                GET_TOKEN(BOOLEAN);
                revprg->force_delay_diff = yylval.boolean;
                GET_TOKEN(EOS);
            } else if (strcmp(yylval.field, "force_decay") == 0) {
                TEST_AND_SET(6);
                GET_TOKEN(BOOLEAN);
                revprg->force_decay = yylval.boolean;
                GET_TOKEN(EOS);
            } else if (strcmp(yylval.field, "mix_with_dry") == 0) {
                TEST_AND_SET(7);
                GET_TOKEN(BOOLEAN);
                revprg->mix_with_dry = yylval.boolean;
                GET_TOKEN(EOS);
            } else if (strcmp(yylval.field, "dry_volume") == 0) {
                TEST_AND_SET(8);
                GET_TOKEN(REAL);
                revprg->dry_volume = yylval.real;
                GET_TOKEN(EOS);
		limit_range(&revprg->dry_volume,
                            VCU_VOLUME_MIN, VCU_VOLUME_MAX);
            } else if (strcmp(yylval.field, "bformat_directivity") == 0) {
                TEST_AND_SET(9);
                GET_TOKEN(REAL);
                revprg->bformat_directivity = yylval.real;
                GET_TOKEN(EOS);
                limit_range(&revprg->bformat_directivity,
                            REVERB_BFORMAT_DIRECTIVITY_MIN,
                            REVERB_BFORMAT_DIRECTIVITY_MAX);
            } else if (strcmp(yylval.field, "xyz_balance") == 0) {
                TEST_AND_SET(10);
                GET_TOKEN(REAL);
                revprg->xyz_balance[0] = yylval.real;
                GET_TOKEN(COMMA);
                GET_TOKEN(REAL);
                revprg->xyz_balance[1] = yylval.real;
                GET_TOKEN(COMMA);
                GET_TOKEN(REAL);
                revprg->xyz_balance[2] = yylval.real;
                GET_TOKEN(EOS);
	    } else {
                elog(LOGDEBUG, "skipping unrecognised field \"%s\"",
                     yylval.string);
                ignore_to_eos();
	    }
	    break;
	default:
	    yy_perror_tok(FIELD, tok);
	    return false;
	}
    }
    if (tok == INVALID_INPUT) {
        return false;
    }
    return true;
}

void
conf_load_static(const char path[])
{
    FILE *stream;
    
    memset(&conf_static, 0, sizeof(conf_static));

    if ((stream = fopen(path, "rt")) == NULL) {
	elog(LOGERROR, "Failed to open \"%s\" for reading: %s.\n",
	     path, strerror(errno));
	exit(EXIT_FAILURE);
    }
    yy_restart(path, stream, LOGERROR);
    if (!parse_static()) {
	exit(EXIT_FAILURE);
    }
    fclose(stream);
    yy_restart(NULL, NULL, LOGERROR);
    has_loaded_static = true;
}

void
conf_load_user(void)
{
    char path[PATH_MAX + 100], pathnew[PATH_MAX + 100], *filename;
    bool_t parse_success = false;
    FILE *stream;
    int n;

    if (!has_loaded_static) {
        elog(LOGERROR, "must load static conf first");
        exit(EXIT_FAILURE);
    }
    sprintf(path, "%s%s", conf_static.storage_path, CONF_FILE_USER);
    filename = path;
    if ((stream = fopen(filename, "rt")) == NULL) {
	sprintf(pathnew, "%s.new", path);
        filename = pathnew;
	stream = fopen(filename, "rt");
    }
    if (stream != NULL) {
	yy_restart(filename, stream, LOGERROR);
	parse_success = parse_conf_user(stream);
	yy_restart(NULL, NULL, LOGERROR);
	fclose(stream);
    }
    if (!parse_success) {
	init_conf_user();
	for (n = 0; n < conf_static.n_uimodules; n++) {
	    if (conf_static.uimodules[n]->loaduserconf != NULL) {
		conf_static.uimodules[n]->loaduserconf(NULL);
	    }
	}
	sprintf(pathnew, "%s.failed", path);
        copyfile(pathnew, path);
    }
    conf_save_user();
}

void
conf_save_user(void)
{
    char path[PATH_MAX + 100], pathnew[PATH_MAX + 100], name[10];
    struct vcu_eqpart *eqpart;
    int n, i, j, type, io;
    bool_t first, first2;
    FILE *stream;

    sprintf(path, "%s%s", conf_static.storage_path, CONF_FILE_USER);
    sprintf(pathnew, "%s.new", path);
    if ((stream = fopen(pathnew, "wt")) == NULL) {
	elog(LOGERROR, "Failed to open \"%s\" for writing: %s.\n",
	     pathnew, strerror(errno));
	exit(EXIT_FAILURE);
    }    
    
    fprintf(stream, "\n\
\"volume\" {\n\
    master: %g;\n\
    group: ",
	    cu.volume.master);
    print_array(stream, REAL, cu.volume.group, cu.channel.group_count);
    fprintf(stream, ";\n    mute:\n      { ");
    FOR_IN_AND_OUT {
        fprintf(stream, "{ ");
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            print_array(stream, BOOLEAN, cu.volume.mute[io][type],
                        cu.channel.count[io][type]);
            fprintf(stream, (type != VCU_CH_WP_COUNT - 1) ?
                    ",\n          " : " ");
        }
        fprintf(stream, io == IN ? "},\n        " : "} };\n");
    }
    fprintf(stream, "    group_mute: ");
    print_array(stream, BOOLEAN, cu.volume.group_mute, cu.channel.group_count);
    fprintf(stream, ";\n\
    master_temp_mute: %s;\n",
            cu.volume.master_temp_mute ? "true" : "false");
    fprintf(stream, "};\n\
\n\
\"signal\" {\n\
    rates: ");
    print_array(stream, INTEGER, cu.signal.rates, cu.signal.n_rates);
    fprintf(stream, ";\n\
    suggested_rate: %d;\n\
    output_format: \"%s\";\n\
    soft_clip: %s;\n\
    dither: %s;\n\
    watch_clock: %s;\n\
    double_precision: %s;\n\
};",
	    cu.signal.suggested_rate,
	    bf_strsampleformat(cu.signal.output_sample_format),
	    cu.signal.use_soft_clip ? "true" : "false",
	    cu.signal.apply_dither ? "true" : "false",
            cu.signal.watch_clock ? "true" : "false",
            cu.signal.double_precision ? "true" : "false");

    fprintf(stream, "\n\n\
\"channel\" {\n\
    count: ");
    print_array(stream, INTEGER, cu.channel.count[OUT], VCU_CH_COUNT);
    fprintf(stream, ";\n\
    mapping:\n      { ");
    strcpy(name, "ARBP");
    first = true;
    FOR_IN_AND_OUT {
        for (type = 0; type < VCU_CH_COUNT; type++) {
            for (n = 0; n < cu.channel.count[io][type]; n++) {
                if (!first) {
                    fprintf(stream, ",\n        ");
                } else {
                    first = false;
                }
                first2 = true;
                fprintf(stream, "{ \"%c%c%d\", { ", io == IN ? 'i' : 'o',
                        name[type], n);
                i = -1;
                while ((i = bit_find(cu.channel.map[io][type][n], i + 1,
                                     cu.channel.count[io][VCU_CH_PHYSICAL] - 1))
                       != -1)
                {
                    if (!first2) {
                        fprintf(stream, ", ");
                    } else {
                        first2 = false;
                    }
                    fprintf(stream, "\"%c%d\"", io == IN ? 'i' : 'o', i);
                }
                fprintf(stream, " } }");
            }
        }
    }
    fprintf(stream, " };\n");
    fprintf(stream, "    speaker_positions:\n      { ");
    for (n = 0; n < conf_static.output.total_open_channels; n++) {
	fprintf(stream, "{ %s, %g, %g, %g }%s",
                cu.channel.speakerpos[n].active ? "true" : "false",
		cu.channel.speakerpos[n].azimut,
		cu.channel.speakerpos[n].elevation,
		cu.channel.speakerpos[n].distance,
		(n != conf_static.output.total_open_channels - 1) ?
		",\n        " : " };\n");
    }
    fprintf(stream, "    stages:\n      { ");
    for (n = 0; n < cu.channel.stage_count; n++) {
	fprintf(stream, "{ %g, %g, %g, %g }%s",
		cu.channel.stage[n].azimut,
		cu.channel.stage[n].elevation,
		cu.channel.stage[n].width,
		cu.channel.stage[n].height,
		(n != cu.channel.stage_count - 1) ?
		",\n        " : "");
    }
    fprintf(stream, " };\n");
    fprintf(stream, "    distance_compensate: %s;\n",
            cu.channel.distance_compensate ? "true" : "false");
    fprintf(stream, "    trim_volumes:\n      { ");
    FOR_IN_AND_OUT {
        fprintf(stream, "{ ");
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            print_array(stream, REAL, cu.channel.trim.volume[io][type],
                        cu.channel.count[io][type]);
            fprintf(stream, (type != VCU_CH_WP_COUNT - 1) ?
                    ",\n          " : " ");
        }
        fprintf(stream, io == IN ? "},\n        " : "} };\n");
    }
    fprintf(stream, "    trim_delays:\n      { ");
    FOR_IN_AND_OUT {
        fprintf(stream, "{ ");
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            print_array(stream, REAL, cu.channel.trim.delay[io][type],
                        cu.channel.count[io][type]);
            fprintf(stream, (type != VCU_CH_WP_COUNT - 1) ?
                    ",\n          " : " ");
        }
        fprintf(stream, io == IN ? "},\n        " : "} };\n");
    }
    
    fprintf(stream, "    groups:\n      { ");
    for (n = 0; n < cu.channel.group_count; n++) {
        fprintf(stream, "{ \"%s\",\n          ", cu.channel.group[n].name);
        FOR_IN_AND_OUT {
            fprintf(stream, "{ ");
            for (type = 0; type < VCU_CH_COUNT; type++) {
                print_array(stream, BOOLEAN_BIT,
                            cu.channel.group[n].channels[io][type],
                            cu.channel.count[io][type]);
                fprintf(stream, (type != VCU_CH_COUNT - 1) ? ", " : " }");
            }
            if (io == OUT) {
                fprintf(stream, "\n        ");
            } else {
                fprintf(stream, ",\n          ");
            }
        }
        fprintf(stream, (n != cu.channel.group_count - 1) ? " },\n        " :
                " }");
    }
    fprintf(stream, " };\n");

    fprintf(stream, "    modes:\n      { ");
    for (n = 0; n < cu.channel.mode_count; n++) {
        fprintf(stream, "{ \"%s\", ", cu.channel.mode[n].name);
        print_array(stream, BOOLEAN_BIT, cu.channel.mode[n].groups,
                    cu.channel.group_count);
        fprintf(stream, ", ");
        print_array(stream, BOOLEAN_BIT, cu.channel.mode[n].eqmodes,
                    cu.eq.mode_count);
        fprintf(stream, ", %d", cu.channel.mode[n].active_eq_mode);
        fprintf(stream, (n != cu.channel.mode_count - 1) ? " },\n        " :
                " }");
    }
    fprintf(stream, " };\n");
    fprintf(stream, "    active_mode: %d;\n", cu.channel.active_mode);
    
    fprintf(stream, "};\n\n\"equalisation\" {\n");

    fprintf(stream, "    modes:\n      { ");
    for (n = 0; n < cu.eq.mode_count; n++){
        fprintf(stream, "{ \"%s\",\n          ", cu.eq.mode[n].name);
        FOR_IN_AND_OUT {
            fprintf(stream, "{ ");
            for (type = 0; type < VCU_CH_WP_COUNT; type++) {
                print_array(stream, INTEGER,
                            cu.eq.mode[n].prgs[io][type],
                            cu.channel.count[io][type]);
                fprintf(stream, (type != VCU_CH_WP_COUNT - 1) ? ", " : " }");
            }
            if (io == OUT) {
                fprintf(stream, "\n        ");
            } else {
                fprintf(stream, ",\n          ");
            }
        }
        fprintf(stream, (n != cu.eq.mode_count - 1) ? " },\n        " : " }");
    }
    fprintf(stream, " };\n");
    fprintf(stream, "    active_mode: %d;\n", cu.eq.active_mode);

    fprintf(stream, "    equalisers:\n      { ");
    for (n = 0; n < cu.eq.eqprg_count; n++) {
        fprintf(stream, "{ \"%s\", ", cu.eq.eqprg[n].name);
        for (i = 0; i < cu.eq.eqprg[n].part_count; i++) {
            eqpart = &cu.eq.eqprg[n].parts[i];
            fprintf(stream, "{ ");
            switch (eqpart->type) {
            case VCU_EQPRG_TYPE_OP_CONVOLVE:
                fprintf(stream, "\"convolve\"");
                break;
            case VCU_EQPRG_TYPE_OP_ADD:
                fprintf(stream, "\"add\"");
                break;
            case VCU_EQPRG_TYPE_OP_SUBTRACT:
                fprintf(stream, "\"subtract\"");
                break;
            case VCU_EQPRG_TYPE_OP_MULTIPLY:
                fprintf(stream, "\"multiply\" / { %g, %d }",
                        eqpart->params.multiply.amp,
                        eqpart->params.multiply.sign);
                break;
            case VCU_EQPRG_TYPE_LOWPASS:
                fprintf(stream,
                        "\"lowpass\" / { %g, %g, \"%s\" }",
                        eqpart->params.lowpass.cutoff,
                        eqpart->params.lowpass.slope,
                        phasestr(eqpart->params.lowpass.phase));
                break;
            case VCU_EQPRG_TYPE_HIGHPASS:
                fprintf(stream,
                        "\"highpass\" / { %g, %g, \"%s\" }",
                        eqpart->params.highpass.cutoff,
                        eqpart->params.highpass.slope,
                        phasestr(eqpart->params.highpass.phase));
                break;
            case VCU_EQPRG_TYPE_BANDPASS:
                fprintf(stream,
                        "\"bandpass\" / { %g, %g, %g, \"%s\" }",
                        eqpart->params.bandpass.cutoff[0],
                        eqpart->params.bandpass.cutoff[1],
                        eqpart->params.bandpass.slope,
                        phasestr(eqpart->params.bandpass.phase));
                break;
            case VCU_EQPRG_TYPE_BANDSTOP:
                fprintf(stream,
                        "\"bandstop\" / { %g, %g, %g, \"%s\" }",
                        eqpart->params.bandstop.cutoff[0],
                        eqpart->params.bandstop.cutoff[1],
                        eqpart->params.bandstop.slope,
                        phasestr(eqpart->params.bandstop.phase));
                break;
            case VCU_EQPRG_TYPE_ALLPASS:
                fprintf(stream, "\"allpass\" / {  }");
                break;
            case VCU_EQPRG_TYPE_CUSTOM:
                fprintf(stream,
                        "\"custom\" / { \"%s\" }",
                        filterprg_prgid_ntoa(eqpart->params.custom.id));
                break;
            case VCU_EQPRG_TYPE_RANDOM_PHASE:
                fprintf(stream,
                        "\"random_phase\" / { %g }",
                        eqpart->params.random_phase.length);
                break;
            case VCU_EQPRG_TYPE_FFT_DYNAMIC:
            case VCU_EQPRG_TYPE_FFT_STATIC:
                if (eqpart->type == VCU_EQPRG_TYPE_FFT_DYNAMIC) {
                    fprintf(stream, "\"fft_dynamic\" / { %d, ",
                            eqpart->params.fft.n_bands);
                } else {
                    fprintf(stream, "\"fft_static\" / { %d, ",
                            eqpart->params.fft.n_bands);
                }
                for (j = 0; j < eqpart->params.fft.n_bands; j++) {
                    fprintf(stream, "%.1f/%g",
                            eqpart->params.fft.band[j],
                            eqpart->params.fft.amp[j]);
                    fprintf(stream,
                            (j != eqpart->params.fft.n_bands - 1) ?
                            ", " : " }");
                }
                break;
            default:
                elog(LOGERROR, "unknown eq program type\n");
                exit(EXIT_FAILURE);
                break;
            }
            if (i < cu.eq.eqprg[n].part_count - 1) {
                fprintf(stream, " }, ");
            } else {
                fprintf(stream, " }");
            }
        }
        if (n == cu.eq.eqprg_count - 1) {
            fprintf(stream, " }");
        } else {
            fprintf(stream, " },\n        ");
        }
    }
    fprintf(stream, " };\n");
    
    if (cu.eq.n_loaded_programs > 0) {
	fprintf(stream, "    programs: { ");
	for (n = 0; n < cu.eq.n_loaded_programs; n++) {
	    fprintf(stream, "\"%s\"%s",
		    filterprg_prgid_ntoa(cu.eq.loaded_programs[n]),
		    (n != cu.eq.n_loaded_programs - 1) ? ", " : " };\n");
	}
    }
    fprintf(stream, "};\n\n\"ambiopole\" {\n\
    xtalk_cancel: ");
    print_array(stream, BOOLEAN, cu.ambiopole.use_xtalk_cancel,
                cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2);
    fprintf(stream, ";\n");
    if (cu.channel.count[OUT][VCU_CH_AMBIOPOLE] > 0) {
	fprintf(stream, "    current_program:\n      { ");
	for (n = 0; n < cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2; n++) {
	    fprintf(stream, "\"%s\"%s",
                    filterprg_prgid_ntoa(cu.ambiopole.current[n]),
		    (n != cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2 - 1) ?
		    ",\n        " : " };\n");
	}
    }
    if (cu.ambiopole.n_loaded_programs > 0) {
	fprintf(stream, "    programs: { ");	
	for (n = 0; n < cu.ambiopole.n_loaded_programs; n++) {
	    fprintf(stream, "\"%s\"%s",
		    filterprg_prgid_ntoa(
			cu.ambiopole.loaded_programs[n]),
		    (n != cu.ambiopole.n_loaded_programs - 1) ?
		    ", " : " };\n");
	}
    }
    fprintf(stream, "};\n\n\"reverb\" {\n\
    benchmark_cpu_count: %d;\n\
    benchmark_cpu_speed: %d;\n\
    benchmark_cpu_model: \"%s\";\n\
    benchmark_memsize: %d;\n\
    iodelay_manually_set: %s;\n\
    max_benchmark_load: %g;\n\
    max_benchmark_length: %g;\n\
    block_sizes: ",
	    cu.reverb.benchmark_sysinfo.cpu_count,
	    cu.reverb.benchmark_sysinfo.cpu_speed,
	    cu.reverb.benchmark_sysinfo.cpu_modelname,
	    cu.reverb.benchmark_sysinfo.mem_total,
            cu.reverb.iodelay_manually_set ? "true" : "false",
            cu.reverb.max_benchmark_load,
            cu.reverb.max_benchmark_length);
    for (n = 0; n < cu.signal.n_rates; n++) {
	fprintf(stream, "{ %u, %u }%s",
		cu.reverb.block_sizes[n],
		cu.reverb.n_blocks[n],
		(n != cu.signal.n_rates - 1) ? ", " : ";\n");
    }
    fprintf(stream, "    current_program: \"%s\";\n",
	    filterprg_prgid_ntoa(cu.reverb.current));
    if (cu.reverb.n_loaded_programs > 0) {
	fprintf(stream, "    programs: { ");	
	for (n = 0; n < cu.reverb.n_loaded_programs; n++) {
	    fprintf(stream, "\"%s\"%s",
		    filterprg_prgid_ntoa(cu.reverb.loaded_programs[n]),
		    (n != cu.reverb.n_loaded_programs - 1) ?
		    ", " : " };\n");
	}
    }
    switch (cu.reverb.response_mode) {
    case VCU_REVERB_RESP_MODE_FULL:
        fprintf(stream, "    response_mode: \"full\";\n");
        break;
    case VCU_REVERB_RESP_MODE_MUTE_DIRECT:
        fprintf(stream, "    response_mode: \"mute_direct\";\n");
        break;
    case VCU_REVERB_RESP_MODE_MUTE_EARLY:
        fprintf(stream, "    response_mode: \"mute_early\";\n");
        break;
    case VCU_REVERB_RESP_MODE_STAGE_ADAPTED:
        fprintf(stream, "    response_mode: \"stage_adapted\";\n");
        break;
    }
    switch (cu.reverb.decorrelate_mode) {
    case VCU_REVERB_DECORR_MODE_NONE:
        fprintf(stream, "    decorrelate_mode: \"none\";\n");
        break;
    case VCU_REVERB_DECORR_MODE_NEEDED:
        fprintf(stream, "    decorrelate_mode: \"needed\";\n");
        break;
    case VCU_REVERB_DECORR_MODE_ALL:
        fprintf(stream, "    decorrelate_mode: \"all\";\n");
        break;
    }
    fprintf(stream, "    force_delay_diff: %s;\n", cu.reverb.force_delay_diff ?
            "true" : "false");
    fprintf(stream, "    force_decay: %s;\n", cu.reverb.force_decay ?
            "true" : "false");
    fprintf(stream, "    mix_with_dry: %s;\n", cu.reverb.mix_with_dry ?
            "true" : "false");
    fprintf(stream, "    dry_volume: %g;\n", cu.reverb.dry_volume);
    fprintf(stream, "    bformat_directivity: %g;\n\
    source_layout: ",
            cu.reverb.bformat_directivity);
    switch (cu.reverb.source_layout) {
    case VCU_REVERB_SOURCE_LAYOUT_CENTER:
        fprintf(stream, "\"center\";\n");
        break;
    case VCU_REVERB_SOURCE_LAYOUT_STEREO:
        fprintf(stream, "\"stereo\";\n");
        break;
    case VCU_REVERB_SOURCE_LAYOUT_LRC:
        fprintf(stream, "\"lrc\";\n");
        break;
    }
    fprintf(stream, "};\n\n");
    fprintf(stream, "\"ambisonics\" {\n\
    input_format: ");
    switch (cu.ambisonics.input_format) {
    case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:
        fprintf(stream, "\"first order\";\n");
        break;
    case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
        fprintf(stream, "\"second order\";\n");
        break;
    case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
        fprintf(stream, "\"uhj\";\n");
        break;
    }
    fprintf(stream, "};\n\n");
    fprintf(stream, "\"ctl\" {\n\
    powersave: %s;\n\
    load_user_config: %s;\n\
    load_revprg_config: %s;\n\
    load_static_config: %s;\n\
    debug: %s;\n",
            cu.ctl.powersave ? "true" : "false",
            cu.ctl.load_user_config ? "true" : "false",
            cu.ctl.load_revprg_config ? "true" : "false",
            cu.ctl.load_static_config ? "true" : "false",
            cu.ctl.debug ? "true" : "false");
    if (cu.ctl.n_fprogrefs > 0) {
        fprintf(stream, "    filter_program_references: { ");
        for (n = 0; n < cu.ctl.n_fprogrefs; n++) {
            fprintf(stream, "{ \"%s\", \"%s\" }",
                    filterprg_prgid_ntoa(cu.ctl.fprogrefs[n]->id),
                    cu.ctl.fprogrefs[n]->path);
            fprintf(stream, (n != cu.ctl.n_fprogrefs - 1) ? ", " : " }");
        }
        fprintf(stream, ";\n");
    }
    fprintf(stream, "};\n");
    
    for (n = 0; n < conf_static.n_uimodules; n++) {
	if (conf_static.uimodules[n]->saveuserconf != NULL) {
	    break;
	}
    }
    if (n < conf_static.n_uimodules) {
	custom_user_config_stream = stream;
	fprintf(stream, "\n# custom module configuration sections below\n\n");
	for (; n < conf_static.n_uimodules; n++) {
	    if (conf_static.uimodules[n]->saveuserconf != NULL) {
		fprintf(stream, "\"%s\" {\n",
			conf_static.uimodules[n]->name());
		conf_static.uimodules[n]->saveuserconf(put_config_token);
		fprintf(stream, "};\n");
	    }
	}
    }
    fflush(stream);
    if (fsync(fileno(stream)) == -1) {
	elog(LOGERROR, "Synching of file \"%s\" failed: %s.\n",
             pathnew, strerror(errno));
	exit(EXIT_FAILURE);
    }
    fclose(stream);
    
    if (copyfile(path, pathnew) != 0) {
	exit(EXIT_FAILURE);
    }
    remove(pathnew);
}
	    
	    

void
conf_load_revprg(struct conf_revprg *revprg,
		 const struct filterprg *fprg)
{
    char path[PATH_MAX + FILTERPRG_MD5SUM_MAX + 100];    
    FILE *stream;
    
    if (revprg == NULL || fprg == NULL) {
	return;
    }

    memset(revprg, 0, sizeof(struct conf_revprg));
    revprg->id = fprg->id;
    revprg->volume = 0;
    revprg->delay = 0;
    revprg->scale_length = 1.0;
    revprg->response_mode = cu.reverb.response_mode;
    revprg->decorrelate_mode = cu.reverb.decorrelate_mode;
    revprg->force_delay_diff = cu.reverb.force_delay_diff;
    revprg->force_decay = cu.reverb.force_decay;
    revprg->mix_with_dry = cu.reverb.mix_with_dry;
    revprg->dry_volume = cu.reverb.dry_volume;
    revprg->bformat_directivity = cu.reverb.bformat_directivity;
    revprg->xyz_balance[0] = 0;
    revprg->xyz_balance[1] = 0;
    revprg->xyz_balance[2] = 0;
    
    sprintf(path, "%s%s.revprg", conf_static.storage_path,
	    filterprg_prgid_ntoa(revprg->id));

    if ((stream = fopen(path, "rt")) == NULL) {
	/* this is ok, we use default values if file is missing or read fails */
	return;
    }
    yy_restart(path, stream, LOGDEBUG);
    parse_revprg(revprg, fprg); /* we don't care about the return value */    
    fclose(stream);
    yy_restart(NULL, NULL, LOGERROR);
    elog(LOGDEBUG, "loaded reverb program config for id %s",
         filterprg_prgid_ntoa(revprg->id));
}

void
conf_save_revprg(const struct conf_revprg *revprg)
{
    char path[PATH_MAX + FILTERPRG_MD5SUM_MAX + 100];    
    FILE *stream;

    if (revprg == NULL) {
	return;
    }
    
    sprintf(path, "%s%s.revprg", conf_static.storage_path,
	    filterprg_prgid_ntoa(revprg->id));

    if ((stream = fopen(path, "wt")) == NULL) {
	elog(LOGDEBUG, "failed to open \"%s\" for writing: %s. "
             "Reverb conf NOT saved.\n",
	     path, strerror(errno));
	return;
    }

    elog(LOGDEBUG, "id: \"%s\"; volume: %g; delay: %g;",
         filterprg_prgid_ntoa(revprg->id),
         revprg->volume, revprg->delay);
    fprintf(stream, "volume: %g;\ndelay: %g;\nscale_length: %g;\n",
	    revprg->volume,
            revprg->delay,
            revprg->scale_length);
    switch (revprg->response_mode) {
    case VCU_REVERB_RESP_MODE_FULL:
        fprintf(stream, "response_mode: \"full\";\n");
        break;
    case VCU_REVERB_RESP_MODE_MUTE_DIRECT:
        fprintf(stream, "response_mode: \"mute_direct\";\n");
        break;
    case VCU_REVERB_RESP_MODE_MUTE_EARLY:
        fprintf(stream, "response_mode: \"mute_early\";\n");
        break;
    case VCU_REVERB_RESP_MODE_STAGE_ADAPTED:
        fprintf(stream, "response_mode: \"stage_adapted\";\n");
        break;
    }
    switch (revprg->decorrelate_mode) {
    case VCU_REVERB_DECORR_MODE_NONE:
        fprintf(stream, "decorrelate_mode: \"none\";\n");
        break;
    case VCU_REVERB_DECORR_MODE_NEEDED:
        fprintf(stream, "decorrelate_mode: \"needed\";\n");
        break;
    case VCU_REVERB_DECORR_MODE_ALL:
        fprintf(stream, "decorrelate_mode: \"all\";\n");
        break;
    }
    fprintf(stream,
            "force_delay_diff: %s;\nforce_decay: %s;\nmix_with_dry: %s;\n"
            "dry_volume: %g;\nbformat_directivity: %g;\n"
            "xyz_balance: %g, %g, %g;\n",
            revprg->force_delay_diff ? "true" : "false",
            revprg->force_decay ? "true" : "false",
            revprg->mix_with_dry ? "true" : "false",
            revprg->dry_volume,
            revprg->bformat_directivity,
            revprg->xyz_balance[0],
            revprg->xyz_balance[1],
            revprg->xyz_balance[2]);
    fclose(stream);
    elog(LOGDEBUG, "saved reverb program config for id %s",
         filterprg_prgid_ntoa(revprg->id));
}

void
conf_create_pidfile(void)
{
    char path[PATH_MAX + 100], s[200];
    FILE *stream;
    bool_t retry;
    pid_t pid;
    int fd, n;

    sprintf(path, "%salmusvcu.pid", conf_static.log_path);

    retry = false;
 pidfile_retry:
    if ((fd = open(path, O_CREAT | O_EXCL | O_WRONLY, 0644)) == -1) {        
        switch (errno) {
        case EEXIST:
            break;
        default:
            elog(LOGERROR, "Could not create pidfile \"%s\": %s",
                 path, strerror(errno));
            exit(EXIT_FAILURE);
        }
        if ((stream = fopen(path, "rt")) == NULL) {
            elog(LOGERROR, "Could not open pidfile \"%s\" for reading: %s",
                 path, strerror(errno));
            exit(EXIT_FAILURE);
        }
        s[199] = '\0';
        if (fgets(s, 199, stream) == NULL) {
            fclose(stream);
            if (retry) {
                if (remove(path) != 0) {
                    elog(LOGERROR, "Could not remove pidfile \"%s\": %s",
                         path, strerror(errno));
                    exit(EXIT_FAILURE);
                }
                goto pidfile_retry;
            }
            sleep(1);
            retry = true;
            goto pidfile_retry;
        }
        fclose(stream);
        pid = atoi(s);
        /* this is the way to check if the pid exists */
        errno = 0;
        if (pid != getpid()) {
            if (kill(pid, 0) == 0 || errno == EPERM) {
                /* process is running (or this process is not allowed to send a
                   signal, which is highly unlikely) */

                /* if we have a proc file system, make an extra check */
                sprintf(s, "/proc/%d/cmdline", (int)pid);
                if ((stream = fopen(s, "rt")) != NULL) {
                    s[0] = '\0';
                    fgets(s, 199, stream);
                    fclose(stream);
                    if (strstr(s, "almusvcu") != NULL) {
                        elog(LOGINFO, "AlmusVCU is already running, aborting.");
                        exit(EXIT_SUCCESS);
                    }
                } else {
                    elog(LOGINFO, "AlmusVCU is already running, aborting.");
                    exit(EXIT_SUCCESS);
                }
            } else if (errno != ESRCH) {
                elog(LOGERROR, "Kill failed: %s.", strerror(errno));
                exit(EXIT_FAILURE);
            }
        }
        if ((fd = open(path, O_CREAT | O_WRONLY | O_TRUNC, 0644)) == -1) {
            elog(LOGERROR, "Could not create pidfile \"%s\": %s",
                 path, strerror(errno));
            exit(EXIT_FAILURE);
        }
    }
    if ((stream = fdopen(fd, "wt")) == NULL) {
        elog(LOGERROR, "fdopen failed: %s", strerror(errno));
        exit(EXIT_FAILURE);
    }
    sprintf(s, "%d", (int)getpid());
    n = fprintf(stream, s);
    if (n != strlen(s)) {
        elog(LOGERROR, "Failed to write PID to file.");
        exit(EXIT_FAILURE);
    }    
    fclose(stream);
}
