/*
 * (c) Copyright 2002 - 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <math.h>
#include <inttypes.h>
#include <malloc.h>
#include <string.h>

#include "alignalloc.h"
#include "defs.h"
#include "log2.h"
#include "elog.h"
#include "tausrand.h"
#include "convolve.h"
#include "fftplan.h"
#include "almusvcu.h"
#include "decorr.h"

static uint32_t tausrand_state[3];
static bool_t seed_is_init = false;

static inline double
cosine_int(double mag1,
           double mag2,
           double freq1,
           double freq2,
           double curfreq)
{
    return (mag1 - mag2) * 0.5 *
        cos(M_PI * (curfreq - freq1) / (freq2 - freq1)) +
        (mag1 + mag2) * 0.5;
}

static double
random_val(double amplitude)
{
    double r;

    r = (double)(tausrand(tausrand_state)) / (double)4294967295.0;
    return 2.0 * amplitude * (r - 0.5);
}

static double
magwin(void *f,
       int len,
       double lo_freq,
       double hi_freq,
       double rate,
       int realsize)
{
    double m, curfreq;
    int start, n;
    
    start = (int)(lo_freq * (double)len / rate);
    if (start == 0) {
        start = 1;
    }
    m = 0.0;
    for (n = start; n < len / 2; n++) {
        curfreq = (double)n * rate / (double)len;
        if (realsize == 4) {
            m += sqrt(((float *)f)[n] * ((float *)f)[n] +
                      ((float *)f)[len-n] * ((float *)f)[len-n]);
        } else {
            m += sqrt(((double *)f)[n] * ((double *)f)[n] +
                      ((double *)f)[len-n] * ((double *)f)[len-n]);
        }
        if (curfreq >= hi_freq) {
            break;
        }
    }
    m /= (double)(n - start);
    return 20 * log10(m);
}

static int
mag_randomisation(void *s,
                  int len,
                  double rate,
                  double mag[],
                  double freq[],
                  double diff_mag,
                  int realsize)
{
    double mstart, mend, mline;
    double offset[100];
    void *fftplan, *f;
    int n, top;

    if (diff_mag <= 0.0) {
        mag[0] = 1.0;
        mag[1] = 1.0;
        freq[0] = 0.0;
        freq[1] = rate / 2.0;
        return 0;
    }
    
    if ((f = alignalloc(len * realsize)) == NULL) {
        elog(LOGDEBUG, "memory allocation failed");
        return VCU_ERR_NO_MEMORY;
    }
    if (realsize == 4) {
        fftplan = fftwf_plan_r2r_1d(len, s, f, FFTW_R2HC, FFTW_ESTIMATE);
        if (fftplan == NULL) {
            free(f);
            elog(LOGDEBUG, "failed to create plan");
            return VCU_ERR_NO_MEMORY;
        }
        fftwf_execute_r2r(fftplan, s, f);
        fftwf_destroy_plan(fftplan);
    } else {
        fftplan = fftw_plan_r2r_1d(len, s, f, FFTW_R2HC, FFTW_ESTIMATE);
        if (fftplan == NULL) {
            free(f);
            elog(LOGDEBUG, "failed to create plan");
            return VCU_ERR_NO_MEMORY;
        }
        fftw_execute_r2r(fftplan, s, f);
        fftw_destroy_plan(fftplan);
    }

    mstart = magwin(f, len, 1800, 2250.0, rate, realsize);
    mend = magwin(f, len, 18000.0, 22600.0, rate, realsize);

    top = 0;
    mag[top++] = 0.0;
    mag[top++] = mstart;
    mag[top++] = magwin(f, len, 2250.0, 2800.0, rate, realsize);
    mag[top++] = magwin(f, len, 2800.0, 3600.0, rate, realsize);
    mag[top++] = magwin(f, len, 3600.0, 4500.0, rate, realsize);
    mag[top++] = magwin(f, len, 4500.0, 5700.0, rate, realsize);
    mag[top++] = magwin(f, len, 5700.0, 7200.0, rate, realsize);
    mag[top++] = magwin(f, len, 7200.0, 9000.0, rate, realsize);
    mag[top++] = magwin(f, len, 9000.0, 11300.0, rate, realsize);
    mag[top++] = magwin(f, len, 11300.0, 14300.0, rate, realsize);
    mag[top++] = magwin(f, len, 14300.0, 18000.0, rate, realsize);
    mag[top] = mend;
    freq[0] = 0.0;
    freq[1] = 2000.0;
    freq[2] = 2500.0;
    freq[3] = 3200.0;
    freq[4] = 4000.0;
    freq[5] = 5000.0;
    freq[6] = 6300.0;
    freq[7] = 8000.0;
    freq[8] = 10000.0;
    freq[9] = 12700.0;
    freq[10] = 16000.0;
    freq[top] = 20000.0;
    freq[top+1] = rate / 2.0;

    for (n = 2; n < top; n++) {
        mline = mstart + (freq[n] - freq[1]) / (freq[top] - freq[1])  *
            (mend - mstart);
        offset[n] = mag[n] - mline;
        if (offset[n] > diff_mag) {
            offset[n] = -0.8 * diff_mag;
        } else if (offset[n] < -diff_mag) {
            offset[n] = 0.8 * diff_mag;
        } else {
            offset[n] = -0.8 * offset[n];
        }
    }
    for (n = 2; n < top; n++) {
        mag[n] = offset[n] + random_val(diff_mag);
    }
    
    mag[0] = 1.0;
    mag[1] = 1.0;
    for (n = 2; n < top; n++) {
        mag[n] = pow(10, mag[n] / 20);
    }
    mag[top] = 1.0;
    mag[top+1] = 1.0;

    return 0;
}

static void *
design_decorr_filter(int len,
                     double rate,
                     double diff_phase,
                     bool_t reduce_diff,
                     double apply_mag[],
                     double apply_mag_freq[],
                     int realsize)
{
    double fftwscale, curfreq, mag, freqmul, amp_prev, amp_next, phase, r;
    int n, i, period, next, prev;
    void *ifftplan, *filter;
    
    if ((ifftplan = fftplan_get_real(log2_get(len), true, true, realsize))
        == NULL)
    {
        elog(LOGDEBUG, "could not get fftplan for len %d", len);
        return NULL;
    }
    if ((filter = alignalloc(len * realsize)) == NULL) {
        elog(LOGDEBUG, "memory allocation failure", len);
        return NULL;
    }
    
    fftwscale = 1.0 / (double)len;
    freqmul = rate / (double)len;
    mag = fftwscale;
    i = 0;

    amp_prev = 0;
    amp_next = random_val(diff_phase);
    period = (int)(500.0 / rate * (double)len);
    prev = 0;
    next = period;

    if (realsize == 4) {
        ((float *)filter)[0] = 0.0;
    } else {
        ((double *)filter)[0] = 0.0;
    }
    for (n = 1; n < len / 2; n++) {
        
        curfreq = (double)n * freqmul;

        if (n == next) {
            amp_prev = amp_next;
            amp_next = random_val(diff_phase);
            prev = next;
            next += period;
            if (next >= len / 2 - 1) {
                amp_next = 0.0;
                next = len;
            }
        }
        r = cosine_int(amp_prev, amp_next, prev, next, n);
        if (reduce_diff) {
            if (curfreq < 2000) {
                /* full phase randomisation below 2kHz */
            } else if (curfreq < 4000) {
                r *= 0.5;
            } else if (curfreq < 8000) {
                r *= 0.25;            
            } else if (curfreq < 16000) {
                r *= 0.125;            
            } else {
                r *= 0.06125;
            }
        }
        phase = -M_PI * (double)n + r;

        if (apply_mag != NULL) {
            if (curfreq > apply_mag_freq[i+1]) {
                i++;
            }
            mag = cosine_int(apply_mag[i], apply_mag[i+1], apply_mag_freq[i],
                             apply_mag_freq[i+1], curfreq);
            mag *= fftwscale;
        }

        if (realsize == 4) {
            ((float *)filter)[n] = mag * cos(phase);
            ((float *)filter)[len - n] = mag * sin(phase);
        } else {
            ((double *)filter)[n] = mag * cos(phase);
            ((double *)filter)[len - n] = mag * sin(phase);
        }
        
    }
    if (realsize == 4) {
        ((float *)filter)[len / 2] = 0.0;
        fftwf_execute_r2r(ifftplan, filter, filter);
    } else {
        ((double *)filter)[len / 2] = 0.0;
        fftw_execute_r2r(ifftplan, filter, filter);
    }
    
    return filter;
}

void
decorr_set_random_seed(uint32_t seed)
{
    seed_is_init = true;
    tausinit(tausrand_state, seed);
}

void
decorr_get_random_state(uint32_t state[3])
{
    memcpy(state, tausrand_state, 12);
}

void
decorr_set_random_state(const uint32_t state[3])
{
    memcpy(tausrand_state, state, 12);
}

void *
decorr_reverb(void *imp,
              int len,
              double rate,
              double mag_diff,
              bool_t do_inplace,
              int realsize)
{
    void *f0, *f1, *f2, *f3, *s0, *s1, *s2, *s3, *s;
    double rd, mag[100], freq[100];
    uint32_t state1[3], state2[3];
    int base, taps, n;
    float rf;

    if (imp == NULL || len < 1 || rate < 1) {
        elog(LOGDEBUG, "invalid parameter");
        return NULL;
    }
    if (((ptrdiff_t)imp & (ALIGNMENT - 1)) != 0) {
        elog(LOGDEBUG, "alignment error");
        return NULL;
    }
    if (!seed_is_init) {
        decorr_set_random_seed(0);
    }
    
    /* Length of the decorrelation filters, about 100 ms */
    if (rate <= 48000.0) {
        taps = 4096;
    } else if (rate <= 96000.0) {
        taps = 8192;
    } else {
        taps = 16384;
    }

    /*
     * The tausrand states are copied to make sure that the same magnitude
     * randomisation is employed in all four stages
     */
    
    /*
     * Direct sound decorrelation filter: small phase variation, even smaller in
     * high frequencies (so the impulse looks fine). The larger phase variation
     * in lower frequencies is to produce some variation more than to be
     * 'correct'. Magnitude variation is also applied.
     */
    memcpy(state1, tausrand_state, 12);
    mag_randomisation(imp,
                      len < (int)(0.0075 * rate) ? len : (int)(0.0075 * rate),
                      rate, mag, freq, mag_diff, realsize);
    f0 = design_decorr_filter(taps, rate, 0.25 * M_PI, true, mag, freq,
                              realsize);

    /*
     * Early reflections decorrelation filter: as above, but with a bit larger
     * phase variation.
     */
    memcpy(state2, tausrand_state, 12);
    memcpy(tausrand_state, state1, 12);
    mag_randomisation(imp,
                      len < (int)(0.050 * rate) ? len : (int)(0.050 * rate),
                      rate, mag, freq, mag_diff, realsize);
    memcpy(tausrand_state, state2, 12);
    f1 = design_decorr_filter(taps, rate, 2 * M_PI, true, mag, freq, realsize);

    /*
     * Early reverb tail decorrelation filter: large phase variation, evenly
     * distributed over all frequencies.
     */
    memcpy(state2, tausrand_state, 12);
    memcpy(tausrand_state, state1, 12);
    mag_randomisation(imp,
                      len < (int)(0.150 * rate) ? len : (int)(0.150 * rate),
                      rate, mag, freq, mag_diff, realsize);
    memcpy(tausrand_state, state2, 12);
    f2 = design_decorr_filter(taps, rate, 40, false, mag, freq, realsize);
    
    /*
     * Reverb tail decorrelation filter: complete phase variation, evenly
     * distributed over all frequencies.
     */
    memcpy(state2, tausrand_state, 12);
    memcpy(tausrand_state, state1, 12);
    mag_randomisation(imp,
                      len < (int)(1.0 * rate) ? len : (int)(1.0 * rate),
                      rate, mag, freq, mag_diff, realsize);
    memcpy(tausrand_state, state2, 12);
    f3 = design_decorr_filter(taps, rate, 2 * taps, false, mag, freq, realsize);

    /*
     * decorrelate each part of the impulse response with the proper filter
     */
    s0 = convolve(f0, taps, imp,
                  len < (int)(0.0075 * rate) ? len : (int)(0.0075 * rate),
                  realsize);
    s1 = convolve(f1, taps, imp,
                  len < (int)(0.050 * rate) ? len : (int)(0.050 * rate),
                  realsize);
    s2 = convolve(f2, taps, imp,
                  len < (int)(0.150 * rate) ? len : (int)(0.150 * rate),
                  realsize);
    s3 = convolve(f3, taps, imp, len, realsize);
    free(f0); free(f1); free(f2); free(f3);

    if (s0 == NULL || s1 == NULL || s2 == NULL || s3 == NULL) {
        free(s0); free(s1); free(s2); free(s3);
        return NULL;
    }

    /*
     * compose the new impulse response by summing the results of the
     * convolutions (with proper fading in/out in the joints)
     */
    if (do_inplace) {
        s = imp;
    } else {
        if ((s = alignalloc(len * realsize)) == NULL) {
            free(s0); free(s1); free(s2); free(s3);
            return NULL;
        }
    }
    base = taps / 2;
    if (realsize == 4) {
        for (n = 0; n < len; n++) {
            if (n < (int)(0.005 * rate)) {
                ((float *)s)[n] = ((float *)s0)[base+n];
            } else if (n < (int)(0.010 * rate)) {
                rf = (float)(n - (int)(0.005 * rate)) / (0.005 * rate);
                ((float *)s)[n] =
                    ((float *)s0)[base+n] * (1.0 - rf) +
                    ((float *)s1)[base+n] * rf;
            } else if (n < (int)(0.045 * rate)) {
                ((float *)s)[n] = ((float *)s1)[base+n];
            } else if (n < (int)(0.055 * rate)) {
                rf = (float)(n - (int)(0.045 * rate)) / (0.010 * rate);
                ((float *)s)[n] =
                    ((float *)s1)[base+n] * (1.0 - rf) +
                    ((float *)s2)[base+n] * rf;
            } else if (n < (int)(0.140 * rate)) {
                ((float *)s)[n] = ((float *)s2)[base+n];
            } else if (n < (int)(0.160 * rate)) {
                rf = (float)(n - (int)(0.140 * rate)) / (0.020 * rate);
                ((float *)s)[n] =
                    ((float *)s2)[base+n] * (1.0 - rf) +
                    ((float *)s3)[base+n] * rf;
            } else {
                ((float *)s)[n] = ((float *)s3)[base+n];
            }
        }
    } else {
        for (n = 0; n < len; n++) {
            if (n < (int)(0.005 * rate)) {
                ((double *)s)[n] = ((double *)s0)[base+n];
            } else if (n < (int)(0.010 * rate)) {
                rd = (double)(n - (int)(0.005 * rate)) / (0.005 * rate);
                ((double *)s)[n] =
                    ((double *)s0)[base+n] * (1.0 - rd) +
                    ((double *)s1)[base+n] * rd;
            } else if (n < (int)(0.045 * rate)) {
                ((double *)s)[n] = ((double *)s1)[base+n];
            } else if (n < (int)(0.055 * rate)) {
                rd = (double)(n - (int)(0.045 * rate)) / (0.010 * rate);
                ((double *)s)[n] =
                    ((double *)s1)[base+n] * (1.0 - rd) +
                    ((double *)s2)[base+n] * rd;
            } else if (n < (int)(0.140 * rate)) {
                ((double *)s)[n] = ((double *)s2)[base+n];
            } else if (n < (int)(0.160 * rate)) {
                rd = (double)(n - (int)(0.140 * rate)) / (0.020 * rate);
                ((double *)s)[n] =
                    ((double *)s2)[base+n] * (1.0 - rd) +
                    ((double *)s3)[base+n] * rd;
            } else {
                ((double *)s)[n] = ((double *)s3)[base+n];
            }
        }
    }
    free(s0); free(s1); free(s2); free(s3);

    return s;
}

#if 0
int
main(void)
{
    double *imp, *s, rate;
    char str[1000];
    int len, n;

    rate = 44100;

    len = 0;
    imp = alignalloc(200000 * sizeof(double));
    s = alignalloc(200000 * sizeof(double));
    while (fgets(str, 999, stdin) != NULL) {
        imp[len++] = atof(str);
    }

    decorr_set_random_seed(200030);

    s = decorr_reverb(imp, len, rate, 3.0, sizeof(double));
    
    for (n = 0; n < len; n++) {
        printf("%.16e\n", s[n]);
    }
    return 0;
}
#endif
