/*
 * (c) Copyright 2002 - 2006 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <string.h>
#include <math.h>

#include "alignalloc.h"
#include "defs.h"
#include "elog.h"
#include "log2.h"
#include "fftplan.h"
#include "tausrand.h"
#include "equaliser.h"
#include "convolve.h"
#include "filterprg.h"
#include "firwindow.h"

struct stackentry {
    int delay[VCU_VALID_RATES_COUNT];
    struct prepfilter filter;
};

bool_t equaliser_debug = false;

static char log_path[PATH_MAX];

static int
get_maxpos(const void *coeffs,
	   int length,
           int realsize)
{
    int n, maxpos = 0;
    float maxvalf = 0;
    double maxvald = 0;

    if (realsize == 4) {
        for (n = 0; n < length; n++) {
            if (maxvalf < fabsf(((const float *)coeffs)[n])) {
                maxvalf = fabsf(((const float *)coeffs)[n]);
                maxpos = n;
            }
        }
    } else {
        for (n = 0; n < length; n++) {
            if (maxvald < fabs(((const double *)coeffs)[n])) {
                maxvald = fabs(((const double *)coeffs)[n]);
                maxpos = n;
            }
        }
    }
    return maxpos;
}

static void
dbabs_on_halfcomplex(void *signal,
		     int len,
                     int realsize)
{
    int n;

    if (realsize == 4) {
        for (n = 1; n < len / 2; n++) {
            ((float *)signal)[n] = ((float *)signal)[len-n] =
                20 * log10(sqrt(((float *)signal)[n] *
                                ((float *)signal)[n] +
                                ((float *)signal)[len-n] *
                                ((float *)signal)[len-n]));
        }
        ((float *)signal)[0] = 20 * log10(fabsf(((float *)signal)[0]));
        if ((len & 0x1) == 0) {
            ((float *)signal)[len/2] =
                20 * log10(fabsf(((float *)signal)[len/2]));
        }
    } else {
        for (n = 1; n < len / 2; n++) {
            ((double *)signal)[n] = ((double *)signal)[len-n] =
                20 * log10(sqrt(((double *)signal)[n] *
                                ((double *)signal)[n] +
                                ((double *)signal)[len-n] *
                                ((double *)signal)[len-n]));
        }
        ((double *)signal)[0] = 20 * log10(fabs(((double *)signal)[0]));
        if ((len & 0x1) == 0) {
            ((double *)signal)[len/2] =
                20 * log10(fabs(((double *)signal)[len/2]));
        }
    }
}

static inline double
cosine_int(double mag1,
           double mag2,
           double freq1,
           double freq2,
           double curfreq)
{
    return (mag1 - mag2) * 0.5 *
        cos(M_PI * (curfreq - freq1) / (freq2 - freq1)) +
        (mag1 + mag2) * 0.5;
}

static void
multiply(struct prepfilter *filter,
         double f)
{
    float ff;
    int ri, n;

    if (filter->realsize == 4) {
        ff = (float)f;
        for (ri = 0; ri < filter->n_rates; ri++) {
            for (n = 0; n < filter->lengths[ri]; n++) {
                ((float *)filter->filters[0][ri])[n] *= ff;
            }
        }
    } else {
        for (ri = 0; ri < filter->n_rates; ri++) {
            for (n = 0; n < filter->lengths[ri]; n++) {
                ((double *)filter->filters[0][ri])[n] *= f;
            }
        }
    }
}

static void *
addsub(const void *in1,
       int len1,
       int delay1,
       const void *in2,
       int len2,
       int delay2,
       bool_t add,
       int realsize,
       int *newlen,
       int *newdelay)
{
    int offset, len, n;
    const void *in;
    void *out;
    
    *newdelay = (delay1 > delay2) ? delay1 : delay2;
    *newlen = (*newdelay) +
        ((len1 - delay1 > len2 - delay2) ? len1 - delay1 : len2 - delay2);

    if ((out = alignalloc((*newlen) * realsize)) == NULL) {
        return NULL;
    }
    memset(out, 0, (*newlen) * realsize);
    if (delay1 > delay2) {
        memcpy(out, in1, len1 * realsize);
        in = in2;
        len = len2;
        offset = (*newdelay) - delay2;
    } else {
        memcpy(out, in2, len2 * realsize);
        in = in1;
        len = len1;
        offset = (*newdelay) - delay1;
    }    
    if (add) {
        if (realsize == 4) {
            for (n = 0; n < len; n++) {
                ((float *)out)[offset + n] += ((const float *)in)[n];
            }
        } else {
            for (n = 0; n < len; n++) {
                ((double *)out)[offset + n] += ((const double *)in)[n];
            }
        }
    } else {
        if (realsize == 4) {
            for (n = 0; n < len; n++) {
                ((float *)out)[offset + n] -= ((const float *)in)[n];
            }
        } else {
            for (n = 0; n < len; n++) {
                ((double *)out)[offset + n] -= ((const double *)in)[n];
            }
        }
    }
    return out;
}

static int
combine_filters(struct prepfilter *filter,
		const struct prepfilter *filter1,
                const int delay1[],
		const struct prepfilter *filter2,
                const int delay2[],
                int operation,
                int outdelay[])
{
    int ri = 0, err;
    
    if (filter == NULL || filter1 == NULL || filter2 == NULL ||
	filter1->n_channels != 1 || filter2->n_channels != 1 ||
	filter1->n_rates != filter2->n_rates || outdelay == NULL ||
        filter1->realsize != filter2->realsize ||
	memcmp(filter1->rates, filter2->rates,
	       filter1->n_rates * sizeof(int)) != 0)
    {
	return VCU_ERR_INVALID_VALUE;
    }
    memset(filter, 0, sizeof(struct prepfilter));
    memcpy(filter->rates, filter1->rates, sizeof(filter1->rates));
    filter->n_rates = filter1->n_rates;
    filter->n_channels = 1;
    filter->realsize = filter1->realsize;
    for (ri = 0; ri < filter1->n_rates; ri++) {
        switch (operation) {
        case VCU_EQPRG_TYPE_OP_CONVOLVE:
            elog(LOGDEBUG, "convolve");
            if ((filter->filters[0][ri] =
                 convolve(filter1->filters[0][ri], filter1->lengths[ri],
                          filter2->filters[0][ri], filter2->lengths[ri],
                          filter->realsize))
                == NULL)
            {
                err = VCU_ERR_NO_MEMORY;
                goto combine_filters_failure;
            }
            filter->lengths[ri] = filter1->lengths[ri] +
                filter2->lengths[ri] - 1;
            outdelay[ri] = delay1[ri] + delay2[ri];
            break;
        case VCU_EQPRG_TYPE_OP_ADD:
        case VCU_EQPRG_TYPE_OP_SUBTRACT:
            elog(LOGDEBUG, "addsub");
            if ((filter->filters[0][ri] =
                 addsub(filter1->filters[0][ri], filter1->lengths[ri],
                        delay1[ri],
                        filter2->filters[0][ri], filter2->lengths[ri],
                        delay2[ri],
                        operation == VCU_EQPRG_TYPE_OP_ADD,
                        filter->realsize,
                        &filter->lengths[ri],
                        &outdelay[ri]))
                == NULL)
            {
                err = VCU_ERR_NO_MEMORY;
                goto combine_filters_failure;
            }
            break;
        default:
            err = VCU_ERR_INVALID_VALUE;
            goto combine_filters_failure;            
        }
        elog(LOGDEBUG, "done (ri: %d)", ri);
    }
    return 0;
 combine_filters_failure:
    for (ri = 0; ri < filter1->n_rates; ri++) {
	free(filter1->filters[0][ri]);
    }
    return err;
}

#define APPLY_KAISER(target, taps, realsize)                                   \
    firwindow_kaiser(target, taps, ((taps) & 1) == 0 ? -1.0 : 0.0, 9, realsize)

#define _FILTER_RESPONSE(normal, zero, type) {                                 \
    fpos = &((type *)filter)[(n_taps - 1) >> 1];                               \
    for (n = -((n_taps - 1) >> 1); n != 0; n++) {                              \
	((type *)fpos)[n] = normal;                                            \
    }                                                                          \
    ((type *)fpos)[0] = zero;                                                  \
    for (n = 1; n <= n_taps >> 1; n++) {                                       \
	((type *)fpos)[n] = normal;                                            \
    }                                                                          \
}

#define FILTER_RESPONSE(normal, zero, realsize) {                              \
    if (realsize == 4) {                                                       \
        _FILTER_RESPONSE(normal, zero, float);                                 \
    } else {                                                                   \
        _FILTER_RESPONSE(normal, zero, double);                                \
    }                                                                          \
} 

static void *
sinc_filter(int n_taps,
            double low_freq,
            double high_freq,
            int type,
            int realsize)
{
    void *filter, *fpos;
    int n;

    if ((filter = alignalloc(n_taps * realsize)) == NULL) {
	elog(LOGERROR, "memory allocation failed");
	return NULL;
    }
    
    low_freq *= M_PI;
    high_freq *= M_PI;
    
    switch (type) {

    case VCU_EQPRG_TYPE_LOWPASS:
	FILTER_RESPONSE(sin((double)n * low_freq) / ((double)n * M_PI),
			low_freq / M_PI, realsize);
	break;
	
    case VCU_EQPRG_TYPE_HIGHPASS:
	FILTER_RESPONSE(-sin((double)n * high_freq) / ((double)n * M_PI),
			1.0 - high_freq / M_PI, realsize);
	break;
	
    case VCU_EQPRG_TYPE_BANDPASS:
	FILTER_RESPONSE((sin((double)n * high_freq) -
			 sin((double)n * low_freq)) / ((double)n * M_PI),
			(high_freq - low_freq) / M_PI, realsize);
	break;
	
    case VCU_EQPRG_TYPE_BANDSTOP:
	FILTER_RESPONSE((sin((double)n * low_freq) -
			 sin((double)n * high_freq)) / ((double)n * M_PI),
			1.0 - (high_freq - low_freq) / M_PI, realsize);
	break;
	
    default:
	elog(LOGERROR, "invalid type (%d)", type);
	free(filter);
	return NULL;
    }    
    return filter;
}

#undef FILTER_RESPONSE

static int
sinc_taps(double slope,
          double freq)
{
#define TAPS_CACHE_SIZE 128
    static float taps_cache_slope[TAPS_CACHE_SIZE];
    static float taps_cache_freq[TAPS_CACHE_SIZE];
    static int taps_cache_taps[TAPS_CACHE_SIZE];
    static int taps_cache_used = 0;
    static int taps_cache_pos = 0;
    
    int n_taps, n, fftsize, old_taps, add_taps, old_add_taps, min_taps;
    float rollfreq, old_rollfreq, target_rollfreq;
    float *tdomain, *fdomain, *fpos, sincfreq;
    bool_t has_undershot, has_overshot;
    void *fftplan;

    elog(LOGDEBUG, "slope %f, freq %f", slope, freq);

    for (n = 0; n < taps_cache_used; n++) {
        if (slope == taps_cache_slope[n] && freq == taps_cache_freq[n]) {
            elog(LOGDEBUG, "return cached value %d", taps_cache_taps[n]);
            return taps_cache_taps[n];
        }
    }
    
    if (slope < 0.0) {
        slope = 0.0;
    }
    if (slope > 1.0) {
        slope = 1.0;
    }
    taps_cache_slope[taps_cache_pos] = slope;
    taps_cache_freq[taps_cache_pos] = freq;
    
    target_rollfreq = slope * freq;

    fftsize = 32768;
    fftplan = fftplan_get_real(log2_get(fftsize), false, false, sizeof(float));
    tdomain = alignalloc(fftsize * sizeof(float));
    fdomain = alignalloc(fftsize * sizeof(float));
    memset(tdomain, 0, fftsize * sizeof(float));

    old_taps = -1;
    old_add_taps = 0;
    add_taps = 0;
    old_rollfreq = 0.0;
    rollfreq = 0.0;
    n_taps = fftsize;
    sincfreq = freq * M_PI;
    min_taps = (int)(4.0 / freq);

    has_undershot = has_overshot = false;
    while (true) {

        memset(tdomain, 0, fftsize * sizeof(float));
        fpos = &tdomain[(n_taps - 1) >> 1];
        for (n = -((n_taps - 1) >> 1); n != 0; n++) {
            fpos[n] = sin((float)n * sincfreq) / ((float)n * M_PI);
        }
        fpos[0] = sincfreq / M_PI;
        for (n = 1; n <= n_taps >> 1; n++) {
            fpos[n] = sin((float)n * sincfreq) / ((float)n * M_PI);
        }
        APPLY_KAISER(tdomain, n_taps, sizeof(float));

        fftwf_execute_r2r(fftplan, tdomain, fdomain);

        old_rollfreq = rollfreq;
        rollfreq = 0.0;
        for (n = 1; n < (fftsize >> 1); n++) {
            if (sqrt(fdomain[n] * fdomain[n] +
                     fdomain[fftsize - n] * fdomain[fftsize - n]) < 0.97)
            {
                rollfreq = 2.0 * (float)n / (float)fftsize;
                break;
            }
        }
        elog(LOGDEBUG, "rolloff frequency: %f (pos %d) with %d taps "
             "(target %f)", rollfreq, n, n_taps, target_rollfreq);
        
        if (add_taps == 1) {
            break;
        }
        
        old_taps = n_taps;
        if (rollfreq < target_rollfreq) {
            has_undershot = true;
            if (!has_overshot) {
                add_taps = n_taps;
            } else {
                add_taps /= 2;
            }
            if (add_taps == 0) {
                add_taps = 1;
            }
            if (n_taps == fftsize) {
                break;
            }
            n_taps += add_taps;
        } else {
            has_overshot = true;
            if (!has_undershot) {
                add_taps = n_taps / 2;
            } else {
                add_taps /= 2;
            }
            if (add_taps == 0) {
                add_taps = 1;
            }
            n_taps -= add_taps;
        }
        
    }

    if (fabsf(old_rollfreq - target_rollfreq) <
        fabsf(rollfreq - target_rollfreq))
    {
        n_taps = old_taps;
        rollfreq = old_rollfreq;
    }
    
    free(tdomain);
    free(fdomain);

    if (n_taps < min_taps) {
        elog(LOGDEBUG, "taps was %d, too low for the given cutoff, setting it "
             "to %d", n_taps, min_taps);
        n_taps = min_taps;
    }

    taps_cache_taps[taps_cache_pos] = n_taps;
    if (taps_cache_used < TAPS_CACHE_SIZE) {
        taps_cache_used++;
    }
    taps_cache_pos++;
    if (taps_cache_pos == TAPS_CACHE_SIZE) {
        taps_cache_pos = 0;
    }
    
    elog(LOGDEBUG, "result: taps %d rollfreq %f", n_taps, rollfreq);
    return n_taps;
}

static void *
design_filter_with_sinc(int *n_taps,
                        double sample_freq,
                        double low_freq,
                        double high_freq,
                        double slope,
                        int phase_type,
                        int type,
                        int realsize)
{
    void *filter = NULL, *filterlo, *filterhi;
    int tapslo, tapshi, n;
    
    if (phase_type != VCU_PHASE_LINEAR) {
        elog(LOGERROR, "only phase type linear supported");
        return NULL;
    }
    low_freq /= sample_freq / 2;
    high_freq /= sample_freq / 2;

    switch (type) {
    case VCU_EQPRG_TYPE_LOWPASS:
        *n_taps = sinc_taps(slope, low_freq);
        filter = sinc_filter(*n_taps, low_freq, 0.0, type, realsize);
        APPLY_KAISER(filter, *n_taps, realsize);
        break;
    case VCU_EQPRG_TYPE_HIGHPASS:
        *n_taps = sinc_taps(slope, high_freq);
        filter = sinc_filter(*n_taps, 0.0, high_freq, type, realsize);
        APPLY_KAISER(filter, *n_taps, realsize);
        break;
    case VCU_EQPRG_TYPE_BANDPASS:
        /* instead of doing bandpass directly, we convolve together a lowpass
           and a highpass filter */
        tapslo = sinc_taps(slope, high_freq);
        tapshi = sinc_taps(slope, low_freq);
        if (tapshi > 128 && tapshi >= 8 * tapslo) {
            tapshi = tapshi - tapslo + 1;
        }
        filterlo = sinc_filter(tapslo, high_freq, 0.0, VCU_EQPRG_TYPE_LOWPASS,
                               realsize);
        APPLY_KAISER(filterlo, tapslo, realsize);
        filterhi = sinc_filter(tapshi, 0.0, low_freq, VCU_EQPRG_TYPE_HIGHPASS,
                               realsize);
        APPLY_KAISER(filterhi, tapshi, realsize);
        filter = convolve(filterlo, tapslo, filterhi, tapshi, realsize);
        *n_taps = tapshi + tapslo - 1;
        free(filterlo);
        free(filterhi);
        break;
    case VCU_EQPRG_TYPE_BANDSTOP:
        /* instead of doing bandpass directly, we add together a highpass and
           a lowpass filter */
        tapslo = sinc_taps(slope, low_freq);
        tapshi = sinc_taps(slope, high_freq);
        filterlo = sinc_filter(tapslo, low_freq, 0.0, VCU_EQPRG_TYPE_LOWPASS,
                               realsize);
        APPLY_KAISER(filterlo, tapslo, realsize);
        filterhi = sinc_filter(tapshi, 0.0, high_freq, VCU_EQPRG_TYPE_HIGHPASS,
                               realsize);
        APPLY_KAISER(filterhi, tapshi, realsize);
        if (filterhi == NULL || filterlo == NULL) {
            free(filterhi);
            free(filterlo);
            return NULL;
        }
        if (tapslo > tapshi) {
            *n_taps = tapslo;
            filter = filterlo;
            if (realsize == 4) {
                for (n = 0; n < tapshi; n++) {
                    ((float *)filter)[tapslo / 2 - tapshi / 2 + n] +=
                        ((float *)filterhi)[n];
                }
            } else {
                for (n = 0; n < tapshi; n++) {
                    ((double *)filter)[tapslo / 2 - tapshi / 2 + n] +=
                        ((double *)filterhi)[n];
                }
            }
            free(filterhi);
        } else {
            *n_taps = tapshi;
            filter = filterhi;
            if (realsize == 4) {
                for (n = 0; n < tapslo; n++) {
                    ((float *)filter)[tapshi / 2 - tapslo / 2 + n] +=
                        ((float *)filterlo)[n];
                }
            } else {
                for (n = 0; n < tapslo; n++) {
                    ((double *)filter)[tapshi / 2 - tapslo / 2 + n] +=
                        ((double *)filterlo)[n];
                }
            }
            free(filterlo);
        }
        break;
    }

    return filter;
}

static void *
make_impulse(int *n_taps,
             int realsize)
{
    void *filter;
    *n_taps = 1;
    if ((filter = alignalloc(realsize)) != NULL) {
        if (realsize == 4) {
            *(float *)filter = 1.0;
        } else {
            *(double *)filter = 1.0;
        }
    }
    return filter;
}

static int
fft_filter_length(int block_size,
                  int sample_rate,
                  int n_bands,
                  const double bands[])
{
    int n, minlength, baselength;
    double minband;


    /* we need at least one wave-length of the lowest band */
    minlength = 1 << log2_roof((int)(((double)sample_rate / bands[0]) + 1));
    elog(LOGDEBUG, "minimal length due to lowest band %f Hz: %d",
         bands[0], minlength);
    if (minlength < block_size) {
        minlength = block_size;
        elog(LOGDEBUG, "minimal length due to block_size: %d", minlength);
    }

    minband = bands[n_bands-1] - bands[0];
    for (n = 1; n < n_bands; n++) {
        if (bands[n] - bands[n-1] < minband) {
            minband = bands[n] - bands[n-1];
        }
    }
    elog(LOGDEBUG, "minband: %f Hz", minband);

    /* for 44.1 - 48 kHz */
    if (minband < 15) {
        baselength = 32768;
    } else if (minband < 25) {
        baselength = 16384;
    } else if (minband < 50) {
        baselength = 8192;
    } else if (minband < 100) {
        baselength = 4096;
    } else if (minband < 500) {
        baselength = 2048;
    } else {
        baselength = 1024;
    }
    /* adapt for sample rate */
    baselength *= (sample_rate / 48001 + 1);
    elog(LOGDEBUG, "minimal length bandwidth: %d", baselength);

    return minlength < baselength ? baselength : minlength;
}

#define DESIGN_FILTER_PROTOTYPE(function_call)                                 \
    int n;                                                                     \
    memset(filter, 0, sizeof(struct prepfilter));                              \
    filter->n_rates = n_rates;                                                 \
    filter->realsize = realsize;                                               \
    for (n = 0; n < n_rates; n++) {                                            \
        filter->rates[n] = rates[n];                                           \
        filter->n_channels = 1;                                                \
        filter->filters[0][n] = function_call;                                 \
        if (filter->filters[0][n] == NULL) {                                   \
            for (n--; n >= 0; n--) {                                           \
                free(filter->filters[0][n]);                                   \
            }                                                                  \
            elog(LOGDEBUG, "failed to design filter");                         \
            return false;                                                      \
        }                                                                      \
    }                                                                          \
    return true;

static bool_t
design_sincfilter(struct prepfilter *filter,
                  double low_freq,
                  double high_freq,
                  const int rates[],
                  int n_rates,
                  double slope,
                  int phase_type,
                  int type,
                  int realsize)
{
    DESIGN_FILTER_PROTOTYPE(
        design_filter_with_sinc(&filter->lengths[n],
                                rates[n],
                                low_freq, high_freq,
                                slope, phase_type, type, realsize)
        );
}

static bool_t
design_noisefilter(struct prepfilter *filter,
                   const int rates[],
                   int n_rates,
                   double length,
                   int realsize)
{
    DESIGN_FILTER_PROTOTYPE(
        eq_design_whitenoise_filter(&filter->lengths[n],
                                    rates[n],
                                    length,
                                    realsize)
        );
}

static bool_t
design_allpass(struct prepfilter *filter,
               const int rates[],
               int n_rates,
               int realsize)
{
    DESIGN_FILTER_PROTOTYPE(make_impulse(&filter->lengths[n], realsize));
}

static bool_t
design_fft(struct prepfilter *filter,
           const int block_sizes[],
           const int rates[],
           int n_rates,
           int n_bands,
           const double bands[],
           const double amps[],
           int realsize)
{
    DESIGN_FILTER_PROTOTYPE(
        eq_design_fft_filter(&filter->lengths[n],
                             rates[n],
                             block_sizes[n],
                             n_bands,
                             bands,
                             amps,
                             realsize)
        );
}

int
eq_sanitycheck(const struct vcu_eqpart eqpart[],
	       int n_parts,
	       const idarray_t *custom_prgs)
{
    struct filterprg *fprg;
    int n, i, pos;

    if (eqpart == NULL || n_parts < 1 ||
	n_parts >= VCU_EQ_COMPONENTS_MAX)
    {
        elog(LOGDEBUG, "invalid input: eqpart: %p; n_parts: %d",
             eqpart, n_parts);
	return VCU_ERR_INVALID_VALUE;
    }
    pos = 0;
    for (n = 0; n < n_parts; n++) {
	switch (eqpart[n].type) {
        case VCU_EQPRG_TYPE_OP_ADD:
        case VCU_EQPRG_TYPE_OP_SUBTRACT:
        case VCU_EQPRG_TYPE_OP_CONVOLVE:
            if (pos < 2) {
                elog(LOGDEBUG, "malplaced operation");
                return VCU_ERR_INVALID_VALUE;
            }
            pos--;
            continue;
        case VCU_EQPRG_TYPE_OP_MULTIPLY:
            if (pos < 1) {
                elog(LOGDEBUG, "malplaced operation");
                return VCU_ERR_INVALID_VALUE;
            }
            continue;
	case VCU_EQPRG_TYPE_LOWPASS:
            if (eqpart[n].params.lowpass.cutoff < 5.0 ||
                eqpart[n].params.lowpass.cutoff > 192000.0)
            {
                elog(LOGDEBUG, "invalid cutoff %f",
                     eqpart[n].params.highpass.cutoff);
                return VCU_ERR_INVALID_VALUE;
            }
            if (eqpart[n].params.lowpass.phase != VCU_PHASE_LINEAR) {
                elog(LOGDEBUG, "invalid phase %d",
                     eqpart[n].params.lowpass.phase);
                return VCU_ERR_INVALID_VALUE;
            }
            break;
	case VCU_EQPRG_TYPE_HIGHPASS:
            if (eqpart[n].params.highpass.cutoff < 5.0 ||
                eqpart[n].params.highpass.cutoff > 192000.0)
            {
                elog(LOGDEBUG, "invalid cutoff %f",
                     eqpart[n].params.highpass.cutoff);
                return VCU_ERR_INVALID_VALUE;
            }
            if (eqpart[n].params.highpass.phase != VCU_PHASE_LINEAR) {
                elog(LOGDEBUG, "invalid phase %d",
                     eqpart[n].params.highpass.phase);
                return VCU_ERR_INVALID_VALUE;
            }
            break;
	case VCU_EQPRG_TYPE_BANDPASS:
            if (eqpart[n].params.bandpass.cutoff[0] < 5.0 ||
                eqpart[n].params.bandpass.cutoff[0] > 192000.0 ||
                eqpart[n].params.bandpass.cutoff[1] < 5.0 ||
                eqpart[n].params.bandpass.cutoff[1] > 192000.0 ||
                eqpart[n].params.bandpass.cutoff[0] >=
                eqpart[n].params.bandpass.cutoff[1])
            {
                elog(LOGDEBUG, "invalid cutoff %f - %f",
                     eqpart[n].params.bandpass.cutoff[0],
                     eqpart[n].params.bandpass.cutoff[1]);
                return VCU_ERR_INVALID_VALUE;
            }
            if (eqpart[n].params.bandpass.phase != VCU_PHASE_LINEAR) {
                elog(LOGDEBUG, "invalid phase %d",
                     eqpart[n].params.bandpass.phase);
                return VCU_ERR_INVALID_VALUE;
            }
            break;
	case VCU_EQPRG_TYPE_BANDSTOP:
            if (eqpart[n].params.bandstop.cutoff[0] < 5.0 ||
                eqpart[n].params.bandstop.cutoff[0] > 192000.0 ||
                eqpart[n].params.bandstop.cutoff[1] < 5.0 ||
                eqpart[n].params.bandstop.cutoff[1] > 192000.0 ||
                eqpart[n].params.bandstop.cutoff[0] >=
                eqpart[n].params.bandstop.cutoff[1])
            {
                elog(LOGDEBUG, "invalid cutoff %f - %f",
                     eqpart[n].params.bandstop.cutoff[0],
                     eqpart[n].params.bandstop.cutoff[1]);
                return VCU_ERR_INVALID_VALUE;
            }
            if (eqpart[n].params.bandstop.phase != VCU_PHASE_LINEAR) {
                elog(LOGDEBUG, "invalid phase %d",
                     eqpart[n].params.bandstop.phase);
                return VCU_ERR_INVALID_VALUE;
            }
            break;
        case VCU_EQPRG_TYPE_ALLPASS:
            break;
	case VCU_EQPRG_TYPE_CUSTOM:
	    i = -1;
	    while ((fprg = idarray_next(custom_prgs, &i)) != NULL) {
		if (memcmp(&fprg->id, &eqpart[n].params.custom.id,
			   sizeof(prgid_t)) == 0)
		{
		    break;
		}
	    }
	    if (fprg == NULL) {
                elog(LOGDEBUG, "could not find program with id %s",
                     filterprg_prgid_ntoa(eqpart[n].params.custom.id));
		return VCU_ERR_INVALID_VALUE;
	    }
	    break;
	case VCU_EQPRG_TYPE_RANDOM_PHASE:
            if (eqpart[n].params.random_phase.length < 1 / 1000.0 ||
                eqpart[n].params.random_phase.length > 10.0)
            {
                elog(LOGDEBUG, "invalid length: %f",
                     eqpart[n].params.random_phase.length);
		return VCU_ERR_INVALID_VALUE;
            }
            break;
        case VCU_EQPRG_TYPE_FFT_DYNAMIC:
        case VCU_EQPRG_TYPE_FFT_STATIC:
            if (eqpart[n].type == VCU_EQPRG_TYPE_FFT_DYNAMIC && n_parts != 1) {
                elog(LOGDEBUG, "dynamic eq must be the single component");
                return VCU_ERR_INVALID_VALUE;
            }
            if (eqpart[n].params.fft.n_bands < 2 ||
                eqpart[n].params.fft.n_bands > VCU_EQPRG_MAX_BANDS)
            {
                elog(LOGDEBUG, "invalid amount of bands");
                return VCU_ERR_INVALID_VALUE;
            }
            for (i = 0; i < eqpart[n].params.fft.n_bands; i++) {
                if (eqpart[n].params.fft.band[i] <
                    VCU_EQPRG_MIN_BAND_FREQUENCY ||
                    eqpart[n].params.fft.band[i] >
                    VCU_EQPRG_MAX_BAND_FREQUENCY ||
                    (i > 0 && eqpart[n].params.fft.band[i] -
                     VCU_FFTEQ_SMALLEST_BAND <
                     eqpart[n].params.fft.band[i-1]))
                {
                    elog(LOGDEBUG, "invalid band");
                    return VCU_ERR_INVALID_VALUE;
                }
                if (eqpart[n].params.fft.amp[i] < VCU_FFTEQ_VOLUME_MIN ||
                    eqpart[n].params.fft.amp[i] > VCU_FFTEQ_VOLUME_MAX)
                {
                    elog(LOGDEBUG, "invalid amplification");
                    return VCU_ERR_INVALID_VALUE;
                }
            }
            break;
	default:
            elog(LOGDEBUG, "invalid type");
	    return VCU_ERR_INVALID_VALUE;
	}
        pos++;
    }
    if (pos != 1) {
        elog(LOGDEBUG, "not 1 element in stack");
        return VCU_ERR_INVALID_VALUE;
    }
    return 0;
}

int
eq_generate(const struct vcu_eqpart eqpart[],
	    int n_parts,
	    const idarray_t *custom_prgs, /* may be NULL */
            const int block_sizes[],
	    const int rates[],
	    int n_rates,
            int realsize,
	    int delay[],
	    struct prepfilter *filter)
{
    int n, i, err, pos, specified_delay[VCU_VALID_RATES_COUNT];
    struct stackentry stack[VCU_EQ_COMPONENTS_MAX];
    struct stackentry stackentry;
    struct filterprg *fprg;
    double f;
    
    if ((err = eq_sanitycheck(eqpart, n_parts, custom_prgs)) < 0) {
        elog(LOGDEBUG, "at least one insane equaliser");
	return err;
    }

    pos = 0;
    memset(stack, 0, sizeof(stack));
    for (n = 0; n < n_parts; n++) {
        specified_delay[0] = -1;
	switch (eqpart[n].type) {
        case VCU_EQPRG_TYPE_OP_ADD:
        case VCU_EQPRG_TYPE_OP_SUBTRACT:
        case VCU_EQPRG_TYPE_OP_CONVOLVE:
            if (pos < 2) {
                elog(LOGDEBUG, "operation without enough operands");
                err = VCU_ERR_INVALID_VALUE;
                goto generate_failure;
            }
	    err = combine_filters(&stackentry.filter,
                                  &stack[pos-1].filter,
                                  stack[pos-1].delay,
                                  &stack[pos-2].filter, 
                                  stack[pos-2].delay,
                                  eqpart[n].type,
                                  stackentry.delay);
	    if (err != 0) {
		goto generate_failure;
	    }
	    filterprg_free_prepfilter_contents(&stack[pos-1].filter);
	    filterprg_free_prepfilter_contents(&stack[pos-2].filter);
            stack[pos-2] = stackentry;
            pos--;
            continue;
        case VCU_EQPRG_TYPE_OP_MULTIPLY:
            if (pos < 1) {
                elog(LOGDEBUG, "operation without enough operands");
                err = VCU_ERR_INVALID_VALUE;
                goto generate_failure;
            }
            f = pow(10, eqpart[n].params.multiply.amp / 20);
            if (eqpart[n].params.multiply.sign < 0) {
                f *= -1;
            }
            multiply(&stack[pos-1].filter, f);
            continue;
	case VCU_EQPRG_TYPE_LOWPASS:
            if (!design_sincfilter(&stack[pos++].filter,
                                   eqpart[n].params.lowpass.cutoff, 0.0,
                                   rates, n_rates,
                                   eqpart[n].params.lowpass.slope,
                                   eqpart[n].params.lowpass.phase,
                                   VCU_EQPRG_TYPE_LOWPASS,
                                   realsize))
            {
		err = VCU_ERR_INVALID_VALUE;
		goto generate_failure;
            }
            break;
	case VCU_EQPRG_TYPE_HIGHPASS:
            if (!design_sincfilter(&stack[pos++].filter,
                                   0.0, eqpart[n].params.highpass.cutoff,
                                   rates, n_rates,
                                   eqpart[n].params.highpass.slope,
                                   eqpart[n].params.highpass.phase,
                                   VCU_EQPRG_TYPE_HIGHPASS,
                                   realsize))
            {
		err = VCU_ERR_INVALID_VALUE;
		goto generate_failure;
            }
            break;
	case VCU_EQPRG_TYPE_BANDPASS:
            if (!design_sincfilter(&stack[pos++].filter,
                                   eqpart[n].params.bandpass.cutoff[0],
                                   eqpart[n].params.bandpass.cutoff[1],
                                   rates, n_rates,
                                   eqpart[n].params.bandpass.slope,
                                   eqpart[n].params.bandpass.phase,
                                   VCU_EQPRG_TYPE_BANDPASS,
                                   realsize))
            {
		err = VCU_ERR_INVALID_VALUE;
		goto generate_failure;
            }
            break;
	case VCU_EQPRG_TYPE_BANDSTOP:
            if (!design_sincfilter(&stack[pos++].filter,
                                   eqpart[n].params.bandstop.cutoff[0],
                                   eqpart[n].params.bandstop.cutoff[1],
                                   rates, n_rates,
                                   eqpart[n].params.bandstop.slope,
                                   eqpart[n].params.bandstop.phase,
                                   VCU_EQPRG_TYPE_BANDSTOP,
                                   realsize))
            {
		err = VCU_ERR_INVALID_VALUE;
		goto generate_failure;
            }
            break;
        case VCU_EQPRG_TYPE_ALLPASS:
            if (!design_allpass(&stack[pos++].filter, rates, n_rates,
                                realsize))
            {
		err = VCU_ERR_INVALID_VALUE;
		goto generate_failure;
            }
            break;
	case VCU_EQPRG_TYPE_CUSTOM:
	    if (custom_prgs == NULL) {
		err = VCU_ERR_INVALID_VALUE;
		goto generate_failure;
	    }	    
	    i = -1;
	    while ((fprg = idarray_next(custom_prgs, &i)) != NULL) {
		if (memcmp(&fprg->id, &eqpart[n].params.custom.id,
			   sizeof(prgid_t)) == 0)
		{
		    break;
		}
	    }
	    if (fprg == NULL) {
		err = VCU_ERR_INVALID_VALUE;
                goto generate_failure;
	    }
            if (fprg->predelay >= 0.0) {
                for (i = 0; i < n_rates; i++) {
                    specified_delay[i] = fprg->predelay * rates[i];
                }
            }
	    if ((err = filterprg_load(fprg, &stack[pos++].filter, NULL,
                                      NULL, 0, fprg->n_channels,
				      rates, n_rates, realsize,
                                      NULL, -1, -1)) != 0)
	    {
                goto generate_failure;
	    }
	    break;
	case VCU_EQPRG_TYPE_RANDOM_PHASE:
            if (!design_noisefilter(&stack[pos++].filter,
                                    rates, n_rates,
                                    eqpart[n].params.random_phase.length,
                                    realsize))
            {
		err = VCU_ERR_INVALID_VALUE;
		goto generate_failure;
            }
            memset(specified_delay, 0, sizeof(specified_delay));
            break;
        case VCU_EQPRG_TYPE_FFT_DYNAMIC:
        case VCU_EQPRG_TYPE_FFT_STATIC:
            if (!design_fft(&stack[pos++].filter,
                            block_sizes,
                            rates, n_rates,
                            eqpart[n].params.fft.n_bands,
                            eqpart[n].params.fft.band,
                            eqpart[n].params.fft.amp,
                            realsize))
            {
		err = VCU_ERR_INVALID_VALUE;
		goto generate_failure;
            }
            break;
	default:
            elog(LOGDEBUG, "unknown type %d", eqpart[n].type);
	    err = VCU_ERR_INVALID_VALUE;
	    goto generate_failure;
	}
        for (i = 0; i < n_rates; i++) {
            if (specified_delay[0] < 0) {
                stack[pos-1].delay[i] =
                    get_maxpos(stack[pos-1].filter.filters[0][i],
                               stack[pos-1].filter.lengths[i],
                               realsize);
            } else {
                /* predelay will be truncated in the conversion */
                stack[pos-1].delay[i] = specified_delay[i];
            }
        }
    }

    if (pos != 1) {
        err = VCU_ERR_INVALID_VALUE;
        goto generate_failure;
    }
    *filter = stack[0].filter;
    if (delay != NULL) {
        for (n = 0; n < n_rates; n++) {
            delay[n] = stack[0].delay[n];
            elog(LOGDEBUG, "Delay for %d Hz: %d samples", rates[n], delay[n]);
        }
    }
    return 0;
 generate_failure:
    for (n = 0; n < pos; n++) {
        filterprg_free_prepfilter_contents(&stack[n].filter);
    }
    return err;
}

#define _BAND_AMP(band, _thru, type)                                           \
    if (amp[band].in == 100000) {                                              \
        amp[band].in = ((type *)fdomain)[n];                                   \
    }                                                                          \
    amp[band].out = ((type *)fdomain)[n];                                      \
    if (fabs(_thru - f) < thru[band]) {                                        \
        thru[band] = fabs(_thru - f);                                          \
        amp[band].thru = ((type *)fdomain)[n];                                 \
    }                                                                          \
    if (amp[band].min > ((type *)fdomain)[n]) {                                \
        amp[band].min = ((type *)fdomain)[n];                                  \
    }                                                                          \
    if (amp[band].max < ((type *)fdomain)[n]) {                                \
        amp[band].max = ((type *)fdomain)[n];                                  \
    }                                                                          \
    amp[band].mean += ((type *)fdomain)[n];                                    \
    values[band]++;

#define BAND_AMP(band, thru)                                                   \
    if (realsize == 4) {                                                       \
        _BAND_AMP(band, thru, float);                                          \
    } else {                                                                   \
        _BAND_AMP(band, thru, double);                                         \
    }

int
eq_properties(const struct prepfilter *filter,
              struct vcu_eqamp amp[VCU_FREQ_ISO_3OCT_BANDS],
              double *peak_amp)
{
    int values[VCU_FREQ_ISO_3OCT_BANDS], len, n, i, fftsize, realsize;
    double thru[VCU_FREQ_ISO_3OCT_BANDS], rate, f, maxamp;
    void *fdomain, *tdomain, *fftplan;
    char path[PATH_MAX + 100];
    FILE *stream = NULL;
    
    if (filter == NULL) {
        elog(LOGDEBUG, "filter == NULL");
        return VCU_ERR_INVALID_VALUE;
    }
    realsize = filter->realsize;

    /* we only care to look at the first sample rate */
    rate = (double)filter->rates[0];
    len = filter->lengths[0];

    if (len < 1) {
        elog(LOGDEBUG, "invalid filter: %d", len);
        return VCU_ERR_INVALID_VALUE;
    }
    
    if (equaliser_debug) {
        for (i = 0; i < filter->n_rates; i++) {
            sprintf(path, "%sequaliser%d", log_path, filter->rates[i]);
            if ((stream = fopen(path, "wt")) == NULL) {
                continue;
            }
            if (realsize == 4) {
                for (n = 0; n < filter->lengths[i]; n++) {
                    fprintf(stream, "%.16e\n",
                            ((float *)filter->filters[i][0])[n]);
                }
            } else {
                for (n = 0; n < filter->lengths[i]; n++) {
                    fprintf(stream, "%.16e\n",
                            ((double *)filter->filters[i][0])[n]);
                }
            }
            fclose(stream);
        }
    }

    if (len < 32768) {
        fftsize = 32768;
        if ((tdomain = alignalloc(fftsize * realsize)) == NULL) {
            elog(LOGDEBUG, "memory allocation failed");
            return VCU_ERR_NO_MEMORY;
        }
        memcpy(tdomain, filter->filters[0][0], len * realsize);
        memset(&((uint8_t *)tdomain)[len * realsize], 0,
               (fftsize - len) * realsize);
    } else {
        fftsize = len;
        tdomain = filter->filters[0][0];
    }
    if ((fdomain = alignalloc(fftsize * realsize)) == NULL) {
        elog(LOGDEBUG, "memory allocation failed");
        return VCU_ERR_NO_MEMORY;
    }
    
    if (log2_get(fftsize) == -1) {
        if (realsize == 4) {
            fftplan = fftwf_plan_r2r_1d(fftsize, tdomain, fdomain,
                                        FFTW_R2HC, FFTW_ESTIMATE);
        } else {
            fftplan = fftw_plan_r2r_1d(fftsize, tdomain, fdomain,
                                       FFTW_R2HC, FFTW_ESTIMATE);
        }
    } else {
        fftplan = fftplan_get_real(log2_get(fftsize), false, false,
                                   realsize);
    }
    if (fftplan == NULL) {
        elog(LOGDEBUG, "failed to get fft plan for len %d", len);
        return VCU_ERR_NO_MEMORY;
    }
    if (realsize == 4) {
        fftwf_execute_r2r(fftplan, tdomain, fdomain);
        if (log2_get(fftsize) == -1) {
            fftwf_destroy_plan(fftplan);
        }
    } else {
        fftw_execute_r2r(fftplan, tdomain, fdomain);
        if (log2_get(fftsize) == -1) {
            fftw_destroy_plan(fftplan);
        }
    }
    dbabs_on_halfcomplex(fdomain, fftsize, realsize);

#if VCU_FREQ_ISO_3OCT_BANDS != 31
#error expecting 31 frequency bands
#endif    
    /*
      ISO Bands:
         20 -    31.5 -  47
         47 -    63 -    94
         94 -   125 -   188
        188 -   250 -   375
        375 -   500 -   750
        750 -  1000 -  1500
       1500 -  2000 -  3000  
       3000 -  4000 -  6000
       6000 -  8000 - 12000
      12000 - 16000 - 20000

      ISO 1/3 oct bands:

         18 -    20 -    23
         23 -    25 -    28
         28 -    31 -    36
         36 -    40 -    45
         45 -    50 -    57
         57 -    63 -    72
         72 -    80 -    90
         90 -   100 -   113
        113 -   125 -   143
        143 -   160 -   180
        180 -   200 -   225
        225 -   250 -   283
        283 -   315 -   358
        358 -   400 -   450
        450 -   500 -   565
        565 -   630 -   715
        715 -   800 -   900
        900 -  1000 -  1125
       1125 -  1250 -  1425
       1425 -  1600 -  1800
       1800 -  2000 -  2250
       2250 -  2500 -  2825
       2825 -  3150 -  3575
       3575 -  4000 -  4500
       4500 -  5000 -  5650
       5650 -  6300 -  7150
       7150 -  8000 -  9000
       9000 - 10000 - 11250
      11250 - 12500 - 14250
      14250 - 16000 - 18000
      18000 - 20000 - 22500
     */
    if (equaliser_debug) {
        sprintf(path, "%sequaliser_mag", log_path);
        stream = fopen(path, "wt");
    }
    memset(thru, 0, sizeof(thru));
    memset(values, 0, sizeof(values));
    if (amp != NULL) {
        memset(amp, 0, VCU_FREQ_ISO_3OCT_BANDS * sizeof(struct vcu_eqamp));
        for (n = 0; n < VCU_FREQ_ISO_3OCT_BANDS; n++) {
            amp[n].in = +100000;
            amp[n].max = -100000;
            amp[n].min = +100000;
            thru[n] = +100000;
        }
    }
    maxamp = -100000;
    for (n = 1; n < (fftsize >> 1); n++) {
        f = (double)n * rate / (double)fftsize;
        if (equaliser_debug && stream != NULL) {
            if (realsize == 4) {
                fprintf(stream, "%.16e %.16e\n", f,
                        ((float *)fdomain)[n]);
            } else {
                fprintf(stream, "%.16e %.16e\n", f,
                        ((double *)fdomain)[n]);
            }
        }
        if (realsize == 4) {
            if (((float *)fdomain)[n] > maxamp) {
                maxamp = ((float *)fdomain)[n];
            }
        } else {
            if (((double *)fdomain)[n] > maxamp) {
                maxamp = ((double *)fdomain)[n];
            }
        }
        if (amp != NULL) {
            if (f >= 18 && f < 23) {
                BAND_AMP(0, 20);
            } else if (f >= 23 && f < 28) {
                BAND_AMP(1, 25);
            } else if (f >= 28 && f < 36) {
                BAND_AMP(2, 31);
            } else if (f >= 36 && f < 45) {
                BAND_AMP(3, 40);
            } else if (f >= 45 && f < 57) {
                BAND_AMP(4, 50);
            } else if (f >= 57 && f < 72) {
                BAND_AMP(5, 63);
            } else if (f >= 72 && f < 90) {
                BAND_AMP(6, 80);
            } else if (f >= 90 && f < 113) {
                BAND_AMP(7, 100);
            } else if (f >= 113 && f < 143) {
                BAND_AMP(8, 125);
            } else if (f >= 143 && f < 180) {
                BAND_AMP(9, 160);
            } else if (f >= 180 && f < 225) {
                BAND_AMP(10, 200);
            } else if (f >= 225 && f < 283) {
                BAND_AMP(11, 250);
            } else if (f >= 283 && f < 358) {
                BAND_AMP(12, 315);
            } else if (f >= 358 && f < 450) {
                BAND_AMP(13, 400);
            } else if (f >= 450 && f < 565) {
                BAND_AMP(14, 500);
            } else if (f >= 565 && f < 715) {
                BAND_AMP(15, 630);
            } else if (f >= 715 && f < 900) {
                BAND_AMP(16, 800);
            } else if (f >= 900 && f < 1125) {
                BAND_AMP(17, 1000);
            } else if (f >= 1125 && f < 1425) {
                BAND_AMP(18, 1250);
            } else if (f >= 1425 && f < 1800) {
                BAND_AMP(19, 1600);
            } else if (f >= 1800 && f < 2250) {
                BAND_AMP(20, 2000);
            } else if (f >= 2250 && f < 2825) {
                BAND_AMP(21, 2500);
            } else if (f >= 2825 && f < 3575) {
                BAND_AMP(22, 3150);
            } else if (f >= 3575 && f < 4500) {
                BAND_AMP(23, 4000);
            } else if (f >= 4500 && f < 5650) {
                BAND_AMP(24, 5000);
            } else if (f >= 5650 && f < 7150) {
                BAND_AMP(25, 6300);
            } else if (f >= 7150 && f < 9000) {
                BAND_AMP(26, 8000);
            } else if (f >= 9000 && f < 11250) {
                BAND_AMP(27, 10000);
            } else if (f >= 11250 && f < 14250) {
                BAND_AMP(28, 12500);
            } else if (f >= 14250 && f < 18000) {
                BAND_AMP(29, 16000);
            } else if (f >= 18000 && f <= 22500) {
                BAND_AMP(30, 20000);
            }
        }
    }
    free(fdomain);
    if (equaliser_debug && stream != NULL) {
        fclose(stream);
    }

    if (amp != NULL) {
        for (n = 0; n < VCU_FREQ_ISO_3OCT_BANDS; n++) {
            if (values[n] != 0) {
                amp[n].mean /= (double)values[n];
            }
        }
    }
    if (peak_amp != NULL) {
        *peak_amp = maxamp;
    }
    return 0;
}

void
eq_init(const char _log_path[])
{
    strncpy(log_path, _log_path, PATH_MAX);
    log_path[PATH_MAX - 1] = '\0';
}

void *
eq_design_fft_filter(int *n_taps,
                     double sample_rate,
                     int block_size,
                     int n_bands,
                     const double bands[],
                     const double amps[],
                     int realsize)
{
    double mag, rad, curfreq, scale, divtaps, tapspi, *mags, *freqs;
    void *ifftplan, *buf;
    int n, i, len;
    
    if (*n_taps <= 0) {
        *n_taps = fft_filter_length(block_size, sample_rate, n_bands, bands);
        elog(LOGDEBUG, "fft filter length: %d taps", *n_taps);
    }

    len = *n_taps;
    scale = 1.0 / (double)len;
    divtaps = 1.0 / (double)len;
    tapspi = -(double)len * M_PI;
    mags = alloca((n_bands + 2) * sizeof(double));
    freqs = alloca((n_bands + 2) * sizeof(double));
    for (n = 0; n < n_bands; n++) {
        elog(LOGDEBUG, "fft filter band %d: %g Hz, %g dB",
             n, bands[n], amps[n]);
        mags[n+1] = pow(10.0, amps[n] / 20.0);
        freqs[n+1] = bands[n] / (double)sample_rate;
    }
    mags[0] = mags[1];
    mags[n+1] = mags[n];
    freqs[0] = 0.0;
    freqs[n+1] = 0.5;
    ifftplan = fftplan_get_real(log2_get(len), true, true, realsize);
    buf = alignalloc(len * realsize);
    if (buf == NULL || ifftplan == NULL) {
        elog(LOGDEBUG, "allocation failure");
        free(buf);
        return NULL;
    }

    mag = cosine_int(mags[0], mags[1], freqs[0], freqs[1], 0) * scale;
    if (realsize == 4) {
        ((float *)buf)[0] = mag;
    } else {
        ((double *)buf)[0] = mag;
    }
    for (n = 1, i = 0; n < len / 2; n++) {
        curfreq = (double)n * divtaps;
        if (curfreq > freqs[i+1]) {
            i++;
        }
        mag = cosine_int(mags[i], mags[i+1], freqs[i], freqs[i+1],
                         curfreq) * scale;
        rad = tapspi * curfreq;
        if (realsize == 4) {
            ((float *)buf)[n] = cos(rad) * mag;
            ((float *)buf)[len - n] = sin(rad) * mag;
        } else {
            ((double *)buf)[n] = cos(rad) * mag;
            ((double *)buf)[len - n] = sin(rad) * mag;
        }
    }
    if (realsize == 4) {
        ((float *)buf)[len / 2] =
            cosine_int(mags[i], mags[i+1], freqs[i], freqs[i+1], 1) * scale;
    } else {
        ((double *)buf)[len / 2] =
            cosine_int(mags[i], mags[i+1], freqs[i], freqs[i+1], 1) * scale;
    }

    /* convert to time-domain */
    if (realsize == 4) {
        fftwf_execute_r2r(ifftplan, buf, buf);
    } else {
        fftw_execute_r2r(ifftplan, buf, buf);
    }

    /*
    {
        FILE *stream;
        stream = fopen("log/fft_filter", "wt");
        if (stream != NULL) {
            for (n = 0; n < len; n++) {
                fprintf(stream, "%.16e\n", out[n]);
            }
            fclose(stream);
        }
    }
    */

    return buf;
}

void *
eq_design_whitenoise_filter(int *n_taps,
                            double sample_rate,
                            double length,
                            int realsize)
{
    static bool_t isinit = false;
    static uint32_t state[3];
    double phase, fftwscale;
    void *ifftplan, *filter;
    int len, n;
    
    if (!isinit) {
        tausinit(state, 0);
        isinit = true;
    }
    
    *n_taps = length * sample_rate;
    len = *n_taps;
    
    if (*n_taps < 4) {
        elog(LOGDEBUG, "too few taps: %d", *n_taps);
        return NULL;
    }
    if ((filter = alignalloc(len * realsize)) == NULL) {
        elog(LOGDEBUG, "memory allocation failed");
        return NULL;
    }

    if (log2_get(len) == -1) {
        if (realsize == 4) {
            ifftplan = fftwf_plan_r2r_1d(length, filter, filter,
                                         FFTW_HC2R, FFTW_ESTIMATE);
        } else {
            ifftplan = fftw_plan_r2r_1d(length, filter, filter,
                                        FFTW_HC2R, FFTW_ESTIMATE);
        }
    } else {
        ifftplan = fftplan_get_real(log2_get(len), true, true, realsize);
    }
    if (ifftplan == NULL) {
        free(filter);
        elog(LOGDEBUG, "failed to get fft plan for len %d", len);
        return NULL;
    }
    
    memset(filter, 0, len * realsize);
    fftwscale = 1.0 / (double)len;
    if (realsize == 4) {
        ((float *)filter)[0] = 0.0;
        for (n = 1; n < len / 2; n++) {
            phase = (double)(tausrand(state) % 10000000);
            ((float *)filter)[n] = fftwscale * cos(phase);
            ((float *)filter)[len - n] = fftwscale * sin(phase);
        }
        if ((len & 0x1) == 0) {
            ((float *)filter)[len / 2] = 0.0;
        }
        fftwf_execute_r2r(ifftplan, filter, filter);
        if (log2_get(len) == -1) {
            fftwf_destroy_plan(ifftplan);
        }
    } else {
        ((double *)filter)[0] = 0.0;
        for (n = 1; n < len / 2; n++) {
            phase = (double)(tausrand(state) % 10000000);
            ((double *)filter)[n] = fftwscale * cos(phase);
            ((double *)filter)[len - n] = fftwscale * sin(phase);
        }
        if ((len & 0x1) == 0) {
            ((double *)filter)[len / 2] = 0.0;
        }
        fftw_execute_r2r(ifftplan, filter, filter);
        if (log2_get(len) == -1) {
            fftw_destroy_plan(ifftplan);
        }
    }
    return filter;
}
