/*
 * (c) Copyright 2002 - 2006 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>

#include <bzlib.h>

#include "alignalloc.h"
#include "defs.h"
#include "elog.h"
#include "bit.h"
#include "grammar.h"
#include "filterprg.h"
#include "resample.h"
#include "simpletar.h"
#include "copyfile.h"
#include "reverb.h"
#include "md5.h"
#include "xstr.h"

/* BruteFIR includes */
#include <numunion.h>
#include <bfmod.h>
#include <swap.h>

#define fprintf(...)
#define bf_exit(...)

#define REALSIZE 4
#define RAW2REAL_NAME raw2realf
#include "raw2real.h"
#undef REALSIZE
#undef RAW2REAL_NAME

#define REALSIZE 8
#define RAW2REAL_NAME raw2reald
#include "raw2real.h"
#undef REALSIZE
#undef RAW2REAL_NAME

#undef bf_exit
#undef fprintf

static struct {
    struct filterprg *array;
    int cap;
    int pos;
} fprg_list = { NULL, 0, 0 };

static int
get_maxpos(const void *coeffs,
	   int length,
           int realsize)
{
    int n, maxpos = 0;
    float maxvalf = 0;
    double maxvald = 0;

    if (realsize == 4) {
        for (n = 0; n < length; n++) {
            if (maxvalf < fabsf(((const float *)coeffs)[n])) {
                maxvalf = fabsf(((const float *)coeffs)[n]);
                maxpos = n;
            }
        }
    } else {
        for (n = 0; n < length; n++) {
            if (maxvald < fabs(((const double *)coeffs)[n])) {
                maxvald = fabs(((const double *)coeffs)[n]);
                maxpos = n;
            }
        }
    }
    return maxpos;
}

/* Get resampler for the given rates. */
struct resampler {
    resample_t *resample;
    int inrate;
    int outrate;
    int used;
};
#define RESAMPLERS_QUEUE_SIZE 10
static resample_t *
get_resampler(int inrate,
	      int outrate,
              int realsize)
{
    static struct resampler resamplers[RESAMPLERS_QUEUE_SIZE];
    static bool_t first_call = true;
    int n, least_used, err;

    if (first_call) {
	memset(resamplers, 0, sizeof(resamplers));
	first_call = false;
    }
    /* check if we have a resampler allocated already */
    for (n = least_used = 0; n < RESAMPLERS_QUEUE_SIZE; n++) {
	if (resamplers[n].inrate == inrate &&
	    resamplers[n].outrate == outrate)
	{
	    break;
	}
	if (resamplers[least_used].used == 0 ||
	    resamplers[n].used < resamplers[least_used].used)
	{
	    least_used = n;
	}
    }

    /* allocate new resampler if necessary */
    if (n == RESAMPLERS_QUEUE_SIZE) {
	n = least_used;
	resample_delete(resamplers[n].resample);
	if ((resamplers[n].resample =
	     resample_new(inrate, outrate, realsize, &err)) == NULL)
	{
	    elog(LOGDEBUG, "resampling failed with error code %d", err);
	    return NULL;
	}
	resamplers[n].used = 0;
	resamplers[n].inrate = inrate;
	resamplers[n].outrate = outrate;
    }
    resamplers[n].used++;
    return resamplers[n].resample;
}

static char *
find_in_header(char header[],
               const char label[])
{
    bool_t in_str;
    char *p, *p1, oldp1;

    p = header;
    while ((p = strstr(p, label)) != NULL) {
        p1 = p;
        /* verify so it is not in a comment or string */
        while (p1 != header && *p1 != '\n' && *p1 != '#') {
            p1--;
        }
        if (*p1 == '#') {
            p += strlen(label);
            continue;
        }
        in_str = false;
        oldp1 = '\0';
        while (p1 != p) {
            if (*p1 == '\"' && oldp1 != '\\') {
                in_str = !in_str;
            }
            oldp1 = *p1;
            p1++;
        }
        if (!in_str) {
            break;
        }
        p += strlen(label);
    }
    return p;
}

static void
coeff_format_convert(void *dest,
		     void *src,
		     int length,
		     int format,
                     double scale,
                     int realsize)
{
    bool_t isfloat, lendian, swap;
    int bytes, sbytes, n;
    double zerodb;
    float scalef;

    isfloat = false;
    lendian = false;
    bytes = bf_sampleformat_size(format);
    switch (format) {
    case BF_SAMPLE_FORMAT_S8:
        sbytes = 1;
        zerodb = 128.0;
        break;
    case BF_SAMPLE_FORMAT_S16_LE:
    case BF_SAMPLE_FORMAT_S16_4LE:
        lendian = true;
    case BF_SAMPLE_FORMAT_S16_BE:
    case BF_SAMPLE_FORMAT_S16_4BE:
        sbytes = 2;
        zerodb = 32768.0;
        break;
    case BF_SAMPLE_FORMAT_S24_LE:
    case BF_SAMPLE_FORMAT_S24_4LE:
        lendian = true;
    case BF_SAMPLE_FORMAT_S24_BE:
    case BF_SAMPLE_FORMAT_S24_4BE:
        sbytes = 3;
        zerodb = 8388608.0;
        break;
    case BF_SAMPLE_FORMAT_S32_LE:
        lendian = true;
    case BF_SAMPLE_FORMAT_S32_BE:
        sbytes = 4;
        zerodb = 2147483648.0;
        break;
    case BF_SAMPLE_FORMAT_FLOAT_LE:
        lendian = true;
    case BF_SAMPLE_FORMAT_FLOAT_BE:
        sbytes = 4;
        isfloat = true;
        zerodb = 1.0;
        break;
    case BF_SAMPLE_FORMAT_FLOAT64_LE:
        lendian = true;
    case BF_SAMPLE_FORMAT_FLOAT64_BE:
        sbytes = 8;
        isfloat = true;
        zerodb = 1.0;
        break;
    default:
	elog(LOGERROR, "invalid format %d", format);
        return;
    }

#ifdef __LITTLE_ENDIAN__
    swap = !lendian;
#endif
#ifdef __BIG_ENDIAN__
    swap = lendian;
#endif
    
    /* this function comes from BruteFIR */
    scale = (1.0 / zerodb) * pow(10, scale / 20.0);
    if (realsize == 4) {
        raw2realf(dest, src, bytes, (bytes - sbytes) << 3, isfloat, 1, swap,
                  length);
        if (scale != 1.0) {
            scalef = (float)scale;
            for (n = 0; n < length; n++) {
                ((float *)dest)[n] *= scalef;
            }
        }
    } else {
        raw2reald(dest, src, bytes, (bytes - sbytes) << 3, isfloat, 1, swap,
                  length);
        if (scale != 1.0) {
            for (n = 0; n < length; n++) {
                ((double *)dest)[n] *= scale;
            }
        }
    }
}

/* Load one channel as specified by 'rate_index' and 'channel_index' form the
   given filter program. */
static void *
load_channel(const struct filterprg *fprg,
	     int rate_index,
	     int channel_index,
             int realsize,
	     int *actual_length)
{
    struct simple_tar_header tar_hdr;
    void *dest, *src = NULL;
    void *retval = NULL;
    int length, err, n;
    unsigned int tmp;
    FILE *stream;

    if ((stream = fopen(fprg->path, "rb")) == NULL) {
	elog(LOGDEBUG, "could not open file \"%s\" for reading: %s",
             fprg->path, strerror(errno));
	return NULL;
    }
    length = (int)(fprg->length * 1.05 * (double)fprg->rates[rate_index] *
		   (double)bf_sampleformat_size(fprg->format));
    if ((dest = malloc(length)) == NULL) {
	elog(LOGDEBUG, "could not allocate memory (%d bytes)", length);
	goto load_channel_return;
    }

    /* we skip the index file... */
    for (n = 0; n < rate_index * fprg->n_channels + channel_index + 1; n++) {
	if (!tar_header_read(stream, NULL, &tar_hdr)) {
	    elog(LOGDEBUG, "could not read tar header");
	    goto load_channel_return;
	}
	if ((fseek(stream, (tar_hdr.size + 511) & ~511, SEEK_CUR)) == -1) {
	    elog(LOGDEBUG, "fseek failed: %s", strerror(errno));
	    goto load_channel_return;
	}
    }
    if (!tar_header_read(stream, NULL, &tar_hdr)) {
        elog(LOGDEBUG, "could not read tar header");
        goto load_channel_return;
    }
    elog(LOGDEBUG, "found file \"%s\" (size %d) in tar archive",
         tar_hdr.filename, tar_hdr.size);
    if ((src = malloc(tar_hdr.size)) == NULL) {
	elog(LOGDEBUG, "could not allocate memory (%d bytes)", tar_hdr.size);
	goto load_channel_return;
    }
    if (fread(src, tar_hdr.size, 1, stream) != 1) {
	elog(LOGDEBUG, "could not read data: %s", strerror(errno));
	goto load_channel_return;
    }
    if ((n = strlen(tar_hdr.filename)) >= 4 &&
        strcmp(&tar_hdr.filename[n-4], ".bz2") == 0)
    {
        tmp = (unsigned int)length;
        if ((err = BZ2_bzBuffToBuffDecompress(dest, &tmp,
                                              src, tar_hdr.size, 0, 0)) !=
            BZ_OK)
        {
            elog(LOGDEBUG, "bz2 decompress failed with error code %d", err);
            goto load_channel_return;
        }
        free(src);
    } else {
        tmp = tar_hdr.size;
        dest = src;
    }
    src = NULL;
    tmp /= bf_sampleformat_size(fprg->format);
    if (tmp > (unsigned int)length) {
        elog(LOGDEBUG, "stored data is longer than specified length");
        goto load_channel_return;
    }
    length = (int)tmp;
    if (actual_length != NULL) {
	*actual_length = length;
    }    
    if ((retval = alignalloc(length * realsize)) == NULL)	{
        elog(LOGDEBUG, "could not allocate memory (%d bytes)",
             length * realsize);
        goto load_channel_return;	
    }
    /*
    {
        char dump_filename[299];
        FILE *stream;
        int n;
        sprintf(dump_filename, "log/lc%d", channel_index);
        stream = fopen(dump_filename, "wt");
        for (n = 0; n < trim_ends[0]; n++) {
            fprintf(stream, "%.16e\n", ((float *)dest)[n]);
        }
        fclose(stream);
    }
    */
    coeff_format_convert(retval, dest, length, fprg->format, fprg->scale,
                         realsize);
    free(dest);
    dest = NULL;
    
 load_channel_return:
    if (retval == NULL) {
        free(dest);
    }
    free(src);
    fclose(stream);
    return retval;
}

/* Prepare filters to fit the target sample rates */
static int
prepare_filter(const struct filterprg *fprg,
	       int channel_index,
	       int n_channels,
	       int n_rates,
	       const int rates[],
               int realsize,
	       void *filter[],
	       int lengths[],
	       void (*progress_callback)(int channel_index,
					 int n_channels,
					 int from_rate,
					 int to_rate,
                                         bool_t deriving),
               int progress_channel_index,
               int progress_n_channels)
{
    int src_lengths[FILTERPRG_RATES_MAX];
    int n, i, err, size, src_n_rates;
    void *data[FILTERPRG_RATES_MAX];
    int maxrate_index = 0;
    resample_t *rs = NULL;
    const int *src_rates;
    bool_t doresample;

    /* check which of filterprg and prepfilter should be used as source */
    if (fprg == NULL || n_rates < 0 || n_rates > FILTERPRG_RATES_MAX ||
	rates == NULL || filter == NULL || lengths == NULL)
    {
	elog(LOGDEBUG, "at least one parameter is invalid");
	return VCU_ERR_INVALID_VALUE;
    }
    src_n_rates = fprg->n_rates;
    src_rates = fprg->rates;

    /* point out the largest rate in source  */
    for (n = 1; n < src_n_rates; n++) {
	if (src_rates[n] > src_rates[maxrate_index]) {
	    maxrate_index = n;
	}
    }
    
    memset(filter, 0, n_rates * sizeof(void *));
    memset(data, 0, sizeof(data));
    
    for (n = 0; n < n_rates; n++) {
	/* see if we need to resample to get the target rate */
	for (i = 0; i < src_n_rates && src_rates[i] != rates[n]; i++);
	if (i == src_n_rates) {
	    doresample = true;
	    i = maxrate_index;
	} else {
	    doresample = false;
	}
	if (progress_callback != NULL) {
	    progress_callback(progress_channel_index,
			      progress_n_channels,
			      doresample ?
			      fprg->rates[maxrate_index] : rates[n],
			      rates[n],
                              false);
	}

	/* if source filter is not available, load it from disk */
	if (data[i] == NULL) {
	    elog(LOGDEBUG, "load channel %d...", channel_index);
	    if ((data[i] = load_channel(fprg, i, channel_index, realsize,
					&src_lengths[i])) == NULL)
	    {
                elog(LOGDEBUG, "load failed");
		err = VCU_ERR_LOAD_FAILED;
		goto prepare_filter_abort;
	    }
	    elog(LOGDEBUG, "finished");
	}
	/* resample if necessary */
	if (doresample) {
	    elog(LOGDEBUG, "resampling from %d Hz to %d Hz...",
		 fprg->rates[maxrate_index], rates[n]);
	    if ((rs = get_resampler(fprg->rates[maxrate_index],
				    rates[n], realsize)) == NULL)
	    {
                elog(LOGDEBUG, "could not get resampler");
		err = VCU_ERR_RESAMPLING_FAILED;
		goto prepare_filter_abort;
	    }
	    size = (int)((double)rates[n] / (double)src_rates[maxrate_index]
			 * (double)src_lengths[maxrate_index]) + 100;	    
	    if ((filter[n] = alignalloc(size * realsize)) == NULL) {
                elog(LOGDEBUG, "memory allocation failed");
		err = VCU_ERR_LOAD_FAILED;
		goto prepare_filter_abort;
	    }
	    memset(filter[n], 0, size * realsize);
	    if ((lengths[n] = resample(rs, data[maxrate_index], filter[n],
				       src_lengths[maxrate_index],
				       size, NULL, NULL)) == -1)
	    {		
                elog(LOGDEBUG, "resampling failed");
		err = VCU_ERR_RESAMPLING_FAILED;
		goto prepare_filter_abort;
	    }
	    elog(LOGDEBUG, "finished");
	} else {
            filter[n] = data[i];
            lengths[n] = src_lengths[i];
        }
    }

    err = 0;

 prepare_filter_abort:
    /* cleanup unused filter buffers */
    for (n = 0; n < src_n_rates; n++) {
	if (data[n] != NULL) {
	    for (i = 0; i < n_rates && filter[i] != data[n]; i++);
	    if (i == n_rates) {
		free(data[n]);
	    }
	}
    }

    if (err != 0) {
	for (n = 0; n < n_rates; n++) {
	    free(filter[n]);
	}
    }
    return err;
}

/* FIXME: sanity check all values parsed from the file */
static bool_t
lexparse_filterprg_header(struct filterprg *fprg)
{
    uint32_t isset[1] = { 0 }, required_fields, forbidden_fields;
    struct simple_tar_header tar_hdr;
    int tok, n, limit, n_files;
    FILE *stream;

    fprg->predelay = -1.0;
    fprg->n_sets = 1;
 restart_parse_loop:
    while ((tok = yylex()) > 0) {
	switch (tok) {
	case FIELD:
	    if (strcmp(yylval.field, "type") == 0) {
		TEST_AND_SET(0);
		GET_TOKEN(STRING);
		if (strcmp(yylval.string, "reverb b-format") == 0) {
		    fprg->type = FILTERPRG_TYPE_REVERB_BFORMAT;
		} else if (strcmp(yylval.string, "reverb traditional") == 0) {
		    fprg->type = FILTERPRG_TYPE_REVERB_TRADITIONAL;
		} else if (strcmp(yylval.string,
				  "crosstalk cancellation") == 0)
		{
		    fprg->type = FILTERPRG_TYPE_CROSSTALK_CANCELLATION;
		} else if (strcmp(yylval.string, "equalisation") == 0) {
		    fprg->type = FILTERPRG_TYPE_EQUALISATION;
		} else {
		    yy_perror("unknown type \"%s\"", yylval.string);
		    return false;
		}
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "comment") == 0) {
		TEST_AND_SET(1);
		GET_TOKEN(STRING);
		/* ignore this field (obsolete) */
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "name") == 0) {
		TEST_AND_SET(2);
		GET_TOKEN(STRING);
		strncpy(fprg->name, yylval.string, FILTERPRG_NAME_MAX);
		fprg->name[FILTERPRG_NAME_MAX - 1] = '\0';
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "longname") == 0) {
		TEST_AND_SET(3);
		GET_TOKEN(STRING);
		strncpy(fprg->longname, yylval.string, FILTERPRG_LONGNAME_MAX);
		fprg->longname[FILTERPRG_LONGNAME_MAX - 1] = '\0';
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "format") == 0) {
		TEST_AND_SET(4);
		GET_TOKEN(STRING);
		for (n = BF_SAMPLE_FORMAT_MIN; n <= BF_SAMPLE_FORMAT_MAX; n++) {
		    if (strcasecmp(bf_strsampleformat(n), yylval.string) == 0) {
			fprg->format = n;
			break;
		    }
		}
		if (n > BF_SAMPLE_FORMAT_MAX) {
		    yy_perror("unknown sample format \"%s\"", yylval.string);
		    return false;
		}
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "source_layout") == 0) {
		TEST_AND_SET(5);
		GET_TOKEN(STRING);
		if (strcmp(yylval.string, "center") == 0) {
		    fprg->source_layout = VCU_REVERB_SOURCE_LAYOUT_CENTER;
		} else if (strcmp(yylval.string, "stereo") == 0) {
		    fprg->source_layout = VCU_REVERB_SOURCE_LAYOUT_STEREO;
		} else if (strcmp(yylval.string, "lrc") == 0) {
		    fprg->source_layout = VCU_REVERB_SOURCE_LAYOUT_LRC;
		} else {
		    yy_perror("unknown source layout \"%s\"", yylval.string);
		    return false;
		}
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "sets") == 0) {
		TEST_AND_SET(6);
                GET_INTEGER();
                if ((fprg->n_sets = (int)yylval.real) < 1) {
                    yy_perror("invalid number of sets");
                    return false;
                }
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "scale") == 0) {
		TEST_AND_SET(7);
		GET_TOKEN(REAL);
                fprg->scale = yylval.real;
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "receivers") == 0) {
		TEST_AND_SET(8);
		for (n = 0, tok = COMMA;
                     n < FILTERPRG_CHANNELS_MAX && tok == COMMA;
                     n++)
                {
                    GET_TOKEN(REAL);
                    fprg->recv_azimuth[n] = yylval.real;
                    GET_TOKEN(SLASH);
                    GET_TOKEN(REAL);
                    fprg->recv_elevation[n] = yylval.real;
		    tok = yylex();
                    if (tok == SLASH) {
                        GET_TOKEN(REAL);
                        fprg->recv_distance[n] = yylval.real;
                        tok = yylex();
                    } else {
                        fprg->recv_distance[n] = 0.0;
                    }
                    if (fprg->recv_azimuth[n] > 360.0 ||
                        fprg->recv_azimuth[n] < -360.0 ||
                        fprg->recv_elevation[n] > 90.0 ||
                        fprg->recv_elevation[n] < -90.0)
                    {
                        yy_perror("invalid receiver angle");
                        return false;
                    }
                    if (fprg->recv_azimuth[n] > 180.0) {
                        fprg->recv_azimuth[n] -= 360.0;
                    }
                    if (fprg->recv_azimuth[n] < -180.0) {
                        fprg->recv_azimuth[n] += 360.0;
                    }
		}
		if (tok != EOS) {
		    yy_perror_tok(EOS, tok);
		    return false;
		}
                fprg->n_receivers = n;
	    } else if (strcmp(yylval.field, "rates") == 0) {
		TEST_AND_SET(9);
		limit = (fprg->n_rates != 0) ? fprg->n_rates :
		    FILTERPRG_RATES_MAX;
		for (n = 0, tok = COMMA; n < limit && tok == COMMA; n++) {
		    GET_INTEGER();
		    fprg->rates[n] = (int)yylval.real;
		    tok = yylex();
		}
		if (tok != EOS) {
		    yy_perror_tok(EOS, tok);
		    return false;
		}
		if (fprg->n_rates != 0 && n != fprg->n_rates) {
		    yy_perror("array length mismatch");
		    return false;
		}
		fprg->n_rates = n;
	    } else if (strcmp(yylval.field, "length") == 0) {
		TEST_AND_SET(10);
		GET_TOKEN(REAL);
		if ((fprg->length = yylval.real) < 0) {
		    yy_perror("negative length");
                    return false;
		}
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "predelay") == 0) {
		TEST_AND_SET(11);
		GET_TOKEN(REAL);
		if ((fprg->predelay = yylval.real) < 0) {
                    fprg->predelay = -1.0; /* unspecified predelay */
		}
		GET_TOKEN(EOS);
	    } else if (strcmp(yylval.field, "subtype") == 0) {
		TEST_AND_SET(12);
		GET_TOKEN(STRING);
		if (strcmp(yylval.string, "b-format-normalised") == 0 ||
                    strcmp(yylval.string, "b-format-normalized") == 0)
                {
		    fprg->subtype = FILTERPRG_SUBTYPE_BFORMAT_NORMALISED;
		} else if (strcmp(yylval.string, "b-format-3db") == 0) {
                    elog(LOGDEBUG, "\"b-format-3db\" is an obsolete subformat");
                } else {
		    yy_perror("unknown subtype \"%s\"", yylval.string);
		    return false;
		}
		GET_TOKEN(EOS);
	    } else {
		yy_pwarning("ignoring field \"%s\"", yylval.string);
		while ((tok = yylex()) != EOS && tok > 0);
		goto restart_parse_loop;
	    }
	    break;
	default:
	    yy_perror_tok(FIELD, tok);
	    return false;
	}
    }

    /* count files */
    if ((stream = fopen(fprg->path, "rb")) == NULL) {
	elog(LOGDEBUG, "could not open file \"%s\" for reading: %s",
             fprg->path, strerror(errno));
	return false;
    }
    elog(LOGDEBUG, "counting files...");
    for (n_files = 0;
         tar_header_read(stream, NULL, &tar_hdr);
         n_files++)
    {
        elog(LOGDEBUG, "%d: \"%s\" (size %d)", n_files, tar_hdr.filename,
             tar_hdr.size);
	if ((fseek(stream, (tar_hdr.size + 511) & ~511, SEEK_CUR)) == -1) {
	    elog(LOGDEBUG, "fseek failed: %s", strerror(errno));
            fclose(stream);
            return false;
	}
    }
    fclose(stream);
    n_files -= 1; /* don't count index file */
    elog(LOGDEBUG, "finished counting files");
    
    if (fprg->n_receivers == 0) {
        fprg->n_receivers = 1;
    }
    elog(LOGDEBUG, "n_receivers: %d; n_rates: %d; n_sets: %d; "
         "source_layout: %d;", fprg->n_receivers, fprg->n_rates, fprg->n_sets,
         fprg->source_layout);
    switch (fprg->type) {
    case FILTERPRG_TYPE_REVERB_TRADITIONAL:
        required_fields = 0x635;
        forbidden_fields = 0x0;
        fprg->n_channels = fprg->n_receivers;
        switch (fprg->source_layout) {
        case VCU_REVERB_SOURCE_LAYOUT_CENTER:
            break;
        case VCU_REVERB_SOURCE_LAYOUT_STEREO:
            fprg->n_channels *= 2;
            break;
        case VCU_REVERB_SOURCE_LAYOUT_LRC:
            fprg->n_channels *= 3;
            break;
        }
        fprg->n_channels *= fprg->n_sets;
        if (!reverb_test_receiver_directions(fprg->recv_azimuth,
                                             fprg->recv_elevation,
                                             fprg->n_receivers))
        {
            yy_perror("at least one invalid receiver direction");
            return false;
        }
        break;
    case FILTERPRG_TYPE_REVERB_BFORMAT:
	required_fields = 0x635;
        forbidden_fields = 0x0;
	fprg->n_channels = 4 * fprg->n_receivers;
        switch (fprg->source_layout) {
        case VCU_REVERB_SOURCE_LAYOUT_CENTER:
            break;
        case VCU_REVERB_SOURCE_LAYOUT_STEREO:
            fprg->n_channels *= 2;
            break;
        case VCU_REVERB_SOURCE_LAYOUT_LRC:
            fprg->n_channels *= 3;
            break;
        }
        fprg->n_channels *= fprg->n_sets;
	break;
    case FILTERPRG_TYPE_EQUALISATION:
	fprg->n_channels = 1;
	required_fields = 0x615;
        forbidden_fields = 0x160;
	break;
    case FILTERPRG_TYPE_CROSSTALK_CANCELLATION:
	if (4 * fprg->n_rates != n_files &&
	    2 * fprg->n_rates != n_files)
	{
	    yy_perror("expected %d or %d files, got %d",
		      2 * fprg->n_rates, 4 * fprg->n_rates, n_files);
	    return false;
	}
	fprg->n_channels = n_files / fprg->n_rates;
	required_fields = 0x615;
        forbidden_fields = 0x160;
	break;
    default:
	yy_perror("filter program type not specified");
	return false;
    }
    elog(LOGDEBUG, "0x%08X 0x%08X 0x%08X",
         isset[0], required_fields, forbidden_fields);
    if ((isset[0] & required_fields) != required_fields) {
	yy_perror("at least one required field was not set");
	return false;
    }
    if ((isset[0] & forbidden_fields) != 0) {
	yy_perror("at least one field not belonging to the "
                  "program type was set");
	return false;
    }
    if (fprg->longname[0] == '\0') {
	strcpy(fprg->longname, fprg->name);
    }
    if (fprg->type != FILTERPRG_TYPE_CROSSTALK_CANCELLATION &&
	fprg->n_channels * fprg->n_rates != n_files)
    {
	yy_perror("expected %d files, got %d",
		  fprg->n_channels * fprg->n_rates, n_files);
	return false;
    }    

    return true;
}

#undef TEST_AND_SET
#undef GET_TOKEN
#undef GET_INTEGER

static int
hexchar(uint8_t c)
{
    if (c >= '0' && c <= '9') {
	return c - '0';
    }
    if (c >= 'a' && c <= 'f') {
	return 10 + c - 'a';
    }
    if (c >= 'A' && c <= 'F') {
	return 10 + c - 'a';
    }
    return 0;
}

static char
hexencode(int n)
{
    if (n < 0 || n > 15) {
        return '#';
    }
    if (n < 10) {
        return '0' + n;
    }
    return 'a' + n - 10;
}

static int
scandir_select(const struct dirent *dirent)
{
    struct filterprg fprg;
    char path[PATH_MAX];
    int slen;

    slen = strlen(dirent->d_name);
    if (slen < 5 || slen > PATH_MAX - 100) {
        return 0;
    }
    if (strstr(dirent->d_name, ".tar") != dirent->d_name + slen - 4) {
	return 0;
    }
    getcwd(path, PATH_MAX);
    path[PATH_MAX-2] = '\0';
    strcat(path, PATH_SEPARATOR_STR);
    if (slen + strlen(path) >= PATH_MAX) {
        return 0;
    }
    strcat(path, dirent->d_name);
    
    if (!filterprg_scan_file(path, &fprg)) {
        return 0;
    }
    if (fprg_list.pos >= fprg_list.cap) {
	fprg_list.cap = fprg_list.pos + 1;
	fprg_list.array = realloc(fprg_list.array,
				  fprg_list.cap * sizeof(struct filterprg));
    }
    memcpy(&fprg_list.array[fprg_list.pos], &fprg, sizeof(struct filterprg));
    fprg_list.pos++;
    
    return 1;
}

bool_t
filterprg_scan_file(const char path[],
                    struct filterprg *fprg)
{
    struct simple_tar_header tar_hdr;
    uint8_t *header;
    FILE *stream;
    fpos_t pos;
    md5_t md5;
    bool_t b;
    
    if ((stream = fopen(path, "rb")) == NULL) {
	return false;
    }
    if (!tar_header_read(stream, NULL, &tar_hdr) ||
        strcmp(tar_hdr.filename, "filter.txt") != 0)
    {
	fclose(stream);
	return false;
    }

    memset(fprg, 0, sizeof(struct filterprg));
    if (tar_hdr.size > 65536) {
	fclose(stream);
	return false;
    }
    header = alloca(tar_hdr.size);
    if (fgetpos(stream, &pos) != 0) {
	fclose(stream);
	return 0;
    }
    if (fread(header, tar_hdr.size, 1, stream) != 1) {
	fclose(stream);
	return 0;
    }
    md5_init(&md5);
    md5_update(&md5, header, tar_hdr.size);
    md5_final(&md5, fprg->id.md5sum);
    fsetpos(stream, &pos);

    strncpy(fprg->path, path, PATH_MAX - 1);
    yy_restart(fprg->path, stream, LOGDEBUG);
    yy_input_mode(true, tar_hdr.size);
    b = lexparse_filterprg_header(fprg);
    fclose(stream);
    yy_restart(NULL, NULL, LOGDEBUG);

    return b;
}

struct filterprg *
filterprg_scan_directory(const char path[],
			 int *n_programs)
{
    struct dirent **namelist;
    char cwd[PATH_MAX];
    int n;

    getcwd(cwd, PATH_MAX);
    if (chdir(path) == -1) {
	elog(LOGDEBUG, "could not change directory to \"%s\": %s", path,
	     strerror(errno));
	return NULL;
    }
    fprg_list.pos = 0;
    /* Sometimes scandir fails by some strange reason, although there have been
       lots of successful reads of entries. This takes care of this problem */
    n = scandir(".", &namelist, scandir_select, NULL);
    *n_programs = fprg_list.pos;
    if (n > 0) {
        while(n--) {
            free(namelist[n]);
        }
        free(namelist);
    }
    chdir(cwd);
    return (*n_programs > 0) ? fprg_list.array : NULL;
}

int
filterprg_load(const struct filterprg *fprg,
	       struct prepfilter *prepf,
	       int delay[], /* may be NULL */
               int channel_list[],
               int base_channel_index,
               int n_channels,
	       const int rates[],
	       int n_rates,
               int realsize,
	       void (*progress_callback)(int channel_index,
					 int n_channels,
					 int from_rate,
					 int to_rate,
                                         bool_t deriving),
               int progress_channel_index_start,
               int progress_n_channels)

{
    int lengths[FILTERPRG_RATES_MAX];
    int n, i, err;
    char *s, *p;

    if (fprg == NULL || prepf == NULL || rates == NULL || n_rates < 0 ||
	n_rates > FILTERPRG_RATES_MAX || base_channel_index < 0 ||
        base_channel_index >= fprg->n_channels || n_channels < 1 ||
        base_channel_index + n_channels > fprg->n_channels)
    {
	elog(LOGDEBUG, "At least one parameter is invalid");
	return VCU_ERR_INVALID_VALUE;
    }
    if (channel_list == NULL) {
        elog(LOGDEBUG, "About to load channel indexes %d - %d from filter "
             "program \"%s\"",
             base_channel_index, base_channel_index + n_channels - 1,
             fprg->path);
    } else {
        s = alloca(n_channels * 10);
        p = s;
        for (n = 0; n < n_channels; n++) {
            sprintf(p, "%d, ", channel_list[n]);
            p += strlen(p);
        }
        p[-2] = '\0';
        elog(LOGDEBUG, "About to load channel indexes { %s } from filter "
             "program \"%s\"", s, fprg->path);
    }
    
    memset(prepf, 0, sizeof(struct prepfilter));
    prepf->n_rates = n_rates;
    memcpy(prepf->rates, rates, n_rates * sizeof(int));
    prepf->n_channels = n_channels;
    prepf->realsize = realsize;

    memset(lengths, 0, sizeof(lengths));
    for (n = 0; n < n_channels; n++) {
	if ((err = prepare_filter(fprg,
                                  channel_list == NULL ?
                                  base_channel_index + n :
                                  channel_list[n],
                                  fprg->n_channels, n_rates, rates,
                                  realsize, prepf->filters[n], lengths,
				  progress_callback,
                                  progress_channel_index_start < 0 ? n :
                                  progress_channel_index_start + n,
                                  progress_n_channels < 0 ? n_channels :
                                  progress_n_channels)) != 0)
	{
            elog(LOGDEBUG, "prepare filter failed");
	    goto filterprg_load_failed;
	}
	if (n == 0) {
	    memcpy(prepf->lengths, lengths, sizeof(lengths));
	} else {
	    for (i = 0; i < n_rates; i++) {
		if (lengths[i] != prepf->lengths[i]) {
                    elog(LOGDEBUG, "invalid length of filter: got %d "
                         "expected %d", lengths[i], prepf->lengths[i]);
		    err = VCU_ERR_INVALID_FORMAT;
		    goto filterprg_load_failed;
		}
	    }
	}
    }

    if (delay != NULL) {
	for (n = 0; n < n_rates; n++) {
	    switch (fprg->type) {
	    default:
	    case FILTERPRG_TYPE_EQUALISATION:
		delay[n] = get_maxpos(prepf->filters[0][n], prepf->lengths[n],
                                      prepf->realsize);
		break;
	    case FILTERPRG_TYPE_REVERB_TRADITIONAL:
                /* FIXME: get from impulse response closest to vector 0/0? */
                delay[n] = 0;
                break;
	    case FILTERPRG_TYPE_REVERB_BFORMAT:
		/* W channel */
		delay[n] = get_maxpos(prepf->filters[0][n], prepf->lengths[n],
                                      prepf->realsize);
		break;
	    case FILTERPRG_TYPE_CROSSTALK_CANCELLATION:
		/* direct-path channel */
		delay[n] = get_maxpos(prepf->filters[0][n], prepf->lengths[n],
                                      prepf->realsize);
		break;		
	    }
            elog(LOGDEBUG, "delay[%d]: %d", n, delay[n]);
	}
    }
    return 0;
    
 filterprg_load_failed:
    filterprg_free_prepfilter_contents(prepf);
    return err;
}

int
filterprg_copy(const struct filterprg *fprg,
               const char destname[],
               const char destlongname[],
               const char destpath[])
{
    char *header, *srcheader, *p, *p1, *path_storage;
    struct simple_tar_header tar_hdr;
    FILE *stream, *srcstream;
    const char *path;
    int pad, err;
    prgid_t id;
    md5_t md5;

    if (fprg == NULL || destname == NULL || destpath == NULL ||
        destlongname == NULL || strlen(destlongname) > 1000 ||
        strlen(destname) > 1000)
    {
        return VCU_ERR_INVALID_VALUE;
    }
    
    if ((srcstream = fopen(fprg->path, "rb")) == NULL) {
        return VCU_ERR_READ_FAILED;
    }
    if (!tar_header_read(srcstream, NULL, &tar_hdr)) {
	fclose(srcstream);
        return VCU_ERR_READ_FAILED;
    }
    if (tar_hdr.size > 65536) {
	fclose(srcstream);
        elog(LOGDEBUG, "invalid header size: %d", tar_hdr.size);
	return VCU_ERR_READ_FAILED;
    }
    header = alloca(tar_hdr.size + strlen(destname) + strlen(destlongname) + 1);
    srcheader = alloca(tar_hdr.size + strlen(destname) + 1);
    if (fread(srcheader, tar_hdr.size, 1, srcstream) != 1) {
	fclose(srcstream);
	return VCU_ERR_READ_FAILED;
    }
    fseek(srcstream, ((tar_hdr.size + 511) & ~511) - tar_hdr.size, SEEK_CUR);
    memcpy(header, srcheader, tar_hdr.size);
    srcheader[tar_hdr.size] = '\0';
    header[tar_hdr.size] = '\0';

    if ((p = find_in_header(header, "name:")) == NULL ||
        (p1 = strchr(find_in_header(srcheader, "name:"), ';')) == NULL)
    {
	fclose(srcstream);
	return VCU_ERR_READ_FAILED;
    }
    p += strlen("name:");
    sprintf(p, " \"%s\"", escape_string(NULL, destname));
    p += strlen(p);
    strcpy(p, p1);
    strcpy(srcheader, header);
    if ((p = find_in_header(header, "longname:")) == NULL ||
        (p1 = strchr(find_in_header(srcheader, "longname:"), ';')) == NULL)
    {
	fclose(srcstream);
	return VCU_ERR_READ_FAILED;
    }
    p += strlen("longname:");
    sprintf(p, " \"%s\"", escape_string(NULL, destlongname));
    p += strlen(p);
    strcpy(p, p1);
    elog(LOGDEBUG, "%d (%s)", strlen(header), header);

    tar_hdr.size = strlen(header);
    if (strstr(destpath, "%s") != NULL) {
        path_storage = alloca(strlen(destpath) + 100);
        md5_init(&md5);
        md5_update(&md5, (uint8_t *)header, tar_hdr.size);
        md5_final(&md5, id.md5sum);
        sprintf(path_storage, destpath, filterprg_prgid_ntoa(id));
        path = path_storage;
    } else {
        path = destpath;
    }
    if ((stream = fopen(path, "wb")) == NULL) {
        elog(LOGDEBUG, "could not open \"%s\" for writing: %s", destpath,
             strerror(errno));
	fclose(srcstream);
        return VCU_ERR_WRITE_FAILED;
    }
    if ((pad = tar_header_write(stream, NULL, &tar_hdr)) < 0) {
	fclose(srcstream);
	fclose(stream);
        return VCU_ERR_WRITE_FAILED;
    }
    /* assume success, if this fails, it fails in copystream anyway */
    fwrite(header, tar_hdr.size, 1, stream);
    if (pad > 0) {
        elog(LOGDEBUG, "pad %d bytes", pad);
        p = alloca(pad);
        memset(p, 0, pad);
        fwrite(p, pad, 1, stream);
    }    
    if ((err = copystream(stream, NULL, srcstream, NULL, -1, true)) != 0) {
        remove(path);
    }
    return err;
}

const char *
filterprg_prgid_ntoa(prgid_t id)
{
    static char md5sum[33];
    int n;

    for (n = 0; n < 16; n++) {
        md5sum[n << 1] = hexencode((id.md5sum[n] & 0xF0) >> 4);
        md5sum[(n << 1) + 1] = hexencode(id.md5sum[n] & 0x0F);
    }
    md5sum[32] = '\0';
    return md5sum;
}

prgid_t
filterprg_prgid_aton(const char s[])
{
    prgid_t id;
    int n;

    memset(&id, 0, sizeof(id));
    if (s == NULL) {
	return id;
    }
    for (n = 0; n < 16; n++){
	if (s[n<<1] == '\0' || s[(n<<1) + 1] == '\0') {
	    break;
	}
	id.md5sum[n] = (hexchar(s[n<<1]) << 4) +  hexchar(s[(n<<1) + 1]);
    }
    return id;
}

void
filterprg_free_prepfilter_contents(struct prepfilter *prepf)
{
    int n, i;
    
    if (prepf == NULL) {
	return;
    }
    for (n = 0; n < prepf->n_channels; n++) {
	for (i = 0; i < prepf->n_rates; i++) {
	    free(prepf->filters[n][i]);
            prepf->filters[n][i] = NULL;
	}	
    }
}

void
filterprg_debug_print_program(const struct filterprg *fprg)
{
    int n;
    
    elog(LOGDEBUG, "filterprg.id: %s", filterprg_prgid_ntoa(fprg->id));
    elog(LOGDEBUG, "filterprg.path: \"%s\"", fprg->path);
    elog(LOGDEBUG, "filterprg.type: %d", fprg->type);
    elog(LOGDEBUG, "filterprg.subtype: %d", fprg->subtype);
    elog(LOGDEBUG, "filterprg.format: %d", fprg->format);
    elog(LOGDEBUG, "filterprg.source_layout: %d", fprg->source_layout);
    elog(LOGDEBUG, "filterprg.n_channels: %d", fprg->n_channels);
    elog(LOGDEBUG, "filterprg.n_sets: %d", fprg->n_sets);
    elog(LOGDEBUG, "filterprg.n_receivers: %d", fprg->n_receivers);
    for (n = 0; n < fprg->n_receivers; n++) {
        elog(LOGDEBUG, "filterprg.recv_azimuth[%d]: %f", n,
             fprg->recv_azimuth[n]);
        elog(LOGDEBUG, "filterprg.recv_elevation[%d]: %f", n,
             fprg->recv_elevation[n]);
        elog(LOGDEBUG, "filterprg.recv_distance[%d]: %f", n,
             fprg->recv_distance[n]);
    }
    elog(LOGDEBUG, "filterprg.name: \"%s\"", fprg->name);
    elog(LOGDEBUG, "filterprg.longname: \"%s\"", fprg->longname);
    elog(LOGDEBUG, "filterprg.n_rates: %d", fprg->n_rates);
    for (n = 0; n < fprg->n_rates; n++) {
	elog(LOGDEBUG, "filterprg.rates[%d]: %d", n, fprg->rates[n]);
    }
    elog(LOGDEBUG, "filterprg.scale: %f", fprg->scale);
    elog(LOGDEBUG, "filterprg.predelay: %f", fprg->predelay);
    elog(LOGDEBUG, "filterprg.length: %f", fprg->length);
}
