/*
 * (c) Copyright 2002 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <string.h>

#include "idarray.h"
#include "elog.h"

struct _idarray_t_ {
    bool_t packed;
    int capacity;
    int elements;
    void **array;
};

idarray_t *
idarray_new(bool_t packed)
{
    idarray_t *idarray = malloc(sizeof(idarray_t));

    if (idarray == NULL) {
	return NULL;
    }
    memset(idarray, 0, sizeof(idarray_t));
    idarray->packed = packed;
    return idarray;
}

idarray_t *
idarray_new_ex(bool_t packed,
	       void *values[],
	       int length,
	       bool_t allocate)
{
    idarray_t *idarray = malloc(sizeof(idarray_t));
    int n;

    if (idarray == NULL) {
	return NULL;
    }
    memset(idarray, 0, sizeof(idarray_t));
    idarray->packed = packed;
    idarray->capacity = length;
    if (allocate) {
	idarray->array = malloc(length * sizeof(void *));
	memcpy(idarray->array, values, length * sizeof(void *));
    } else {
	idarray->array = values;
    }
    for (n = 0; n < length; n++) {
	if (idarray->array[n] != NULL) {
	    idarray->elements++;
	}
    }
    return idarray;
}

void
idarray_delete(idarray_t *idarray)
{
    if (idarray == NULL) {
	return;
    }
    free(idarray->array);
    free(idarray);
}

int
idarray_new_id(idarray_t *idarray,
	       void *value)
{
    void **array;
    int n;

    if (idarray == NULL || value == NULL) {
	return -1;
    }
    if (idarray->elements == idarray->capacity) {
	idarray->capacity += 16;
	if ((array = realloc(idarray->array,
			     idarray->capacity * sizeof(void *))) == NULL)
	{
	    idarray->capacity -= 16;
	    return -1;
	}
	memset(&array[idarray->elements], 0, 16 * sizeof(void *));
	idarray->array = array;
	idarray->array[idarray->elements] = value;
	idarray->elements++;
	return idarray->elements - 1;
    }
    for (n = 0; n < idarray->capacity; n++) {
	if (idarray->array[n] == NULL) {
	    idarray->array[n] = value;
	    idarray->elements++;
	    return n;
	}
    }
    return -1;
}

void *
idarray_delete_id(idarray_t *idarray,
		  int id)
{
    void *value;
    
    if (id < 0 || id >= idarray->capacity || idarray->array[id] == NULL) {
	return NULL;
    }
    value = idarray->array[id];
    idarray->array[id] = NULL;
    idarray->elements--;
    if (idarray->packed) {
	memmove(&idarray->array[id], &idarray->array[id + 1],
		(idarray->elements - id) * sizeof(void *));
        idarray->array[idarray->elements] = NULL;
    }
    return value;
}

int
idarray_length(const idarray_t *idarray)
{
    return idarray->elements;
}

void *
idarray_get(const idarray_t *idarray,
	    int id)
{
    if (id < 0 || id >= idarray->capacity) {
	return NULL;
    }
    return idarray->array[id];
}

void *
idarray_next(const idarray_t *idarray,
	     int *id)
{
    if (*id < -1) {
	*id = -1;
    }
    (*id) += 1;
    while (*id < idarray->capacity) {
	if (idarray->array[*id] != NULL) {
	    return idarray->array[*id];
	}
	(*id)++;
    }
    return NULL;
}

int
idarray_get_first_id(const idarray_t *idarray)
{
    int id;

    for (id = 0; id < idarray->capacity; id++) {
	if (idarray->array[id] != NULL) {
	    return id;
	}
    }
    return -1;
}

int
idarray_id2index(const idarray_t *idarray,
		 int id)
{
    int n, i;
    
    if (id < 0 || id >= idarray->capacity) {
	return -1;
    }
    for (n = i = 0; n < id; n++) {
	if (idarray->array[n] != NULL) {
	    i++;
	}
    }
    return i;
}
