/*
 * (c) Copyright 2002 - 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */

/*
 * Mouse protocol internals learnt from GPM
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "textgui.h"
#include "keyinput.h"
#include "elog.h"
#include "almusvcu.h"

#define AUX_SEND_ID    0xF2

#define AUX_SET_RES        0xE8  /* Set resolution */
#define AUX_SET_SCALE11    0xE6  /* Set 1:1 scaling */ 
#define AUX_SET_SCALE21    0xE7  /* Set 2:1 scaling */
#define AUX_GET_SCALE      0xE9  /* Get scaling factor */
#define AUX_SET_STREAM     0xEA  /* Set stream mode */
#define AUX_SET_SAMPLE     0xF3  /* Set sample rate */ 
#define AUX_ENABLE_DEV     0xF4  /* Enable aux device */
#define AUX_DISABLE_DEV    0xF5  /* Disable aux device */
#define AUX_RESET          0xFF  /* Reset aux device */
#define AUX_ACK            0xFA  /* Command byte ACK. */ 

#define MB_LEFT  1
#define MB_RIGHT 2
#define MB_UP    4
#define MB_DOWN  8

#define MOUSE_TYPE_UNDEFINED 0
#define MOUSE_TYPE_IMPS2 1

static char *device_path = NULL;
static int mouse_type = MOUSE_TYPE_UNDEFINED;

static bool_t
read_data(int fd,
          uint8_t *data,
          int count,
          int try_timeout)
{
    struct pollfd pollfd;
    int n, i = 0;

    do {
        memset(&pollfd, 0, sizeof(pollfd));
        pollfd.fd = fd;
        pollfd.events = POLLIN;
        if (poll(&pollfd, 1, try_timeout) == 0) {
            return false;
        }
	if ((n = read(fd, &data[i], 1)) < 1) {
	    if (n == 0 || errno != EINTR) {
		elog(LOGERROR, "reading from mouse failed: %s",
                     strerror(errno));
		return false;
	    }
	    continue;
	}
	i += n;
    } while (i != count);
    return true;
}

static int
event_imps2(int fd)
{
    static int buttons = 0;
    int prev_buttons;
    uint8_t data[4];

    if (!read_data(fd, data, 4, 10)) {
        elog(LOGERROR, "Failed to get mouse event");
        return KI_KEY_FAILURE;
    }
    
    prev_buttons = buttons;
    buttons = !!(data[0] & 1) * MB_LEFT | !!(data[0] & 2) * MB_RIGHT;
    
    if ((prev_buttons & MB_LEFT) && !(buttons & MB_LEFT)) {
        return KI_KEY_LEFT;
    }
    if ((prev_buttons & MB_RIGHT) && !(buttons & MB_RIGHT)) {
        return KI_KEY_RIGHT;
    }
    
    if (data[0] & 0xc0) {
        return KI_KEY_IGNORE;
    }
    
    if (data[3] == 0xff) {
        return KI_KEY_UP;
    }
    if (data[3] == 0x01) {
        return KI_KEY_DOWN;
    }
    
    return KI_KEY_IGNORE;
}

static int
init_imps2(int fd)
{
    static uint8_t s1[] = { 243, 200, 243, 100, 243, 80 };
    static uint8_t s2[] = { 246, 230, 244, 243, 100, 232, 3 };
    uint8_t c;
    int n;

    write(fd, s1, sizeof (s1));
    usleep(30000);
    write(fd, s2, sizeof (s2));
    usleep(30000);
    n = 0;
    while (read_data(fd, &c, 1, 0)) n++;
    elog(LOGDEBUG, "flushed %d bytes", n);
    return fd;
}

static int
keyin_init(void)
{
    int fd;

    if ((fd = open(device_path, O_RDWR | O_NONBLOCK)) == -1) {
	elog(LOGERROR, "Could not open mouse device \"%s\": %s",
             device_path, strerror(errno));
	return -1;
    }    
    switch (mouse_type) {
    case MOUSE_TYPE_IMPS2:
        return init_imps2(fd);
    }
    
    return -1;
}

static int
keyin_getkey(int fd,
	     int *repeats)
{
    *repeats = 0;
    switch (mouse_type) {
    case MOUSE_TYPE_IMPS2:
        return event_imps2(fd);
    }
    return KI_KEY_IGNORE;
}

const char *
vcu_name(void)
{
    return "ki_mouse";
}

static struct keyinput keyin_mouse = {
    keyin_init,
    keyin_getkey
};

void *
vcu_share(void)
{
    return (void *)&keyin_mouse;
}

#define GET_TOKEN(token, error)                                                \
    if (get_config_token(&lval) != token) {                                    \
        elog(LOGERROR, "Parse error: " error );                                \
        return false;                                                          \
    }

bool_t
vcu_preinit(int (*get_config_token)(union vcu_lexval *lexval))
{
    union vcu_lexval lval;
    int tok;

    while ((tok = get_config_token(&lval)) > 0) {
	switch (tok) {
	case VCU_LEXVAL_FIELD:
	    if (strcmp(lval.field, "type") == 0) {
		GET_TOKEN(VCU_LEXVAL_STRING, "expected string for type field");
                if (strcmp(lval.string, "imps2") == 0) {
                    mouse_type = MOUSE_TYPE_IMPS2;
                } else {
                    elog(LOGERROR, "Parse error: unknown mouse type \"%s\"",
                         lval.string);
                }
		GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
	    } else if (strcmp(lval.field, "device") == 0) {
		GET_TOKEN(VCU_LEXVAL_STRING, "expected string for "
                          "device field");
                device_path = strdup(lval.string);
		GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
            }
	    break;
	default:
	    elog(LOGERROR, "Parse error: expected field.");
	    return false;
	}
    }
    if (device_path == NULL) {
        elog(LOGERROR, "\"device\" not set");
        return false;
    }
    if (mouse_type == MOUSE_TYPE_UNDEFINED) {
        elog(LOGERROR, "\"type\" not set");
        return false;
    }
    return true;
}

#undef GET_TOKEN

