/*
 * (c) Copyright 2002 - 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/poll.h>

#include "defs.h"
#include "keyinput.h"
#include "elog.h"
#include "bit.h"

struct keyin {
    struct keyinput k;
    int spipe[2];
    int repeats;
    int key;
};

static struct keyin *inputs[FD_SETSIZE];
static uint32_t fdset[FD_SETSIZE/32+1];
static int maxfd;

bool_t
keyinput_init(struct keyinput _inputs[],
	      int n_inputs)
{
    int n, fd;

    memset(inputs, 0, sizeof(inputs));
    memset(fdset, 0, sizeof(fdset));
    maxfd = 0;
    for (n = 0; n < n_inputs; n++) {
	if ((fd = _inputs[n].init()) == -1) {
	    elog(LOGERROR, "Key input initialisation failed.");
	    return false;
	}
        if (fd > maxfd) {
            maxfd = fd;
        }
        bit_set(fdset, fd);
	inputs[fd] = malloc(sizeof(struct keyin));
        memset(inputs[fd], 0, sizeof(struct keyin));
        if (pipe(inputs[fd]->spipe) == -1) {
            elog(LOGERROR, "Failed to create pipe: %s.", strerror(errno));
            return false;
        }
        inputs[fd]->key = KI_KEY_IGNORE;
	memcpy(&inputs[fd]->k, &_inputs[n], sizeof(struct keyinput));
    }
    return true;
}

void
keyinput_setfds(fd_set *rfds)
{
    int fd;

    fd = -1;
    while ((fd = bit_find(fdset, fd + 1, maxfd)) != -1) {
        FD_SET(fd, rfds);
        FD_SET(inputs[fd]->spipe[0], rfds);
    }
}

int
keyinput_getkey(int fd,
                bool_t drop_queued)
{
    struct pollfd pollfd;
    int n;

    if (fd < 0 || fd >= FD_SETSIZE || inputs[fd] == NULL) {
        return KI_KEY_FAILURE;
    }
    if (inputs[fd]->repeats > 0) {
        if (drop_queued) {
            inputs[fd]->repeats = 1;
        }
	if (--inputs[fd]->repeats == 0) {
	    if (read(inputs[fd]->spipe[0], &n, 1) != 1) {
		elog(LOGERROR, "Read failed: %s.", strerror(errno));
	    }
	}
	return inputs[fd]->key;
    }
    inputs[fd]->key = inputs[fd]->k.getkey(fd, &inputs[fd]->repeats);
    if (inputs[fd]->key == KI_KEY_FAILURE) {
        return KI_KEY_FAILURE;
    }
    if (drop_queued) {
        memset(&pollfd, 0, sizeof(pollfd));
        pollfd.fd = fd;
        pollfd.events = POLLIN;
        while (poll(&pollfd, 1, 0) == 1) {
            n = inputs[fd]->k.getkey(fd, &n);
            if (n == KI_KEY_FAILURE) {
                return KI_KEY_FAILURE;
            }
        }
        inputs[fd]->repeats = 0;
    }
    if (inputs[fd]->repeats > 0) {
	if (write(inputs[fd]->spipe[1], &n, 1) != 1) {
	    elog(LOGERROR, "Write failed: %s.", strerror(errno));
	    inputs[fd]->repeats = 0;
	}
    }
    return inputs[fd]->key;
}

void
keyinput_flush_queue(int fd)
{
    struct pollfd pollfd;
    int n, max;
    
    if (fd >= FD_SETSIZE || (fd >= 0 && inputs[fd] == NULL)) {
        return;
    }

    if (fd < 0) {
        fd = -1;
        max = maxfd;
    } else {
        max = fd;
        fd--;
    }
    while ((fd = bit_find(fdset, fd + 1, max)) != -1) {    
        inputs[fd]->repeats = 0;
        memset(&pollfd, 0, sizeof(pollfd));
        pollfd.fd = fd;
        pollfd.events = POLLIN;
        while (poll(&pollfd, 1, 0) == 1) {
            n = inputs[fd]->k.getkey(fd, &n);
            if (n == KI_KEY_FAILURE) {
                return;
            }
        }
    }
}

