/*
 * (c) Copyright 2002 - 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

#include "simpletar.h"
#include "apui_defs.h"
#include "copyfile.h"
#include "elog.h"

bool_t
tar_header_read(FILE *stream,
                gzFile gzfile,
                struct simple_tar_header *tar_hdr)
{
    char header[512];

    if (gzfile != NULL) {
        if (gzread(gzfile, header, 512) != 512) {
            return false;
        }
    } else {
        if (fread(header, 512, 1, stream) != 1) {
            return false;
        }
    }
    memcpy(tar_hdr->filename, header, 100);
    if (tar_hdr->filename[99] != '\0' || tar_hdr->filename[0] == '\0') {
	return false;
    }
    if (header[135] != '\0' && header[135] != ' ') {
	return false;
    }
    tar_hdr->size = strtol(&header[124], NULL, 8);
    return true;
}

int
tar_header_write(FILE *stream,
                 gzFile gzfile,
                 const struct simple_tar_header *tar_hdr)
{
    char header[512];
    int i, sum;

    memset(header, 0, sizeof(header));
    strcpy(header, tar_hdr->filename);
    strcpy(&header[100], "0100644");
    strcpy(&header[108], "0000000");
    strcpy(&header[116], "0000000");
    sprintf(&header[124], "%.11o", tar_hdr->size);
    sprintf(&header[136], "%.11o", (uint32_t)time(NULL));
    strcpy(&header[148], "       ");
    strcpy(&header[155], " 0");
    strcpy(&header[257], "ustar  ");
    strcpy(&header[265], "root");
    strcpy(&header[297], "root");

    for(sum = i = 0; i < 512; i++) {
        sum += 0xFF & header[i];
    }
    sprintf(&header[148], "%.6o", sum);

    if (gzfile != NULL) {
        if (gzwrite(gzfile, header, 512) != 512) {
            return -1;
        }
    } else {
        if (fwrite(header, 512, 1, stream) != 1) {
            return -1;
        }
    }
    return TAR_PAD_SIZE(tar_hdr->size);
}
        
bool_t
tar_add(const char path[],
        FILE *stream,
        gzFile gzfile)
{
    struct simple_tar_header tar_hdr;
    FILE *srcstream;
    struct stat buf;
    uint8_t *padmem;
    const char *p;
    int pad, err;

    if (stat(path, &buf) == -1) {
        elog(LOGDEBUG, "stat on \"%s\" failed: %s",
             path, strerror(errno));
        return false;
    }
    tar_hdr.size = (int)buf.st_size;
    if (tar_hdr.size < 0) {
        elog(LOGDEBUG, "invalid size: %d", tar_hdr.size);
        return false;
    }
    if ((p = strrchr(path, PATH_SEPARATOR_CHR)) == NULL) {
        p = path;
    } else {
        p++;
    }
    strncpy(tar_hdr.filename, p, sizeof(tar_hdr.filename));
    tar_hdr.filename[sizeof(tar_hdr.filename)-1] = '\0';
    if ((pad = tar_header_write(stream, gzfile, &tar_hdr)) < 0) {        
        elog(LOGDEBUG, "failed to write header");
        return false;
    }
    if ((srcstream = fopen(path, "rb")) == NULL) {
        elog(LOGDEBUG, "failed to open \"%s\" for reading: %s",
             path, strerror(errno));
        return false;
    }
    if ((err = copystream(stream, gzfile, srcstream, NULL, -1, false)) != 0) {
        elog(LOGDEBUG, "copystream failed: %d", err);
        fclose(srcstream);
        return false;
    }
    fclose(srcstream);
    if (pad > 0) {
        elog(LOGDEBUG, "pad %d bytes", pad);
        padmem = alloca(pad);
        memset(padmem, 0, pad);
        if (gzfile != NULL) {
            if (gzwrite(gzfile, padmem, pad) != pad) {
                return false;
            }
        } else {
            if (fwrite(padmem, pad, 1, stream) != 1) {
                elog(LOGDEBUG, "fwrite failed: %s", strerror(errno));
                return false;
            }
        }
    }
    return true;
}
