/*
 * (c) Copyright 2002 - 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <sys/time.h>

#include "textwin.h"
#include "elog.h"

/* ncurses macros that conflicts function pointers */
#ifdef clear
#undef clear
#endif

#ifdef refresh
#undef refresh
#endif

#define SCROLLSTR_TABSIZE 256

struct scrollstr {
    char *str;
    int x;
    int y;
    int len;
    int pos;
    int diff;
    int timer;
    bool_t no_reverse;
    struct timeval interval;    
};

struct tw_screen {
    struct scrollstr *scrollstrtab[SCROLLSTR_TABSIZE];    
    char **chars;
};

static int n_outputs;
static struct textwin_output *outputs;
static char **dirchars;
static int xsize, ysize;
static bool_t isclear = false;
static char **cscr, **oscr, **scr[2];
static int (*register_timer)(void *param,
			     void (*callback)(void *param),
			     struct timeval *timeval);
static void (*unregister_timer)(int id);
static struct scrollstr *scrollstrtab[SCROLLSTR_TABSIZE];
static int scroll_id = 0;
static int n_scrollstrings = 0;
static bool_t print_scrollstrings = true;

static inline char
translate_char(char c,
	       int n)
{
    switch (c) {
    case TWC_LEFT:
	return dirchars[n][0];
    case TWC_RIGHT:
	return dirchars[n][1];
    case TWC_UP:
	return dirchars[n][2];
    case TWC_DOWN:
	return dirchars[n][3];
    default:
	return c;
    }
}

static inline void
translate_string(char s[],
		 int n)
{
    while (*s != '\0') {
	*s = translate_char(*s, n);
	s++;
    }
}

static void
scrollstr_callback(void *param)
{
    struct scrollstr *ss;
    struct timeval tv;
    int id, slen, n;

    id = (int)param;
    if ((ss = scrollstrtab[id]) == NULL) {
	return;
    }
    tv = ss->interval;
    if ((slen = strlen(ss->str)) <= ss->len) {
        if (print_scrollstrings) {
            tw_putstring(ss->x, ss->y, ss->str);
            for (n = slen; n < ss->len; n++) {
                tw_putchar(ss->x + n, ss->y, ' ');
            }
        }
    } else {
        if (print_scrollstrings) {
            for (n = 0; n < ss->len; n++) {
                tw_putchar(ss->x + n, ss->y, ss->str[ss->pos + n]);
            }
        }
	if (ss->pos + ss->len == slen) {
	    if (ss->no_reverse) {
		ss->diff = -ss->pos;
	    } else {
		ss->diff = -1;
	    }
	    timeradd(&tv, &tv, &tv);
	    timeradd(&tv, &tv, &tv);
	} else if (ss->pos == 0) {
	    ss->diff = 1;
	    timeradd(&tv, &tv, &tv);
	    timeradd(&tv, &tv, &tv);
	}
	ss->pos += ss->diff;
    }
    tw_refresh();
    ss->timer = register_timer((void *)id, scrollstr_callback, &tv);
}

bool_t
tw_init(struct textwin_output _outputs[],
	int _n_outputs,
	int _xsize,
	int _ysize,	
	int (*_register_timer)(void *param,
			       void (*callback)(void *param),
			       struct timeval *timeval),
	void (*_unregister_timer)(int id))
{
    int xsize_range[2], ysize_range[2], n, i;
    char c;

    memset(scrollstrtab, 0, sizeof(scrollstrtab));
    
    xsize = _xsize;
    ysize = _ysize;
    n_outputs = _n_outputs;
    register_timer = _register_timer;
    unregister_timer = _unregister_timer;
    for (n = 0, c = ' '; n < 2; n++, c = 'x') {
	scr[n] = malloc(ysize * sizeof(char *));
	for (i = 0; i < ysize; i++) {
	    scr[n][i] = malloc(xsize * sizeof(char));
	    memset(scr[n][i], c, xsize * sizeof(char));
	}
    }
    cscr = scr[0];
    oscr = scr[1];
    outputs = malloc(n_outputs * sizeof(struct textwin_output));
    memcpy(outputs, _outputs, n_outputs * sizeof(struct textwin_output));
    dirchars = malloc(n_outputs * sizeof(char *));
    for (n = 0; n < n_outputs; n++) {
	dirchars[n] = malloc(4 * sizeof(char));
	outputs[n].properties(xsize_range,
			      ysize_range,
			      dirchars[n]);
	if (xsize_range[0] > xsize ||
	    xsize_range[1] < xsize ||
	    ysize_range[0] > ysize ||
	    ysize_range[1] < ysize)
	{
	    elog(LOGERROR, "size is out of range for output index %d "
                 "(given: %d/%d, range: %d-%d/%d-%d)\n",
		 n, xsize, ysize, xsize_range[0], xsize_range[1],
		 ysize_range[0], ysize_range[1]);
	    return false;
	}
	if (outputs[n].init != NULL) {
	    if (!outputs[n].init(xsize, ysize)) {
		elog(LOGERROR, "failed to initialise output index %d\n", n);
		return false;
	    }
	}
    }
    tw_clear();
    tw_refresh();
    return true;
}

int
tw_xsize(void)
{
    return xsize;
}

int
tw_ysize(void)
{
    return ysize;
}

void
tw_clear(void)
{
    int y;

    isclear = true;
    for (y = 0; y < ysize; y++) {
	memset(cscr[y], ' ', xsize * sizeof(char));
    }
    for (y = 0; y < SCROLLSTR_TABSIZE; y++) {
	if (scrollstrtab[y] != NULL) {
	    tw_del_scrollstring(y);
	}
    }
}

char
tw_getchar(int x,
	   int y)
{
    if (x < 0 || x >= xsize || y < 0 || y >= ysize) {
	return '\0';
    }
    return cscr[y][x];
}

void
tw_putchar(int x,
	   int y,
	   char c)
{
    if (x < 0 || x >= xsize || y < 0 || y >= ysize) {
	return;
    }
    isclear = false;
    cscr[y][x] = c;
}

void
tw_putstring(int x,
	     int y,
	     const char s[])
{
    int len;
    
    if (x < 0 || x >= xsize || y < 0 || y >= ysize || s == NULL) {
	return;
    }
    len = strlen(s);
    if (x + len >= xsize) {
	len = xsize - x;
    }
    if (len > 0) {
	isclear = false;
	memcpy(&cscr[y][x], s, len * sizeof(char));
    }
}

void
tw_refresh(void)
{
    bool_t change = false;
    int n, x, y, i, ix = 0;
    char *s;
    
    s = alloca((xsize + 1) * sizeof(char));
    for (n = 0; n < n_outputs; n++) {
	if (isclear && outputs[n].clear != NULL) {
	    outputs[n].clear();
	    change = true;
	    continue;
	}
	for (y = 0; y < ysize; y++) {
	    if (outputs[n].putstring != NULL) {
		for (x = 0, i = 0; x < xsize; x++) {
		    if (cscr[y][x] != oscr[y][x]) {
			if (i == 0) {
			    ix = x;
			}
			s[i++] = cscr[y][x];
		    } else if (i > 0) {
			s[i] = '\0';
			translate_string(s, n);
			outputs[n].putstring(ix, y, s);
			i = 0;
			change = true;
		    }
		}
		if (i > 0) {
		    s[i] = '\0';
		    translate_string(s, n);
		    outputs[n].putstring(ix, y, s);
		    i = 0;
		    change = true;
		}
	    } else {
		for (x = 0; x < xsize; x++) {
		    if (cscr[y][x] != oscr[y][x]) {			
			outputs[n].putchar(x, y,
					   translate_char(cscr[y][x], n));
			change = true;
		    }
		}
	    }
	}
    }
    if (change) {
	for (y = 0; y < ysize; y++) {
	    memcpy(oscr[y], cscr[y], xsize * sizeof(char));
	}
	for (n = 0; n < n_outputs; n++) {
	    if (outputs[n].refresh != NULL) {
		outputs[n].refresh();
	    }
	}
    }
}

void
tw_printf(int x,
	  int y,
	  const char format[],
	  ...)
{
    va_list arg;
    char s[xsize + 1];

    va_start(arg, format);
    vsnprintf(s, xsize + 1, (const char *)format, arg);
    va_end(arg);
    tw_putstring(x, y, s);
}

void
tw_del_scrollstring(int id)
{
    if (id < 0 || id >= SCROLLSTR_TABSIZE || scrollstrtab[id] == NULL) {
	return;
    }
    unregister_timer(scrollstrtab[id]->timer);
    free(scrollstrtab[id]->str);
    free(scrollstrtab[id]);
    scrollstrtab[id] = NULL;
    n_scrollstrings--;
}

void
tw_del_scrollstring_line(int y)
{
    int id;
    
    for (id = 0; id < SCROLLSTR_TABSIZE; id++) {
	if (scrollstrtab[id] != NULL && scrollstrtab[id]->y == y) {
	    tw_del_scrollstring(id);
	}
    }    
}

int
tw_get_scrollstring_pos(int id,
			int *direction)
{
    if (id < 0 || id >= SCROLLSTR_TABSIZE || scrollstrtab[id] == NULL) {
	return -1;
    }
    if (direction != NULL) {
	*direction = scrollstrtab[id]->diff;
    }
    return scrollstrtab[id]->pos;
}

int
tw_add_scrollstring(int x,
		    int y,
		    const char s[],
		    int len,
		    int startpos,
		    int direction,
		    bool_t no_reverse)
{
    struct scrollstr *ss;
    int id;

    if (n_scrollstrings == SCROLLSTR_TABSIZE / 2) {
        elog(LOGDEBUG, "too many scrollstrings");
        return -1;
    }
    id = scroll_id++;    
    if (scroll_id == SCROLLSTR_TABSIZE) {
        scroll_id = 0;
    }
    if (startpos < 0 || strlen(s) <= len) {
	startpos = 0;
    } else if (startpos + len > strlen(s)) {
	startpos = strlen(s) - len;
    }
    
    ss = malloc(sizeof(struct scrollstr));
    ss->pos = startpos;
    ss->x = x;
    ss->y = y;
    ss->len = len;
    ss->str = strdup(s);
    ss->no_reverse = !!no_reverse;
    if (direction >= 0) {
	ss->diff = 1;
    } else {
	ss->diff = -1;
    }
    ss->timer = -1;
    ss->interval.tv_sec = 0;
    ss->interval.tv_usec = 300 * 1000;

    scrollstrtab[id] = ss;
    scrollstr_callback((void *)id);

    n_scrollstrings++;
    return id;
}

void
tw_print_scrollstrings(bool_t do_print)
{
    print_scrollstrings = !!do_print;
}

struct tw_screen *
tw_store_screen(void)
{
    struct tw_screen *screen;
    int y;

    screen = malloc(sizeof(struct tw_screen));
    memset(screen->scrollstrtab, 0, sizeof(scrollstrtab));
    for (y = 0; y < SCROLLSTR_TABSIZE; y++) {
        if (scrollstrtab[y] != NULL) {
            screen->scrollstrtab[y] = malloc(sizeof(struct scrollstr));
            memcpy(screen->scrollstrtab[y], scrollstrtab[y],
                   sizeof(struct scrollstr));
            screen->scrollstrtab[y]->str =
                strdup(scrollstrtab[y]->str);
        }
    }
    screen->chars = malloc(ysize * sizeof(char *));
    for (y = 0; y < ysize; y++) {
        screen->chars[y] = malloc(xsize * sizeof(char));
        memcpy(screen->chars[y], cscr[y], xsize * sizeof(char));
    }
    return screen;
}

void
tw_restore_screen(const struct tw_screen *screen)
{
    int y;

    for (y = 0; y < ysize; y++) {
        memcpy(cscr[y], screen->chars[y], xsize * sizeof(char));
    }
    for (y = 0; y < SCROLLSTR_TABSIZE; y++) {
        if (screen->scrollstrtab[y] != NULL) {
            tw_add_scrollstring(screen->scrollstrtab[y]->x,
                                screen->scrollstrtab[y]->y,
                                screen->scrollstrtab[y]->str,
                                screen->scrollstrtab[y]->len,
                                screen->scrollstrtab[y]->pos,
                                screen->scrollstrtab[y]->diff,
                                screen->scrollstrtab[y]->no_reverse);
        }
    }
}

void
tw_delete_screen(struct tw_screen *screen)
{
    int y;

    if (screen == NULL) {
        return;
    }
    for (y = 0; y < ysize; y++) {
        free(screen->chars[y]);
    }
    for (y = 0; y < SCROLLSTR_TABSIZE; y++) {
        if (screen->scrollstrtab[y] != NULL) {
            free(screen->scrollstrtab[y]->str);
            free(screen->scrollstrtab[y]);
        }
    }
    free(screen->chars);
    free(screen);
}

void
tw_brightness(int value)
{
    int n;

    if (value < 1 || value > 4) {
	return;
    }
    
    for (n = 0; n < n_outputs; n++) {
	if (outputs[n].brightness != NULL) {
	    outputs[n].brightness(value);
	}
    }
}
